Package: neuromapr
Title: Spatial Null Models and Transforms for Brain Map Comparison
Version: 0.2.2
Authors@R: c(
    person("Athanasia Mo", "Mowinckel", , "a.m.mowinckel@psykologi.uio.no",
           role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0002-5756-0223"))
    )
Description: Implements spatial null models and coordinate-space
    transformations for statistical comparison of brain maps, following the
    framework described in Markello et al. (2022)
    <doi:10.1038/s41592-022-01625-w>. Provides variogram-matching
    surrogates (Burt et al. 2020), Moran spectral randomization (Wagner &
    Dray 2015), and spin-based permutation tests (Alexander-Bloch et al.
    2018). Includes an R interface to the 'neuromaps' annotation registry
    for browsing, downloading, and comparing brain map annotations from the
    Open Science Framework ('OSF').
    Integrates with 'ciftiTools' for coordinate-space transforms.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
Config/testthat/parallel: true
Depends: R (>= 4.2.0)
Imports: cli, dplyr, ggplot2, gifti, httr2, igraph, lifecycle, rlang,
        stats, tibble, tools, withr
Suggests: ciftiTools, clue, freesurferformats, jsonlite, knitr,
        rmarkdown, RNifti, RSpectra, testthat (>= 3.0.0), vcr (>=
        1.0.0)
VignetteBuilder: knitr
URL: https://github.com/lcbc-uio/neuromapr,
        https://lcbc-uio.github.io/neuromapr/,
        https://doi.org/10.5281/zenodo.18648937
BugReports: https://github.com/lcbc-uio/neuromapr/issues
NeedsCompilation: no
Packaged: 2026-04-29 17:55:05 UTC; athanasm
Author: Athanasia Mo Mowinckel [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-5756-0223>)
Maintainer: Athanasia Mo Mowinckel <a.m.mowinckel@psykologi.uio.no>
Repository: CRAN
Date/Publication: 2026-04-30 07:50:03 UTC
Built: R 4.6.0; ; 2026-05-01 21:42:16 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.matrix,null_distribution)
S3method(plot,null_distribution)
S3method(print,neuromaps_enhanced_comparison)
S3method(print,null_distribution)
S3method(summary,null_distribution)
export(annot_to_gifti)
export(check_wb_command)
export(clear_neuromaps_cache)
export(compare_maps)
export(fetch_neuromaps_annotation)
export(fsmorph_to_gifti)
export(generate_nulls)
export(get_parcel_centroids)
export(get_surface_distance)
export(make_surf_graph)
export(neuromaps_available)
export(new_null_distribution)
export(null_alexander_bloch)
export(null_baum)
export(null_burt2018)
export(null_burt2020)
export(null_cornblath)
export(null_moran)
export(null_spin_hungarian)
export(null_spin_vasa)
export(parcellate)
export(parcels_to_vertices)
export(permtest_metric)
export(read_brain_map_values)
export(resample_images)
export(transform_to_space)
export(unparcellate)
export(vertex_areas)
export(vertices_to_parcels)
importFrom(lifecycle,deprecated)
importFrom(rlang,.data)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# neuromapr <a href="https://lcbc-uio.github.io/neuromapr/"><img src="man/figures/logo.png" align="right" height="138" alt="neuromapr website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/lcbc-uio/neuromapr/actions/workflows/R-CMD-check.yml/badge.svg)](https://github.com/lcbc-uio/neuromapr/actions/workflows/R-CMD-check.yml)
[![code-quality](https://github.com/lcbc-uio/neuromapr/actions/workflows/code-quality.yaml/badge.svg)](https://github.com/lcbc-uio/neuromapr/actions/workflows/code-quality.yaml)
[![test-coverage](https://github.com/lcbc-uio/neuromapr/actions/workflows/test-coverage.yaml/badge.svg)](https://github.com/lcbc-uio/neuromapr/actions/workflows/test-coverage.yaml)
[![Cross-validate](https://github.com/lcbc-uio/neuromapr/actions/workflows/cross-validate.yml/badge.svg)](https://github.com/lcbc-uio/neuromapr/actions/workflows/cross-validate.yml)
[![DOI](https://zenodo.org/badge/1156352155.svg)](https://doi.org/10.5281/zenodo.18648936)
[![CRAN
status](https://www.r-pkg.org/badges/version/neuromapr)](https://CRAN.R-project.org/package=neuromapr)
<!-- badges: end -->

Brain maps are spatially autocorrelated — nearby regions have similar
values just because they are neighbours. Standard correlation tests
ignore this and produce inflated p-values. neuromapr generates
spatially-constrained surrogate maps that preserve autocorrelation
structure, so you can test whether the similarity between two brain maps
is more than a spatial artifact.

The package implements the framework described in [Markello et
al. (2022)](https://doi.org/10.1038/s41592-022-01625-w) and provides
eight null model methods, parcellation utilities, geodesic surface
distance, and coordinate-space transforms.

This package was co-developed with [Claude
Code](https://platform.claude.com/docs/en/docs/agents-and-tools/claude-code/overview)
(Anthropic’s Claude Opus 4.6). While the test suite is thorough and
defaults are aligned with the
[neuromaps](https://netneurolab.github.io/neuromaps/) Python reference
implementation, discrepancies may still exist at this early stage. We
encourage users to cross-validate results against
[neuromaps](https://github.com/netneurolab/neuromaps) and to [report any
issues](https://github.com/lcbc-uio/neuromapr/issues).

## Installation

Install from CRAN:

``` r
install.packages("neuromapr")
```

Or install from the lcbc-uio
[r-universe](https://lcbc-uio.r-universe.dev/):

``` r
options(repos = c(
  lcbcuio = "https://lcbc-uio.r-universe.dev",
  CRAN = "https://cloud.r-project.org"
))

install.packages("neuromapr")
```

Or install the development version from GitHub:

``` r
# install.packages("pak")
pak::pak("lcbc-uio/neuromapr")
```

## Quick start

``` r
library(neuromapr)
```

Compare two brain maps with a spatial null model in three lines:

``` r
set.seed(42)
n <- 100
distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))

map_a <- rnorm(n)
map_b <- 0.4 * map_a + rnorm(n, sd = 0.8)

compare_maps(
  map_a,
  map_b,
  null_method = "burt2020",
  distmat = distmat,
  n_perm = 500L,
  seed = 1,
  verbose = FALSE
)
#> Generating burt2020 nulls ■■■■■■■■■■                        29% | ETA:  6s
#> Generating burt2020 nulls ■■■■■■■■■■■■■■■■■■■■■             66% | ETA:  4s
#> 
#> ── Brain Map Comparison
#> Method: pearson
#> r = 0.5302, p = 1.4e-08
#> n = 100
#> ────────────────────────────────────────────────────────────────────────────────
#> Null model: burt2020 (500 permutations)
#> p_null = 0.002
```

The `p_null` accounts for spatial autocorrelation. The parametric `p`
does not.

## Null model methods

neuromapr provides eight methods across three families:

| Family         | Method                      | Input required               |
|----------------|-----------------------------|------------------------------|
| Distance-based | `burt2020` (variogram)      | Distance matrix              |
|                | `burt2018` (SAR model)      | Distance matrix              |
|                | `moran` (spectral)          | Distance matrix              |
| Spin-based     | `alexander_bloch`           | Sphere coordinates           |
|                | `spin_vasa` (greedy)        | Sphere coordinates           |
|                | `spin_hungarian` (optimal)  | Sphere coordinates           |
| Parcel spin    | `baum` (max overlap)        | Sphere coords + parcellation |
|                | `cornblath` (majority vote) | Sphere coords + parcellation |

All methods are accessed through `generate_nulls()` or directly via
their dedicated functions (e.g., `null_burt2020()`, `null_moran()`).

## Key features

**Parcellation utilities** — aggregate vertex data into parcels, map
parcel values back to vertices, and compute parcel centroids using
average, surface, or geodesic methods.

``` r
vertex_data <- rnorm(1000)
labels <- rep(1:10, each = 100)

parcel_means <- vertices_to_parcels(vertex_data, labels)
head(parcel_means)
#>           1           2           3           4           5           6 
#>  0.11795689 -0.03426519 -0.05073563  0.04712785  0.06088535 -0.04200267
```

**Geodesic surface distance** — shortest-path distances along a
triangular mesh, for analyses where Euclidean distance through the brain
does not reflect cortical proximity.

``` r
vertices <- matrix(
  c(0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 1, 0),
  ncol = 3,
  byrow = TRUE
)
faces <- matrix(c(1, 2, 3, 2, 3, 4), ncol = 3, byrow = TRUE)

get_surface_distance(vertices, faces)
#>      [,1]     [,2]     [,3] [,4]
#> [1,]    0 1.000000 1.000000    2
#> [2,]    1 0.000000 1.414214    1
#> [3,]    1 1.414214 0.000000    1
#> [4,]    2 1.000000 1.000000    0
```

**Custom metric testing** — `permtest_metric()` runs a permutation test
with any metric function, optionally using spatial null surrogates:

``` r
result <- permtest_metric(
  map_a,
  map_b,
  metric_func = function(x, y) mean(abs(x - y)),
  n_perm = 200L,
  seed = 1
)
result$p_value
#> [1] 1
```

**Format conversions** — convert FreeSurfer `.annot` and morphometry
files to GIFTI with `annot_to_gifti()` and `fsmorph_to_gifti()`.

## Citation

If you use neuromapr, please cite the package and the neuromaps
framework it implements. You can get the package citation from R with:

``` r
citation("neuromapr")
```

    #> Warning in citation("neuromapr"): could not determine year for 'neuromapr' from
    #> package DESCRIPTION file

Mowinckel A (????). *neuromapr: Spatial Null Models and Transforms for
Brain Map Comparison*. R package version 0.2.1,
<https://github.com/lcbc-uio/neuromapr>.

<https://doi.org/10.5281/zenodo.18648937>

> Markello RD, Hansen JY, Liu Z-Q, et al. (2022). neuromaps: structural
> and functional interpretation of brain maps. *Nature Methods*, 19,
> 1472–1480. <doi:10.1038/s41592-022-01625-w>

Individual null model methods have their own citations documented in
each function’s help page.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)

## ----setup--------------------------------------------------------------------
library(neuromapr)

## ----simulate-data------------------------------------------------------------
set.seed(42)
n <- 100
coords <- matrix(rnorm(n * 3), ncol = 3)
distmat <- as.matrix(dist(coords))

map_a <- rnorm(n)
map_b <- 0.4 * map_a + rnorm(n, sd = 0.8)

## ----simple-compare-----------------------------------------------------------
result <- compare_maps(map_a, map_b, verbose = FALSE)
result

## ----null-comparison----------------------------------------------------------
result_null <- compare_maps(
  map_a, map_b,
  null_method = "burt2020",
  distmat = distmat,
  n_perm = 500L,
  seed = 1,
  verbose = FALSE
)
result_null

## ----null-distribution-plot, fig.cap = "Null correlations from 500 variogram-matching surrogates. The dashed red line marks the observed correlation."----
null_df <- data.frame(r = result_null$null_r)
obs_r <- result_null$r

ggplot2::ggplot(null_df, ggplot2::aes(x = r)) +
  ggplot2::geom_histogram(
    bins = 30, fill = "steelblue", alpha = 0.7
  ) +
  ggplot2::geom_vline(
    xintercept = obs_r,
    linetype = "dashed", color = "red", linewidth = 1
  ) +
  ggplot2::labs(
    x = "Null correlation (r)",
    y = "Count"
  ) +
  ggplot2::theme_minimal()

## ----precompute---------------------------------------------------------------
nulls <- generate_nulls(
  map_a,
  method = "moran",
  distmat = distmat,
  n_perm = 200L,
  seed = 42
)
nulls

## ----plot-nulls, fig.cap = "Built-in plot method for a null_distribution object."----
plot(nulls, parcel = 1L)

## ----reuse-nulls--------------------------------------------------------------
compare_maps(map_a, map_b, nulls = nulls, verbose = FALSE)

## ----permtest-----------------------------------------------------------------
mae <- function(a, b) mean(abs(a - b))

result_mae <- permtest_metric(
  map_a, rnorm(n),
  metric_func = mae,
  n_perm = 200L,
  seed = 1
)
result_mae$observed
result_mae$p_value

## ----permtest-spatial---------------------------------------------------------
result_spatial <- permtest_metric(
  map_a, rnorm(n),
  metric_func = mae,
  n_perm = 200L,
  seed = 1,
  null_method = "burt2020",
  distmat = distmat
)
result_spatial$p_value

describe("permtest_metric", {
  it("returns correct structure with random permutation", {
    set.seed(42)
    x <- rnorm(50)
    y <- x + rnorm(50, sd = 0.5)

    result <- permtest_metric(x, y, n_perm = 99L, seed = 1)
    expect_true(is.list(result))
    expect_true("observed" %in% names(result))
    expect_true("null_values" %in% names(result))
    expect_true("p_value" %in% names(result))
    expect_true("n_perm" %in% names(result))
    expect_equal(length(result$null_values), 99)
    expect_equal(result$n_perm, 99L)
  })

  it("observed matches direct correlation", {
    set.seed(42)
    x <- rnorm(30)
    y <- rnorm(30)

    result <- permtest_metric(x, y, n_perm = 10L, seed = 1)
    expect_equal(result$observed, stats::cor(x, y))
  })

  it("accepts custom metric function", {
    set.seed(42)
    x <- rnorm(30)
    y <- rnorm(30)
    mae <- function(a, b) mean(abs(a - b))

    result <- permtest_metric(x, y, metric_func = mae, n_perm = 10L, seed = 1)
    expect_equal(result$observed, mae(x, y))
  })

  it("p_value is between 0 and 1", {
    set.seed(42)
    x <- rnorm(30)
    y <- rnorm(30)

    result <- permtest_metric(x, y, n_perm = 99L, seed = 1)
    expect_true(result$p_value >= 0 && result$p_value <= 1)
  })

  it("errors for length mismatch", {
    expect_error(
      permtest_metric(1:5, 1:3, n_perm = 10L),
      "same length"
    )
  })

  it("errors for non-numeric input", {
    expect_error(
      permtest_metric(letters[1:5], 1:5, n_perm = 10L),
      "numeric"
    )
  })

  it("works with spatial null method", {
    set.seed(42)
    n <- 20
    data_x <- rnorm(n)
    data_y <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))

    result <- permtest_metric(
      data_x, data_y,
      n_perm = 5L, seed = 1,
      null_method = "burt2018",
      distmat = distmat
    )
    expect_equal(length(result$null_values), 5)
    expect_true(is.numeric(result$p_value))
  })
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)

## ----setup--------------------------------------------------------------------
library(neuromapr)

## ----vertices-to-parcels------------------------------------------------------
set.seed(42)
n_vertices <- 1000
vertex_data <- rnorm(n_vertices)

labels <- sample(
  c(0, 1:10), n_vertices,
  replace = TRUE, prob = c(0.1, rep(0.09, 10))
)

parcel_values <- vertices_to_parcels(vertex_data, labels)
parcel_values

## ----custom-summary-----------------------------------------------------------
parcel_sd <- vertices_to_parcels(
  vertex_data, labels, summary_func = sd
)
parcel_sd

## ----parcels-to-vertices------------------------------------------------------
vertex_filled <- parcels_to_vertices(parcel_values, labels)
head(vertex_filled, 20)

## ----custom-fill--------------------------------------------------------------
vertex_filled_zero <- parcels_to_vertices(
  parcel_values, labels, fill = 0
)
sum(vertex_filled_zero == 0)

## ----parcellate-vectors-------------------------------------------------------
parcellated <- parcellate(vertex_data, labels)
unparcellated <- unparcellate(parcellated, labels)

all.equal(
  parcels_to_vertices(parcel_values, labels),
  unparcellated
)

## ----centroids-setup----------------------------------------------------------
vertices <- matrix(rnorm(n_vertices * 3), ncol = 3)

## ----centroid-average---------------------------------------------------------
centroids_avg <- get_parcel_centroids(
  vertices, labels, method = "average"
)
head(centroids_avg)

## ----centroid-surface---------------------------------------------------------
centroids_surf <- get_parcel_centroids(
  vertices, labels, method = "surface"
)
head(centroids_surf)

## ----parcel-distmat-----------------------------------------------------------
parcel_distmat <- as.matrix(dist(centroids_avg))
dim(parcel_distmat)

## ----parcel-nulls-------------------------------------------------------------
parcel_map <- rnorm(nrow(centroids_avg))
nulls <- generate_nulls(
  parcel_map,
  method = "moran",
  distmat = parcel_distmat,
  n_perm = 100L,
  seed = 1
)
nulls

## ----parcel-spin--------------------------------------------------------------
n_lh <- 680
n_rh <- 680
sphere_coords <- list(
  lh = matrix(rnorm(n_lh * 3), ncol = 3),
  rh = matrix(rnorm(n_rh * 3), ncol = 3)
)
parcellation <- rep(1:68, each = 20)
parcel_data <- rnorm(68)

nulls_baum <- null_baum(
  parcel_data, sphere_coords, parcellation,
  n_perm = 50L, seed = 1
)
nulls_baum

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.5
)

## ----setup--------------------------------------------------------------------
library(neuromapr)

## ----shared-data--------------------------------------------------------------
set.seed(42)
n <- 80
coords_3d <- matrix(rnorm(n * 3), ncol = 3)
distmat <- as.matrix(dist(coords_3d))

map_x <- rnorm(n)

## ----burt2020-----------------------------------------------------------------
nulls_vario <- generate_nulls(
  map_x,
  method = "burt2020",
  distmat = distmat,
  n_perm = 100L,
  seed = 1
)
nulls_vario

## ----burt2018-----------------------------------------------------------------
nulls_sar <- generate_nulls(
  map_x,
  method = "burt2018",
  distmat = distmat,
  n_perm = 100L,
  seed = 1
)
nulls_sar

## ----sar-params---------------------------------------------------------------
nulls_sar$params$rho
nulls_sar$params$d0

## ----moran--------------------------------------------------------------------
nulls_moran <- generate_nulls(
  map_x,
  method = "moran",
  distmat = distmat,
  n_perm = 100L,
  seed = 1,
  procedure = "singleton"
)
nulls_moran

## ----spin-coords--------------------------------------------------------------
n_lh <- 40
n_rh <- 40
sphere_coords <- list(
  lh = matrix(rnorm(n_lh * 3), ncol = 3),
  rh = matrix(rnorm(n_rh * 3), ncol = 3)
)

vertex_data <- rnorm(n_lh + n_rh)

## ----alexander-bloch----------------------------------------------------------
nulls_ab <- null_alexander_bloch(
  vertex_data, sphere_coords,
  n_perm = 100L, seed = 1
)
nulls_ab

## ----spin-vasa----------------------------------------------------------------
nulls_vasa <- null_spin_vasa(
  vertex_data, sphere_coords,
  n_perm = 100L, seed = 1
)
nulls_vasa

## ----spin-hungarian, eval = rlang::is_installed("clue")-----------------------
nulls_hung <- null_spin_hungarian(
  vertex_data, sphere_coords,
  n_perm = 100L, seed = 1
)
nulls_hung

## ----parcel-setup-------------------------------------------------------------
n_parcel_lh <- 340
n_parcel_rh <- 340
parcel_coords <- list(
  lh = matrix(rnorm(n_parcel_lh * 3), ncol = 3),
  rh = matrix(rnorm(n_parcel_rh * 3), ncol = 3)
)
parcellation <- rep(1:68, each = 10)
parcel_data <- rnorm(68)

## ----baum---------------------------------------------------------------------
nulls_baum <- null_baum(
  parcel_data, parcel_coords, parcellation,
  n_perm = 50L, seed = 1
)
nulls_baum

## ----cornblath----------------------------------------------------------------
nulls_corn <- null_cornblath(
  parcel_data, parcel_coords, parcellation,
  n_perm = 50L, seed = 1
)
nulls_corn

## ----comparison-plot, fig.cap = "Surrogate distributions from variogram-matching (left) and Moran spectral randomization (right). The dashed red line is the original data distribution.", fig.width = 7, fig.height = 3.5----
df <- data.frame(
  value = c(
    as.vector(nulls_vario$nulls[, 1:50]),
    as.vector(nulls_moran$nulls[, 1:50])
  ),
  method = rep(
    c("burt2020 (variogram)", "moran (spectral)"),
    each = n * 50
  )
)

ggplot2::ggplot(df, ggplot2::aes(x = value)) +
  ggplot2::geom_density(fill = "steelblue", alpha = 0.5) +
  ggplot2::facet_wrap(~method) +
  ggplot2::geom_density(
    data = data.frame(value = map_x),
    color = "red", linetype = "dashed", linewidth = 0.8
  ) +
  ggplot2::labs(x = "Value", y = "Density") +
  ggplot2::theme_minimal()

## ----permtest-----------------------------------------------------------------
mae <- function(a, b) mean(abs(a - b))

result <- permtest_metric(
  map_x, rnorm(n),
  metric_func = mae,
  n_perm = 200L,
  seed = 1
)
result$observed
result$p_value

## ----permtest-spatial---------------------------------------------------------
result_spatial <- permtest_metric(
  map_x, rnorm(n),
  metric_func = mae,
  n_perm = 200L,
  seed = 1,
  null_method = "burt2020",
  distmat = distmat
)
result_spatial$p_value

## ----custom-nulls-------------------------------------------------------------
custom_nulls <- matrix(rnorm(n * 50), nrow = n, ncol = 50)
nd <- new_null_distribution(
  custom_nulls,
  method = "custom",
  observed = map_x
)
nd
summary(nd)

#' Aggregate vertex data into parcels
#'
#' Summarises vertex-level data by parcellation labels.
#'
#' @param data Numeric vector of vertex-level values.
#' @param labels Integer vector of parcel labels. `0` and `NA` are treated as
#'   medial wall and excluded.
#' @param summary_func Function to summarise each parcel (default: [mean]).
#'
#' @return Named numeric vector of parcel-level values.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' data <- c(1.0, 2.0, 3.0, 4.0)
#' labels <- c(1L, 1L, 2L, 2L)
#' vertices_to_parcels(data, labels)
#' @export
vertices_to_parcels <- function(data, labels, summary_func = mean) {
  if (length(data) != length(labels)) {
    cli::cli_abort(paste(
      "{.arg data} ({length(data)}) and {.arg labels}",
      "({length(labels)}) must have the same length."
    ))
  }
  valid <- !is.na(labels) & labels != 0
  result <- tapply(data[valid], labels[valid], summary_func)
  stats::setNames(as.numeric(result), names(result))
}

#' Map parcel data back to vertices
#'
#' Expands parcel-level values to a vertex-level vector using parcellation
#' labels.
#'
#' @param parcel_data Named numeric vector of parcel values
#'   (names match labels).
#' @param labels Integer vector of parcel labels. `0` and `NA` are medial wall.
#' @param fill Value for medial wall vertices (default: `NA_real_`).
#'
#' @return Numeric vector of `length(labels)`.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' parcel_data <- c("1" = 10, "2" = 20)
#' labels <- c(1L, 1L, 2L, 2L, 0L)
#' parcels_to_vertices(parcel_data, labels)
#' @export
parcels_to_vertices <- function(parcel_data, labels, fill = NA_real_) {
  out <- rep(fill, length(labels))
  valid <- !is.na(labels) & labels != 0
  label_names <- as.character(labels[valid])
  matched <- parcel_data[label_names]
  out[valid] <- unname(matched)
  out
}

#' Parcellate brain map data
#'
#' High-level wrapper that reads data and parcellation from file paths or
#' vectors, then aggregates vertices into parcels.
#'
#' @param data Numeric vector or file path to a GIFTI/NIfTI brain map.
#' @param parcellation Integer vector of labels or file path to a GIFTI label
#'   file.
#' @param summary_func Function to summarise each parcel (default: [mean]).
#'
#' @return Named numeric vector of parcel-level values.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' data <- c(1.0, 2.0, 3.0, 4.0)
#' labels <- c(1L, 1L, 2L, 2L)
#' parcellate(data, labels)
#' @export
parcellate <- function(data, parcellation, summary_func = mean) {
  if (is.character(data) && length(data) == 1) {
    data <- read_brain_map_values(data)
  }
  if (is.character(parcellation) && length(parcellation) == 1) {
    parcellation <- read_parcellation_labels(parcellation)
  }
  validate_parcellation(parcellation, length(data))
  vertices_to_parcels(data, parcellation, summary_func)
}

#' Unparcellate brain map data
#'
#' Inverse of [parcellate()]: maps parcel-level values back to vertices.
#'
#' @param parcel_data Named numeric vector of parcel values.
#' @param parcellation Integer vector of labels or file path to a GIFTI label
#'   file.
#' @param fill Value for medial wall vertices (default: `NA_real_`).
#'
#' @return Numeric vector of vertex-level values.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' parcel_data <- c("1" = 10, "2" = 20)
#' labels <- c(1L, 1L, 2L, 2L, 0L)
#' unparcellate(parcel_data, labels)
#' @export
unparcellate <- function(parcel_data, parcellation, fill = NA_real_) {
  if (is.character(parcellation) && length(parcellation) == 1) {
    parcellation <- read_parcellation_labels(parcellation)
  }
  parcels_to_vertices(parcel_data, parcellation, fill)
}

#' Compute parcel centroids
#'
#' Finds the centroid of each parcel using one of three methods.
#'
#' @param vertices Numeric matrix (n x 3) of vertex coordinates.
#' @param labels Integer vector of parcel labels. `0` and `NA` are medial wall.
#' @param method Centroid method: `"average"` (coordinate means), `"surface"`
#'   (vertex closest to the average centroid), or `"geodesic"` (vertex
#'   minimizing sum of geodesic distances within parcel).
#' @param faces Integer matrix (m x 3) of face indices. Required for
#'   `"geodesic"` method.
#'
#' @return Numeric matrix (n_parcels x 3) with rownames set to parcel labels.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' vertices <- matrix(rnorm(30), ncol = 3)
#' labels <- c(1L, 1L, 1L, 2L, 2L, 2L, 3L, 3L, 3L, 3L)
#' get_parcel_centroids(vertices, labels, method = "average")
#' @export
get_parcel_centroids <- function(vertices,
                                 labels,
                                 method = c("average", "surface", "geodesic"),
                                 faces = NULL) {
  method <- match.arg(method)
  if (!is.matrix(vertices) || ncol(vertices) != 3) {
    cli::cli_abort("{.arg vertices} must be a matrix with 3 columns.")
  }
  validate_parcellation(labels, nrow(vertices))

  if (method == "geodesic" && is.null(faces)) {
    cli::cli_abort("{.arg faces} is required for geodesic centroid method.")
  }

  valid <- !is.na(labels) & labels != 0
  ulabels <- sort(unique(labels[valid]))
  centroids <- matrix(0, nrow = length(ulabels), ncol = 3)
  rownames(centroids) <- ulabels

  for (i in seq_along(ulabels)) {
    idx <- which(labels == ulabels[i])
    parcel_verts <- vertices[idx, , drop = FALSE]

    if (method == "average") {
      centroids[i, ] <- colMeans(parcel_verts)
    } else if (method == "surface") {
      avg <- colMeans(parcel_verts)
      dists <- rowSums(sweep(parcel_verts, 2, avg)^2)
      centroids[i, ] <- parcel_verts[which.min(dists), ]
    } else {
      gdist <- get_surface_distance(vertices, faces, source_vertices = idx)
      gdist_within <- gdist[, idx, drop = FALSE]
      sum_dists <- rowSums(gdist_within)
      best <- which.min(sum_dists)
      centroids[i, ] <- parcel_verts[best, ]
    }
  }

  centroids
}

#' @noRd
#' @keywords internal
read_parcellation_labels <- function(path) {
  gii <- gifti::read_gifti(path)
  as.integer(gii$data[[1]])
}

describe("vertices_to_parcels", {
  it("aggregates by labels", {
    data <- c(1, 2, 3, 4, 5, 6)
    labels <- c(1, 1, 2, 2, 3, 3)
    result <- vertices_to_parcels(data, labels)
    expect_equal(unname(result), c(1.5, 3.5, 5.5))
  })

  it("excludes medial wall (0 and NA)", {
    data <- c(1, 2, 3, 4)
    labels <- c(0, 1, NA, 1)
    result <- vertices_to_parcels(data, labels)
    expect_equal(length(result), 1)
    expect_equal(unname(result), 3)
  })

  it("errors for length mismatch", {
    expect_error(
      vertices_to_parcels(1:3, 1:4),
      "same length"
    )
  })

  it("accepts custom summary function", {
    data <- c(1, 2, 3, 4)
    labels <- c(1, 1, 2, 2)
    result <- vertices_to_parcels(data, labels, summary_func = max)
    expect_equal(unname(result), c(2, 4))
  })
})

describe("parcels_to_vertices", {
  it("maps parcel values to vertex positions", {
    parcel_data <- c("1" = 10, "2" = 20)
    labels <- c(1, 1, 2, 2, 0)
    result <- parcels_to_vertices(parcel_data, labels)
    expect_equal(result, c(10, 10, 20, 20, NA_real_))
  })

  it("fills medial wall with custom value", {
    parcel_data <- c("1" = 10)
    labels <- c(1, 0)
    result <- parcels_to_vertices(parcel_data, labels, fill = -1)
    expect_equal(result, c(10, -1))
  })
})

describe("parcellate / unparcellate roundtrip", {
  it("roundtrips through parcellate and unparcellate", {
    data <- c(1, 2, 3, 4, 5, 6)
    labels <- c(1, 1, 2, 2, 3, 3)
    parcelled <- parcellate(data, labels)
    unparcelled <- unparcellate(parcelled, labels)
    expect_equal(unparcelled, c(1.5, 1.5, 3.5, 3.5, 5.5, 5.5))
  })
})

describe("get_parcel_centroids", {
  it("computes average centroids", {
    vertices <- matrix(c(
      0, 0, 0,
      2, 0, 0,
      0, 2, 0,
      2, 2, 0
    ), ncol = 3, byrow = TRUE)
    labels <- c(1, 1, 2, 2)

    centroids <- get_parcel_centroids(vertices, labels, method = "average")
    expect_equal(nrow(centroids), 2)
    expect_equal(centroids[1, ], c(1, 0, 0))
    expect_equal(centroids[2, ], c(1, 2, 0))
  })

  it("computes surface centroids", {
    vertices <- matrix(c(
      0, 0, 0,
      2, 0, 0,
      10, 0, 0
    ), ncol = 3, byrow = TRUE)
    labels <- c(1, 1, 1)

    centroids <- get_parcel_centroids(vertices, labels, method = "surface")
    expect_equal(nrow(centroids), 1)
    expect_equal(centroids[1, ], c(2, 0, 0))
  })

  it("errors for geodesic without faces", {
    vertices <- matrix(rnorm(9), ncol = 3)
    labels <- c(1, 1, 1)
    expect_error(
      get_parcel_centroids(vertices, labels, method = "geodesic"),
      "faces"
    )
  })
})

describe("validate_parcellation", {
  it("errors for wrong length", {
    expect_error(validate_parcellation(1:3, 5), "length")
  })

  it("errors for non-numeric", {
    expect_error(validate_parcellation(letters[1:3], 3), "integer")
  })
})

#' Create a null distribution object
#'
#' @param nulls Numeric matrix (n x n_perm) of surrogate values.
#' @param method Character string identifying the null model method.
#' @param observed Numeric vector of original data values.
#' @param params Named list of algorithm parameters.
#'
#' @return A `null_distribution` object.
#'
#' @name null_distribution
#' @examples
#' nulls <- matrix(rnorm(30), nrow = 3, ncol = 10)
#' nd <- new_null_distribution(nulls, "test", observed = c(1, 2, 3))
#' print(nd)
#' summary(nd)
#' @export
new_null_distribution <- function(nulls, method, observed, params = list()) {
  if (!is.matrix(nulls) || !is.numeric(nulls)) {
    cli::cli_abort("{.arg nulls} must be a numeric matrix.")
  }
  if (!is.character(method) || length(method) != 1L) {
    cli::cli_abort("{.arg method} must be a single character string.")
  }
  if (!is.numeric(observed)) {
    cli::cli_abort("{.arg observed} must be numeric.")
  }
  if (length(observed) != nrow(nulls)) {
    cli::cli_abort(paste(
      "{.arg observed} length ({length(observed)})",
      "must match rows in {.arg nulls} ({nrow(nulls)})."
    ))
  }
  structure(
    list(
      nulls = nulls,
      method = method,
      observed = observed,
      params = params,
      n_perm = ncol(nulls),
      n = nrow(nulls)
    ),
    class = c("null_distribution", "list")
  )
}

#' @noRd
#' @keywords internal
validate_null_distribution <- function(x) {
  if (!inherits(x, "null_distribution")) {
    cli::cli_abort("{.arg x} must be a {.cls null_distribution} object.")
  }
  if (!is.matrix(x$nulls)) {
    cli::cli_abort("{.field nulls} must be a matrix.")
  }
  if (length(x$observed) != nrow(x$nulls)) {
    cli::cli_abort(paste(
      "{.field observed} length ({length(x$observed)})",
      "must match rows in {.field nulls} ({nrow(x$nulls)})."
    ))
  }
  invisible(x)
}

#' @rdname null_distribution
#' @param x A `null_distribution` object.
#' @param ... Ignored.
#' @export
print.null_distribution <- function(x, ...) {
  cli::cli_h3("Null Distribution")
  cli::cli_ul(c(
    "Method: {x$method}",
    "Permutations: {x$n_perm}",
    "Observations: {x$n}"
  ))
  invisible(x)
}

#' @rdname null_distribution
#' @param object A `null_distribution` object.
#' @export
summary.null_distribution <- function(object, ...) {
  null_means <- rowMeans(object$nulls)
  null_sds <- apply(object$nulls, 1, stats::sd)
  list(
    method = object$method,
    n_perm = object$n_perm,
    n = object$n,
    null_mean = null_means,
    null_sd = null_sds,
    observed = object$observed
  )
}

#' @rdname null_distribution
#' @export
as.matrix.null_distribution <- function(x, ...) {
  x$nulls
}

#' @rdname null_distribution
#' @param parcel Integer index of the parcel to plot.
#' @export
plot.null_distribution <- function(x, parcel = 1L, ...) {
  df <- data.frame(value = x$nulls[parcel, ])
  obs <- x$observed[parcel]
  ggplot2::ggplot(df, ggplot2::aes(x = .data$value)) +
    ggplot2::geom_histogram(bins = 30) +
    ggplot2::geom_vline(
      xintercept = obs,
      linetype = "dashed",
      color = "firebrick"
    ) +
    ggplot2::labs(
      title = paste("Null distribution:", x$method),
      subtitle = paste("Parcel", parcel),
      x = "Value",
      y = "Count"
    )
}

describe("new_null_distribution", {
  it("errors for non-numeric observed", {
    nulls <- matrix(1:10, nrow = 5, ncol = 2)
    expect_error(
      new_null_distribution(nulls, "test", letters[1:5]),
      "numeric"
    )
  })

  it("creates a valid null_distribution object", {
    nulls <- matrix(rnorm(100), nrow = 10, ncol = 10)
    observed <- rnorm(10)
    nd <- new_null_distribution(nulls, "burt2020", observed, list(nh = 25))

    expect_s3_class(nd, "null_distribution")
    expect_equal(nd$method, "burt2020")
    expect_equal(nd$n_perm, 10)
    expect_equal(nd$n, 10)
    expect_identical(nd$nulls, nulls)
    expect_identical(nd$observed, observed)
    expect_equal(nd$params, list(nh = 25))
  })
})

describe("validate_null_distribution", {
  it("passes for valid objects", {
    nulls <- matrix(rnorm(50), nrow = 5, ncol = 10)
    nd <- new_null_distribution(nulls, "moran", rnorm(5))
    expect_invisible(validate_null_distribution(nd))
  })

  it("errors for non-null_distribution", {
    expect_error(validate_null_distribution(list()), "null_distribution")
  })

  it("errors when observed length mismatches nulls rows", {
    nulls <- matrix(rnorm(50), nrow = 5, ncol = 10)
    expect_error(
      new_null_distribution(nulls, "moran", rnorm(3)),
      "observed.*length"
    )
  })

  it("errors for non-numeric nulls", {
    expect_error(
      new_null_distribution("not a matrix", "test", 1),
      "numeric matrix"
    )
  })

  it("errors for non-character method", {
    expect_error(
      new_null_distribution(matrix(1:4, 2, 2), 42, 1:2),
      "character string"
    )
  })

  it("errors for corrupted nulls field", {
    nulls <- matrix(rnorm(50), nrow = 5, ncol = 10)
    nd <- new_null_distribution(nulls, "moran", rnorm(5))
    nd$nulls <- "broken"
    expect_error(validate_null_distribution(nd), "matrix")
  })

  it("errors for corrupted observed length", {
    nulls <- matrix(rnorm(50), nrow = 5, ncol = 10)
    nd <- new_null_distribution(nulls, "moran", rnorm(5))
    nd$observed <- rnorm(3)
    expect_error(validate_null_distribution(nd), "observed.*length")
  })
})

describe("print.null_distribution", {
  it("prints method, permutations, and observations", {
    nulls <- matrix(rnorm(50), nrow = 5, ncol = 10)
    nd <- new_null_distribution(nulls, "moran", rnorm(5))
    expect_snapshot(res <- print(nd))
    expect_identical(res, nd)
  })
})

describe("summary.null_distribution", {
  it("returns expected summary fields", {
    nulls <- matrix(rnorm(50), nrow = 5, ncol = 10)
    nd <- new_null_distribution(nulls, "burt2020", rnorm(5))
    s <- summary(nd)

    expect_equal(s$method, "burt2020")
    expect_equal(s$n_perm, 10)
    expect_equal(s$n, 5)
    expect_length(s$null_mean, 5)
    expect_length(s$null_sd, 5)
  })
})

describe("as.matrix.null_distribution", {
  it("returns the nulls matrix", {
    nulls <- matrix(1:20, nrow = 4, ncol = 5)
    nd <- new_null_distribution(nulls, "moran", 1:4)
    expect_identical(as.matrix(nd), nulls)
  })
})

describe("plot.null_distribution", {
  it("returns a ggplot object when ggplot2 available", {
    skip_if_not_installed("ggplot2")
    nulls <- matrix(rnorm(100), nrow = 10, ncol = 10)
    nd <- new_null_distribution(nulls, "burt2020", rnorm(10))
    p <- plot(nd, parcel = 1)
    expect_s3_class(p, "gg")
  })
})

fixture_dir <- function() {
  testthat::test_path("fixtures", "python")
}

skip_if_no_python_fixtures <- function() {
  dir <- fixture_dir()
  if (!dir.exists(dir) || length(list.files(dir, pattern = "\\.json$")) == 0) {
    testthat::skip(paste(
      "Python reference fixtures not found",
      "(run tools/generate-reference-data.py)"
    ))
  }
}

load_python_fixture <- function(name) {
  path <- file.path(fixture_dir(), paste0(name, ".json"))
  if (!file.exists(path)) {
    testthat::skip(paste0("Fixture not found: ", name, ".json"))
  }
  jsonlite::fromJSON(path, simplifyVector = TRUE)
}

describe("cross-validation with Python reference", {
  skip_if_no_python_fixtures()

  inputs <- load_python_fixture("shared_inputs")
  n <- inputs$n
  distmat <- as.matrix(inputs$distmat)
  data <- inputs$data
  data_y <- inputs$data_y

  describe("weight matrix (inverse distance)", {
    ref <- load_python_fixture("weight_matrix_inverse_distance")

    it("matches Python output", {
      w <- compute_weight_matrix(distmat, kernel = "inverse_distance")
      expect_equal(w, as.matrix(ref$w), tolerance = 1e-10)
    })
  })

  describe("weight matrix (exponential)", {
    ref <- load_python_fixture("weight_matrix_exponential")

    it("matches Python output", {
      w <- compute_weight_matrix(distmat, kernel = "exponential")
      expect_equal(w, as.matrix(ref$w), tolerance = 1e-10)
    })
  })

  describe("SAR weight matrix", {
    ref <- load_python_fixture("sar_weight_matrix")

    it("matches Python output", {
      w <- build_sar_weights(distmat, ref$d0)
      expect_equal(w, as.matrix(ref$w), tolerance = 1e-10)
    })

    it("has zero diagonal", {
      w <- build_sar_weights(distmat, ref$d0)
      expect_equal(diag(w), rep(0, n))
    })

    it("has unit row sums", {
      w <- build_sar_weights(distmat, ref$d0)
      expect_equal(rowSums(w), rep(1, n), tolerance = 1e-10)
    })
  })

  describe("MEM eigendecomposition", {
    ref <- load_python_fixture("mem")
    w <- compute_weight_matrix(distmat, kernel = "inverse_distance")
    mem <- compute_mem(w, n)

    it("eigenvalues match", {
      r_vals <- sort(mem$values, decreasing = TRUE)
      py_vals <- sort(ref$eigenvalues, decreasing = TRUE)
      min_len <- min(length(r_vals), length(py_vals))
      expect_equal(
        r_vals[seq_len(min_len)],
        py_vals[seq_len(min_len)],
        tolerance = 1e-8
      )
    })

    it("eigenvectors match up to sign", {
      r_vecs <- mem$vectors
      py_vecs <- as.matrix(ref$eigenvectors)
      min_cols <- min(ncol(r_vecs), ncol(py_vecs))

      r_order <- order(mem$values, decreasing = TRUE)
      py_order <- order(ref$eigenvalues, decreasing = TRUE)
      r_vecs <- r_vecs[, r_order[seq_len(min_cols)], drop = FALSE]
      py_vecs <- py_vecs[, py_order[seq_len(min_cols)], drop = FALSE]

      for (j in seq_len(min_cols)) {
        same_sign <- sum((r_vecs[, j] - py_vecs[, j])^2)
        flip_sign <- sum((r_vecs[, j] + py_vecs[, j])^2)
        expect_lt(min(same_sign, flip_sign), 1e-8)
      }
    })
  })

  describe("Moran double-centered matrix", {
    ref <- load_python_fixture("moran_double_centered")

    it("matches Python output", {
      w <- compute_weight_matrix(distmat, kernel = "inverse_distance")
      sym_w <- (w + t(w)) / 2
      row_means <- rowMeans(sym_w)
      grand_mean <- mean(row_means)
      dbl_r <- sym_w - outer(row_means, row_means, "+") + grand_mean
      expect_equal(dbl_r, as.matrix(ref$dbl), tolerance = 1e-10)
    })
  })

  describe("variogram computation", {
    ref <- load_python_fixture("variogram")

    it("matches Python output with fixed indices", {
      idx_r <- sort(ref$idx) + 1L
      vg <- compute_variogram(
        data,
        distmat,
        nh = ref$nh,
        pv = ref$pv,
        ns = ref$ns,
        idx = idx_r
      )
      expect_equal(vg$gamma, ref$gamma, tolerance = 1e-10)
      expect_equal(vg$distance, ref$bin_centers, tolerance = 1e-10)
    })
  })

  describe("ZYZ Euler rotation matrix", {
    ref <- load_python_fixture("rotation_matrix")

    it("matches Python output for fixed angles", {
      alpha <- ref$alpha
      beta <- ref$beta
      gamma <- ref$gamma
      ca <- cos(alpha)
      sa <- sin(alpha)
      cb <- cos(beta)
      sb <- sin(beta)
      cg <- cos(gamma)
      sg <- sin(gamma)

      rot <- matrix(
        c(
          # nolint: object_name
          cg * cb * ca - sg * sa,
          sg * cb * ca + cg * sa,
          -sb * ca,
          -cg * cb * sa - sg * ca,
          -sg * cb * sa + cg * ca,
          sb * sa,
          cg * sb,
          sg * sb,
          cb
        ),
        nrow = 3,
        ncol = 3
      )

      expect_equal(rot, as.matrix(ref$R), tolerance = 1e-12)
    })
  })

  describe("cost matrix", {
    ref <- load_python_fixture("cost_matrix")

    it("matches Python output", {
      original <- as.matrix(ref$original)
      rotated <- as.matrix(ref$rotated)
      cost <- compute_cost_matrix(original, rotated)
      expect_equal(cost, as.matrix(ref$cost), tolerance = 1e-10)
    })
  })

  describe("rank matching", {
    ref <- load_python_fixture("rank_match")

    it("matches Python output", {
      matched <- rank_match(ref$surrogate, ref$target)
      expect_equal(matched, ref$matched)
    })
  })

  describe("correlations", {
    ref <- load_python_fixture("correlation_reference")

    it("Pearson r matches Python", {
      r_val <- cor(data, data_y, method = "pearson")
      expect_equal(r_val, ref$pearson_r, tolerance = 1e-6)
    })

    it("Spearman r matches Python", {
      r_val <- cor(data, data_y, method = "spearman")
      expect_equal(r_val, ref$spearman_r, tolerance = 1e-6)
    })

    it("compare_maps r matches Python", {
      result <- compare_maps(data, data_y, verbose = FALSE)
      expect_equal(result$r, ref$pearson_r, tolerance = 1e-6)
    })
  })

  describe("null distribution statistics", {
    ref <- load_python_fixture("null_distribution_stats")

    it("SAR null correlations have similar distribution", {
      sar_r <- null_burt2018(data, distmat, n_perm = ref$n_perm, seed = 1L)
      sar_surr <- as.matrix(sar_r)
      null_r <- apply(sar_surr, 2, function(s) cor(s, data_y))
      expect_equal(mean(null_r), ref$sar_null_r_mean, tolerance = 0.15)
      expect_equal(sd(null_r), ref$sar_null_r_std, tolerance = 0.15)
    })

    it("SAR surrogates preserve rank distribution", {
      expect_true(ref$sar_rank_preserved)
    })

    it("Moran null correlations have similar distribution", {
      moran_r <- null_moran(
        data, distmat,
        n_perm = ref$n_perm, seed = 1L, procedure = "singleton"
      )
      moran_surr <- as.matrix(moran_r)
      null_r <- apply(moran_surr, 2, function(s) cor(s, data_y))
      expect_equal(mean(null_r), ref$moran_null_r_mean, tolerance = 0.15)
      expect_equal(sd(null_r), ref$moran_null_r_std, tolerance = 0.15)
    })
  })

  describe("burt2020 surrogate properties", {
    it("preserves rank distribution", {
      skip_if(
        !file.exists(file.path(fixture_dir(), "burt2020_stats.json")),
        paste(
          "burt2020 stats fixture not available",
          "(brainsmash not installed in Python)"
        )
      )
      ref <- load_python_fixture("burt2020_stats")
      expect_true(ref$rank_preserved)
    })
  })
})

#' List available neuromaps annotations
#'
#' Query the neuromaps registry to see which brain map annotations are
#' available for download. Data is fetched from the
#' [neuromaps](https://github.com/netneurolab/neuromaps) project's GitHub
#' repository and cached for the session.
#'
#' @template neuromaps-params
#' @param tags Character vector of tags. All must match (AND logic).
#' @param format Filter by format (`"surface"` or `"volume"`).
#'
#' @details
#' All string filter parameters (`source`, `desc`, `space`, `density`,
#' `resolution`, `hemisphere`, `format`) are treated as **R regular
#' expressions** and matched with [grepl()]. For example,
#' `source = "^beliveau$"` matches exactly, while `source = "bel"` matches
#' any source containing `"bel"`. Set `fixed = TRUE` for literal string
#' matching. The `tags` parameter always uses exact matching (AND logic).
#'
#' When used in [fetch_neuromaps_annotation()], `source`, `desc`, and `space`
#' are exact matches.
#'
#' @param refresh Logical. If `TRUE`, forces a fresh download of the
#'   registry data, ignoring any session cache.
#' @param fixed Logical. If `TRUE`, filter strings are matched literally
#'   rather than as regular expressions.
#'
#' @return A tibble of available annotations with columns: source, desc,
#'   space, den, res, hemi, format, fname, full_desc, tags, N, age.
#' @export
#' @examplesIf interactive()
#' neuromaps_available()
#' neuromaps_available(source = "beliveau")
#' neuromaps_available(tags = "pet")
neuromaps_available <- function(
  source = NULL,
  desc = NULL,
  space = NULL,
  density = NULL,
  resolution = NULL,
  hemisphere = NULL,
  tags = NULL,
  format = NULL,
  refresh = FALSE,
  fixed = FALSE
) {
  registry <- build_neuromaps_registry(refresh = refresh)
  filter_neuromaps_registry(
    registry,
    source = source,
    desc = desc,
    space = space,
    density = density,
    resolution = resolution,
    hemisphere = hemisphere,
    tags = tags,
    format = format,
    fixed = fixed
  )
}


#' Download a neuromaps annotation
#'
#' Download brain map annotation files from the neuromaps OSF repository.
#' Surface annotations return two files (left and right hemispheres),
#' volume annotations return one.
#'
#' @template neuromaps-params
#' @param data_dir Directory for cached downloads. Defaults to
#'   `neuromaps_cache_dir()`.
#' @param overwrite Re-download even if cached file exists.
#' @param verbose Print progress messages.
#'
#' @return Character vector of downloaded file path(s).
#' @export
#' @examplesIf interactive()
#' fetch_neuromaps_annotation("abagen", "genepc1", "fsaverage", density = "10k")
fetch_neuromaps_annotation <- function(
  source,
  desc,
  space,
  density = NULL,
  resolution = NULL,
  hemisphere = NULL,
  data_dir = neuromaps_cache_dir(),
  overwrite = FALSE,
  verbose = TRUE
) {
  if (!is.null(density) && !is.null(resolution)) {
    cli::cli_abort(
      "{.arg density} and {.arg resolution} are mutually exclusive."
    )
  }

  entries <- resolve_neuromaps_entries(
    source = source,
    desc = desc,
    space = space,
    density = density,
    resolution = resolution,
    hemisphere = hemisphere
  )

  paths <- character(nrow(entries))
  for (i in seq_len(nrow(entries))) {
    entry <- entries[i, ]
    destfile <- file.path(data_dir, "annotations", entry$rel_path, entry$fname)
    url <- osf_download_url(
      entry$osf_project, entry$osf_file_id
    )
    paths[i] <- download_neuromaps_file(
      url = url,
      destfile = destfile,
      checksum = entry$checksum,
      overwrite = overwrite,
      verbose = verbose
    )
  }

  paths
}

describe("neuromaps registry (integration)", {
  it("fetches osf.json from GitHub", {
    skip_if_not_installed("vcr")

    the$osf_json <- NULL

    vcr::use_cassette("neuromaps_osf_json", {
      result <- fetch_neuromaps_osf_json()
    })

    expect_type(result, "list")
    expect_true("annotations" %in% names(result))
    expect_true(length(result$annotations) > 0)

    the$osf_json <- NULL
  })

  it("fetches meta.json from GitHub", {
    skip_if_not_installed("vcr")

    the$meta_json <- NULL

    vcr::use_cassette("neuromaps_meta_json", {
      result <- fetch_neuromaps_meta_json()
    })

    expect_type(result, "list")
    expect_true("annotations" %in% names(result))
    expect_true(length(result$annotations) > 0)

    the$meta_json <- NULL
  })

  it("builds and queries the full registry", {
    skip_if_not_installed("vcr")

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL

    vcr::use_cassette("neuromaps_registry_full", {
      result <- neuromaps_available()
    })

    expect_s3_class(result, "tbl_df")
    expect_true(nrow(result) > 10)
    expect_true(all(
      c("source", "desc", "space", "hemi", "format") %in%
        names(result)
    ))

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL
  })

  it("filters the live registry by source", {
    skip_if_not_installed("vcr")

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL

    vcr::use_cassette("neuromaps_registry_filter_source", {
      result <- neuromaps_available(source = "abagen")
    })

    expect_true(nrow(result) > 0)
    expect_true(all(grepl("abagen", result$source)))

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL
  })

  it("filters by format", {
    skip_if_not_installed("vcr")

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL

    vcr::use_cassette("neuromaps_registry_filter_format", {
      surfaces <- neuromaps_available(format = "surface")
      volumes <- neuromaps_available(format = "volume")
    })

    expect_true(nrow(surfaces) > 0)
    expect_true(nrow(volumes) > 0)
    expect_true(all(surfaces$format == "surface"))
    expect_true(all(volumes$format == "volume"))

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL
  })
})


describe("neuromaps download (integration)", {
  it("downloads a real annotation from OSF", {
    skip_on_cran()
    skip_if_offline()

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL

    tmp_dir <- withr::local_tempdir()

    paths <- fetch_neuromaps_annotation(
      source = "abagen",
      desc = "genepc1",
      space = "fsaverage",
      density = "10k",
      hemisphere = "L",
      data_dir = tmp_dir,
      verbose = FALSE
    )

    expect_length(paths, 1)
    expect_true(file.exists(paths))
    expect_true(file.size(paths) > 0)
    expect_true(grepl("hemi-L", paths))
    expect_true(grepl("\\.func\\.gii$", paths))

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL
  })

  it("uses cache on second download", {
    skip_on_cran()
    skip_if_offline()

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL

    tmp_dir <- withr::local_tempdir()

    paths1 <- fetch_neuromaps_annotation(
      source = "abagen",
      desc = "genepc1",
      space = "fsaverage",
      density = "10k",
      hemisphere = "L",
      data_dir = tmp_dir,
      verbose = FALSE
    )

    mtime1 <- file.mtime(paths1)

    paths2 <- fetch_neuromaps_annotation(
      source = "abagen",
      desc = "genepc1",
      space = "fsaverage",
      density = "10k",
      hemisphere = "L",
      data_dir = tmp_dir,
      verbose = FALSE
    )

    expect_equal(paths1, paths2)
    expect_equal(file.mtime(paths2), mtime1)

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL
  })

  it("reads and compares downloaded files", {
    skip_on_cran()
    skip_if_offline()

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL

    tmp_dir <- withr::local_tempdir()

    paths <- fetch_neuromaps_annotation(
      source = "abagen",
      desc = "genepc1",
      space = "fsaverage",
      density = "10k",
      data_dir = tmp_dir,
      verbose = FALSE
    )

    result <- compare_maps(paths[1], paths[1])
    expect_s3_class(result, "neuromaps_comparison")
    expect_equal(result$r, 1)
    expect_true(result$n > 0)

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL
  })
})

describe("neuromaps_available", {
  it("returns full registry when no filters", {
    local_mocked_bindings(
      build_neuromaps_registry = function(refresh = FALSE) {
        tibble::tibble(
          source = c("a", "b"), desc = c("x", "y"),
          space = "fs", den = "10k", res = NA,
          hemi = "L", format = "surface",
          fname = c("f1", "f2"),
          full_desc = c("A map", "B map"),
          tags = list("pet", "mri"),
          N = c(10, 20), age = c("30", "40")
        )
      }
    )

    result <- neuromaps_available()
    expect_s3_class(result, "tbl_df")
    expect_equal(nrow(result), 2)
  })

  it("applies filters", {
    local_mocked_bindings(
      build_neuromaps_registry = function(refresh = FALSE) {
        tibble::tibble(
          source = c("abagen", "beliveau"),
          desc = c("x", "y"),
          space = "fs", den = "10k", res = NA,
          hemi = "L", format = "surface",
          fname = c("f1", "f2"),
          full_desc = c("A", "B"),
          tags = list("genetics", "pet"),
          N = c(6, 18), age = c("30", "40")
        )
      }
    )

    result <- neuromaps_available(source = "beliveau")
    expect_equal(nrow(result), 1)
    expect_equal(result$source, "beliveau")
  })
})


describe("fetch_neuromaps_annotation", {
  it("errors when both density and resolution provided", {
    expect_error(
      fetch_neuromaps_annotation(
        "a", "b", "c",
        density = "10k", resolution = "1mm"
      ),
      "mutually exclusive"
    )
  })

  it("downloads resolved entries", {
    downloaded <- list()
    tmp_dir <- withr::local_tempdir()

    local_mocked_bindings(
      resolve_neuromaps_entries = function(...) {
        tibble::tibble(
          source = "test", desc = "m", space = "fs",
          den = "10k", res = NA, hemi = c("L", "R"),
          fname = c("left.gii", "right.gii"),
          rel_path = c("test/m/fs/", "test/m/fs/"),
          checksum = c("aaa", "bbb"),
          osf_project = c("4mw3a", "4mw3a"),
          osf_file_id = c("f1", "f2")
        )
      },
      download_neuromaps_file = function(url, destfile, ...) {
        downloaded[[length(downloaded) + 1]] <<- destfile
        destfile
      }
    )

    result <- fetch_neuromaps_annotation(
      "test", "m", "fs",
      density = "10k", data_dir = tmp_dir
    )
    expect_length(result, 2)
    expect_length(downloaded, 2)
  })
})

#' Build an igraph from a triangular surface mesh
#'
#' Extracts unique edges from triangular faces, computes Euclidean edge weights,
#' and returns an igraph graph object suitable for geodesic distance
#' computation.
#'
#' @param vertices Numeric matrix (n x 3) of vertex coordinates.
#' @param faces Integer matrix (m x 3) of face indices (1-indexed).
#'
#' @return An `igraph` graph object with weighted edges.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' vertices <- matrix(
#'   c(0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1),
#'   nrow = 4, byrow = TRUE
#' )
#' faces <- matrix(c(1L, 2L, 3L, 2L, 3L, 4L), nrow = 2, byrow = TRUE)
#' g <- make_surf_graph(vertices, faces)
#' @export
make_surf_graph <- function(vertices, faces) {
  if (!is.matrix(vertices) || ncol(vertices) != 3) {
    cli::cli_abort("{.arg vertices} must be a matrix with 3 columns.")
  }
  if (!is.matrix(faces) || ncol(faces) != 3) {
    cli::cli_abort("{.arg faces} must be a matrix with 3 columns.")
  }
  edges <- rbind(
    faces[, c(1, 2)],
    faces[, c(2, 3)],
    faces[, c(1, 3)]
  )
  edges <- t(apply(edges, 1, sort))
  edges <- unique(edges)

  weights <- sqrt(rowSums((vertices[edges[, 1], ] - vertices[edges[, 2], ])^2))

  igraph::graph_from_edgelist(edges, directed = FALSE) |>
    igraph::set_edge_attr("weight", value = weights)
}

#' Compute geodesic distances on a surface mesh
#'
#' Builds a graph from a triangular mesh and computes shortest-path (Dijkstra)
#' distances between vertices.
#'
#' @param vertices Numeric matrix (n x 3) of vertex coordinates.
#' @param faces Integer matrix (m x 3) of face indices (1-indexed).
#' @param source_vertices Optional integer vector of source vertex indices.
#'   If `NULL`, computes the full n x n distance matrix.
#'
#' @return Numeric distance matrix. If `source_vertices` is provided, returns
#'   a `length(source_vertices) x n` matrix; otherwise an `n x n` matrix.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' vertices <- matrix(
#'   c(0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1),
#'   nrow = 4, byrow = TRUE
#' )
#' faces <- matrix(c(1L, 2L, 3L, 2L, 3L, 4L), nrow = 2, byrow = TRUE)
#' get_surface_distance(vertices, faces)
#' @export
get_surface_distance <- function(vertices, faces, source_vertices = NULL) {
  g <- make_surf_graph(vertices, faces)
  if (is.null(source_vertices)) {
    igraph::distances(g, algorithm = "dijkstra")
  } else {
    igraph::distances(g, v = source_vertices, algorithm = "dijkstra")
  }
}

describe("make_surf_graph", {
  it("builds a weighted igraph from a simple mesh", {
    skip_if_not_installed("igraph")
    vertices <- matrix(c(
      0, 0, 0,
      1, 0, 0,
      0, 1, 0,
      1, 1, 0
    ), ncol = 3, byrow = TRUE)
    faces <- matrix(c(1, 2, 3, 2, 3, 4), ncol = 3, byrow = TRUE)

    g <- make_surf_graph(vertices, faces)
    expect_s3_class(g, "igraph")
    expect_true(igraph::is_weighted(g))
    expect_equal(igraph::vcount(g), 4)
  })

  it("errors for non-3-column vertices", {
    expect_error(
      make_surf_graph(matrix(1:4, ncol = 2), matrix(1:3, ncol = 3)),
      "3 columns"
    )
  })

  it("errors for non-3-column faces", {
    expect_error(
      make_surf_graph(matrix(1:6, ncol = 3), matrix(1:4, ncol = 2)),
      "3 columns"
    )
  })
})

describe("get_surface_distance", {
  it("returns full distance matrix by default", {
    skip_if_not_installed("igraph")
    vertices <- matrix(c(
      0, 0, 0,
      1, 0, 0,
      0, 1, 0
    ), ncol = 3, byrow = TRUE)
    faces <- matrix(c(1, 2, 3), ncol = 3)

    dm <- get_surface_distance(vertices, faces)
    expect_equal(nrow(dm), 3)
    expect_equal(ncol(dm), 3)
    expect_equal(dm[1, 1], 0)
    expect_true(dm[1, 2] > 0)
  })

  it("returns subset when source_vertices given", {
    skip_if_not_installed("igraph")
    vertices <- matrix(c(
      0, 0, 0,
      1, 0, 0,
      0, 1, 0
    ), ncol = 3, byrow = TRUE)
    faces <- matrix(c(1, 2, 3), ncol = 3)

    dm <- get_surface_distance(vertices, faces, source_vertices = 1L)
    expect_equal(nrow(dm), 1)
    expect_equal(ncol(dm), 3)
  })
})

describe("compute_distance_matrix geodesic", {
  it("dispatches to geodesic when method specified", {
    skip_if_not_installed("igraph")
    vertices <- matrix(c(
      0, 0, 0,
      1, 0, 0,
      0, 1, 0
    ), ncol = 3, byrow = TRUE)
    faces <- matrix(c(1, 2, 3), ncol = 3)

    dm <- compute_distance_matrix(vertices, method = "geodesic", faces = faces)
    expect_equal(nrow(dm), 3)
  })

  it("errors without faces for geodesic", {
    vertices <- matrix(rnorm(9), ncol = 3)
    expect_error(
      compute_distance_matrix(vertices, method = "geodesic"),
      "faces"
    )
  })
})

#' Alexander-Bloch spin test null model
#'
#' Original vertex-level spin test: rotates coordinates and assigns each rotated
#' vertex the value of its nearest original vertex (no optimal matching).
#'
#' @template null-params
#' @param coords List with `$lh` and `$rh` matrices of spherical coordinates
#'   (n x 3 each).
#' @param rotation Rotation generation method: `"euler"` (ZYZ Euler angles,
#'   default, matches neuromaps Python) or `"rodrigues"` (Rodrigues axis-angle
#'   formula).
#'
#' @return A [null_distribution] object.
#'
#' @references
#' Alexander-Bloch AF et al. (2018) NeuroImage 175:111-120.
#' doi:10.1016/j.neuroimage.2018.04.023
#'
#' @examples
#' coords <- list(lh = matrix(rnorm(30), 10, 3), rh = matrix(rnorm(30), 10, 3))
#' data <- rnorm(20)
#' nd <- null_alexander_bloch(data, coords, n_perm = 10L, seed = 1L)
#' @export
null_alexander_bloch <- function(data, coords, n_perm = 1000L, seed = NULL,
                                 rotation = c("euler", "rodrigues")) {
  validate_data(data)
  validate_coords(coords)
  rotation <- match.arg(rotation)
  n_lh <- nrow(coords$lh)
  n_rh <- nrow(coords$rh)
  n <- n_lh + n_rh
  if (length(data) != n) {
    cli::cli_abort(
      "Length of {.arg data} ({length(data)}) must match total vertices ({n})."
    )
  }

  rotated <- rotate_coords(
    coords$lh, coords$rh, n_perm, seed,
    rotation = rotation
  )
  nulls <- matrix(0, nrow = n, ncol = n_perm)

  msg <- "Generating alexander_bloch nulls"
  for (i in cli::cli_progress_along(seq_len(n_perm), msg)) {
    cost_lh <- compute_cost_matrix(coords$lh, rotated$lh[, , i])
    cost_rh <- compute_cost_matrix(coords$rh, rotated$rh[, , i])
    assign_lh <- apply(cost_lh, 1, which.min)
    assign_rh <- apply(cost_rh, 1, which.min)
    nulls[seq_len(n_lh), i] <- data[assign_lh]
    nulls[n_lh + seq_len(n_rh), i] <- data[n_lh + assign_rh]
  }

  new_null_distribution(nulls, "alexander_bloch", data, list(n_perm = n_perm))
}

describe("null_alexander_bloch", {
  it("returns null_distribution with correct dims", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    data <- rnorm(n_lh + n_rh)

    result <- null_alexander_bloch(data, coords, n_perm = 5L, seed = 1)
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "alexander_bloch")
    expect_equal(nrow(result$nulls), 20)
    expect_equal(ncol(result$nulls), 5)
  })

  it("errors for data length mismatch", {
    coords <- list(
      lh = matrix(rnorm(15), ncol = 3),
      rh = matrix(rnorm(15), ncol = 3)
    )
    expect_error(
      null_alexander_bloch(1:3, coords, n_perm = 2L),
      "must match"
    )
  })

  it("produces different values per permutation", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    data <- rnorm(n_lh + n_rh)

    result <- null_alexander_bloch(data, coords, n_perm = 3L, seed = 1)
    expect_false(all(result$nulls[, 1] == result$nulls[, 2]))
  })
})

#' Spin-test null models for brain maps
#'
#' Generate spatially-constrained null distributions using spin-based
#' permutation of spherical coordinates.
#'
#' @template null-params
#' @param coords List with `$lh` and `$rh` matrices of spherical coordinates
#'   (n x 3 each).
#' @param rotation Rotation generation method: `"euler"` (ZYZ Euler angles,
#'   default, matches neuromaps Python) or `"rodrigues"` (Rodrigues axis-angle
#'   formula).
#'
#' @return A [null_distribution] object.
#'
#' @references
#' Alexander-Bloch AF et al. (2018) NeuroImage 175:111-120.
#' doi:10.1016/j.neuroimage.2018.04.023
#'
#' Vasa F et al. (2018) Cerebral Cortex 28:3293-3303.
#' doi:10.1093/cercor/bhx195
#'
#' Markello RD, Misic B (2021) NeuroImage 236:118052.
#' doi:10.1016/j.neuroimage.2021.118052
#'
#' @examples
#' coords <- list(lh = matrix(rnorm(30), 10, 3), rh = matrix(rnorm(30), 10, 3))
#' data <- rnorm(20)
#' nd <- null_spin_vasa(data, coords, n_perm = 10L, seed = 1L)
#' @export
null_spin_vasa <- function(data, coords, n_perm = 1000L,
                           seed = NULL,
                           rotation = c("euler", "rodrigues")) {
  validate_data(data)
  validate_coords(coords)
  rotation <- match.arg(rotation)
  n_lh <- nrow(coords$lh)
  n_rh <- nrow(coords$rh)
  n <- n_lh + n_rh
  if (length(data) != n) {
    cli::cli_abort(
      "Length of {.arg data} ({length(data)}) must match total parcels ({n})."
    )
  }

  rotated <- rotate_coords(
    coords$lh, coords$rh, n_perm, seed,
    rotation = rotation
  )
  nulls <- matrix(0, nrow = n, ncol = n_perm)

  msg <- "Generating spin_vasa nulls"
  for (i in cli::cli_progress_along(seq_len(n_perm), msg)) {
    cost_lh <- compute_cost_matrix(coords$lh, rotated$lh[, , i])
    cost_rh <- compute_cost_matrix(coords$rh, rotated$rh[, , i])
    assign_lh <- assign_parcels_vasa(cost_lh)
    assign_rh <- assign_parcels_vasa(cost_rh)
    nulls[seq_len(n_lh), i] <- data[assign_lh]
    nulls[n_lh + seq_len(n_rh), i] <- data[n_lh + assign_rh]
  }

  new_null_distribution(
    nulls, "spin_vasa", data, list(n_perm = n_perm)
  )
}

#' @rdname null_spin_vasa
#' @export
null_spin_hungarian <- function(
    data, coords, n_perm = 1000L, seed = NULL,
    rotation = c("euler", "rodrigues")) {
  rlang::check_installed(
    "clue",
    reason = "for Hungarian assignment in spin tests"
  )
  validate_data(data)
  validate_coords(coords)
  rotation <- match.arg(rotation)
  n_lh <- nrow(coords$lh)
  n_rh <- nrow(coords$rh)
  n <- n_lh + n_rh
  if (length(data) != n) {
    cli::cli_abort(
      "Length of {.arg data} ({length(data)}) must match total parcels ({n})."
    )
  }

  rotated <- rotate_coords(
    coords$lh, coords$rh, n_perm, seed,
    rotation = rotation
  )
  nulls <- matrix(0, nrow = n, ncol = n_perm)

  msg <- "Generating spin_hungarian nulls"
  for (i in cli::cli_progress_along(seq_len(n_perm), msg)) {
    cost_lh <- compute_cost_matrix(coords$lh, rotated$lh[, , i])
    cost_rh <- compute_cost_matrix(coords$rh, rotated$rh[, , i])
    assign_lh <- assign_parcels_hungarian(cost_lh)
    assign_rh <- assign_parcels_hungarian(cost_rh)
    nulls[seq_len(n_lh), i] <- data[assign_lh]
    nulls[n_lh + seq_len(n_rh), i] <- data[n_lh + assign_rh]
  }

  new_null_distribution(
    nulls, "spin_hungarian", data, list(n_perm = n_perm)
  )
}

#' @noRd
#' @keywords internal
random_rotation_matrix <- function(
    method = c("euler", "rodrigues")) {
  method <- match.arg(method)
  switch(method,
    euler = random_rotation_euler(),
    rodrigues = random_rotation_rodrigues()
  )
}

#' @noRd
#' @keywords internal
random_rotation_euler <- function() {
  alpha <- stats::runif(1, 0, 2 * pi)
  beta <- acos(stats::runif(1, -1, 1))
  gamma <- stats::runif(1, 0, 2 * pi)
  ca <- cos(alpha)
  sa <- sin(alpha)
  cb <- cos(beta)
  sb <- sin(beta)
  cg <- cos(gamma)
  sg <- sin(gamma)
  matrix(c(
    cg * cb * ca - sg * sa,
    sg * cb * ca + cg * sa,
    -sb * ca,
    -cg * cb * sa - sg * ca,
    -sg * cb * sa + cg * ca,
    sb * sa,
    cg * sb,
    sg * sb,
    cb
  ), nrow = 3, ncol = 3)
}

#' @noRd
#' @keywords internal
random_rotation_rodrigues <- function() {
  u <- stats::rnorm(3)
  u <- u / sqrt(sum(u^2))
  theta <- stats::runif(1, 0, 2 * pi)
  ct <- cos(theta)
  st <- sin(theta)
  ux <- u[1]
  uy <- u[2]
  uz <- u[3]
  matrix(c(
    ct + ux^2 * (1 - ct),
    uy * ux * (1 - ct) + uz * st,
    uz * ux * (1 - ct) - uy * st,
    ux * uy * (1 - ct) - uz * st,
    ct + uy^2 * (1 - ct),
    uz * uy * (1 - ct) + ux * st,
    ux * uz * (1 - ct) + uy * st,
    uy * uz * (1 - ct) - ux * st,
    ct + uz^2 * (1 - ct)
  ), nrow = 3, ncol = 3)
}

#' @noRd
#' @keywords internal
rotate_coords <- function(coords_lh, coords_rh, n_perm,
                          seed = NULL, rotation = "euler") {
  if (!is.null(seed)) withr::local_seed(seed)
  lh_rotated <- array(0, dim = c(nrow(coords_lh), 3, n_perm))
  rh_rotated <- array(0, dim = c(nrow(coords_rh), 3, n_perm))

  for (i in seq_len(n_perm)) {
    rot1 <- random_rotation_matrix(rotation)
    rot2 <- random_rotation_matrix(rotation)
    lh_rotated[, , i] <- coords_lh %*% rot1
    rh_rotated[, , i] <- coords_rh %*% rot2
  }
  list(lh = lh_rotated, rh = rh_rotated)
}

#' @noRd
#' @keywords internal
compute_cost_matrix <- function(original, rotated) {
  a <- rowSums(original^2)
  b <- rowSums(rotated^2)
  outer(a, b, "+") - 2 * tcrossprod(original, rotated)
}

#' @noRd
#' @keywords internal
assign_parcels_vasa <- function(cost_matrix) {
  n <- nrow(cost_matrix)
  assignment <- integer(n)
  available <- rep(TRUE, n)

  for (i in seq_len(n)) {
    row_order <- order(cost_matrix[i, ])
    for (j in row_order) {
      if (available[j]) {
        assignment[i] <- j
        available[j] <- FALSE
        break
      }
    }
  }
  assignment
}

#' @noRd
#' @keywords internal
assign_parcels_hungarian <- function(cost_matrix) {
  sol <- clue::solve_LSAP(cost_matrix)
  as.integer(sol)
}

describe("random_rotation_matrix", {
  it("returns a 3x3 orthogonal matrix", {
    rot <- random_rotation_matrix()
    expect_equal(dim(rot), c(3, 3))
    expect_equal(rot %*% t(rot), diag(3), tolerance = 1e-10)
  })

  it("has determinant of 1", {
    rot <- random_rotation_matrix()
    expect_equal(det(rot), 1, tolerance = 1e-10)
  })
})

describe("rotate_coords", {
  it("returns arrays of correct dimensions", {
    lh <- matrix(rnorm(15), ncol = 3)
    rh <- matrix(rnorm(12), ncol = 3)
    result <- rotate_coords(lh, rh, n_perm = 3, seed = 42)

    expect_equal(dim(result$lh), c(5, 3, 3))
    expect_equal(dim(result$rh), c(4, 3, 3))
  })

  it("is reproducible with seed", {
    lh <- matrix(rnorm(15), ncol = 3)
    rh <- matrix(rnorm(12), ncol = 3)
    r1 <- rotate_coords(lh, rh, n_perm = 3, seed = 42)
    r2 <- rotate_coords(lh, rh, n_perm = 3, seed = 42)
    expect_equal(r1$lh, r2$lh, tolerance = 1e-10)
    expect_equal(r1$rh, r2$rh, tolerance = 1e-10)
  })
})

describe("compute_cost_matrix", {
  it("computes squared distances between original and rotated", {
    original <- matrix(c(1, 0, 0, 0, 1, 0), nrow = 2, ncol = 3, byrow = TRUE)
    rotated <- original
    cost <- compute_cost_matrix(original, rotated)
    expect_equal(diag(cost), c(0, 0))
    expect_true(cost[1, 2] > 0)
  })
})

describe("assign_parcels_vasa", {
  it("returns a valid permutation", {
    cost <- matrix(c(1, 10, 10, 1), nrow = 2)
    assignment <- assign_parcels_vasa(cost)
    expect_length(assignment, 2)
    expect_equal(sort(assignment), 1:2)
  })

  it("assigns to nearest available", {
    cost <- matrix(c(1, 2, 3, 4), nrow = 2)
    assignment <- assign_parcels_vasa(cost)
    expect_equal(assignment[1], 1)
    expect_equal(assignment[2], 2)
  })
})

describe("assign_parcels_hungarian", {
  it("returns optimal assignment", {
    skip_if_not_installed("clue")
    cost <- matrix(c(1, 10, 10, 1), nrow = 2)
    assignment <- assign_parcels_hungarian(cost)
    expect_equal(assignment, c(1L, 2L))
  })
})

describe("null_spin_vasa", {
  it("generates null distribution of correct dimensions", {
    set.seed(42)
    n_lh <- 5
    n_rh <- 5
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    data <- rnorm(n_lh + n_rh)
    result <- null_spin_vasa(data, coords, n_perm = 3L, seed = 1)

    expect_s3_class(result, "null_distribution")
    expect_equal(result$n, 10)
    expect_equal(result$n_perm, 3)
    expect_equal(result$method, "spin_vasa")
  })

  it("errors when data length mismatches coords", {
    coords <- list(
      lh = matrix(rnorm(9), ncol = 3),
      rh = matrix(rnorm(9), ncol = 3)
    )
    expect_error(null_spin_vasa(1:5, coords), "total parcels")
  })

  it("null values are permutations of original data", {
    set.seed(42)
    coords <- list(
      lh = matrix(rnorm(15), ncol = 3),
      rh = matrix(rnorm(15), ncol = 3)
    )
    data <- c(1:10) * 1.0
    result <- null_spin_vasa(data, coords, n_perm = 5L, seed = 1)

    for (i in seq_len(5)) {
      lh_vals <- sort(result$nulls[1:5, i])
      rh_vals <- sort(result$nulls[6:10, i])
      expect_equal(lh_vals, sort(data[1:5]))
      expect_equal(rh_vals, sort(data[6:10]))
    }
  })
})

describe("null_spin_hungarian", {
  it("generates null distribution", {
    skip_if_not_installed("clue")
    set.seed(42)
    n_lh <- 5
    n_rh <- 5
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    data <- rnorm(n_lh + n_rh)
    result <- null_spin_hungarian(data, coords, n_perm = 3L, seed = 1)

    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "spin_hungarian")
    expect_equal(result$n_perm, 3)
  })

  it("is reproducible with seed", {
    skip_if_not_installed("clue")
    set.seed(42)
    coords <- list(
      lh = matrix(rnorm(15), ncol = 3),
      rh = matrix(rnorm(15), ncol = 3)
    )
    data <- rnorm(10)
    r1 <- null_spin_hungarian(data, coords, n_perm = 3L, seed = 1)
    r2 <- null_spin_hungarian(data, coords, n_perm = 3L, seed = 1)
    expect_equal(r1$nulls, r2$nulls, tolerance = 1e-10)
  })
})

#' @noRd
#' @keywords internal
read_surface_coordinates <- function(path) {
  ext <- tools::file_ext(path)
  if (ext == "gii" || grepl("\\.surf\\.gii$", path)) {
    g <- gifti::read_gifti(path)
    g$data[[1]]
  } else {
    cli::cli_abort("Unsupported surface file format: {.file {path}}")
  }
}

#' @noRd
#' @keywords internal
compute_distance_matrix <- function(coords,
                                    method = c("euclidean", "geodesic"),
                                    faces = NULL) {
  method <- match.arg(method)
  if (is.list(coords) && !is.null(coords$lh) && !is.null(coords$rh)) {
    coords <- rbind(coords$lh, coords$rh)
  }
  if (!is.matrix(coords) || ncol(coords) != 3) {
    cli::cli_abort("{.arg coords} must be a matrix with 3 columns (x, y, z).")
  }
  if (method == "geodesic") {
    if (is.null(faces)) {
      cli::cli_abort("{.arg faces} is required for geodesic distance.")
    }
    get_surface_distance(coords, faces)
  } else {
    stats::dist(coords) |> as.matrix()
  }
}

#' @noRd
#' @keywords internal
compute_knn <- function(distmat, k) {
  n <- nrow(distmat)
  k <- min(k, n - 1L)
  indices <- matrix(0L, nrow = n, ncol = k)
  distances <- matrix(0, nrow = n, ncol = k)
  for (i in seq_len(n)) {
    d <- distmat[i, ]
    d[i] <- Inf
    top_k <- order(d)[seq_len(k)]
    indices[i, ] <- top_k
    distances[i, ] <- d[top_k]
  }
  list(indices = indices, distances = distances)
}

#' @noRd
#' @keywords internal
validate_data <- function(data, arg = "data") {
  if (!is.numeric(data)) {
    cli::cli_abort("{.arg {arg}} must be numeric.")
  }
  if (anyNA(data)) {
    cli::cli_abort("{.arg {arg}} must not contain NA values.")
  }
  invisible(data)
}

#' @noRd
#' @keywords internal
validate_distmat <- function(distmat, n, arg = "distmat") {
  if (!is.matrix(distmat)) {
    cli::cli_abort("{.arg {arg}} must be a matrix.")
  }
  if (nrow(distmat) != n || ncol(distmat) != n) {
    cli::cli_abort(
      "{.arg {arg}} must be {n} x {n}, got {nrow(distmat)} x {ncol(distmat)}."
    )
  }
  invisible(distmat)
}

#' Read vertex/voxel values from a brain map file
#'
#' Reads GIFTI (`.func.gii`) or NIfTI (`.nii.gz`) files and returns the
#' values as a numeric vector. Used internally by [compare_maps()] when
#' file paths are passed instead of numeric vectors.
#'
#' @param path Path to a `.func.gii` (GIFTI) or `.nii.gz` (NIfTI) file.
#'
#' @return A numeric vector of map values.
#' @examples
#' \dontrun{
#' read_brain_map_values("cortical_thickness.func.gii")
#' read_brain_map_values("brain_volume.nii.gz")
#' }
#' @export
read_brain_map_values <- function(path) {
  if (grepl("\\.surf\\.gii$", path, ignore.case = TRUE)) {
    cli::cli_abort(c(
      "{.file {path}} is a surface geometry file, not a data file.",
      "i" = "Expected a {.file .func.gii} or {.file .label.gii} file."
    ))
  }
  if (grepl("\\.gii$", path, ignore.case = TRUE)) {
    gii <- gifti::read_gifti(path)
    return(as.numeric(gii$data[[1]]))
  }
  ext <- tolower(tools::file_ext(path))
  if (ext == "gz" || ext == "nii") {
    rlang::check_installed("RNifti", reason = "to read NIfTI volume files")
    img <- RNifti::readNifti(path)
    return(as.numeric(img))
  }
  cli::cli_abort("Unsupported file format: {.file {path}}")
}

#' @noRd
#' @keywords internal
validate_coords <- function(coords) {
  if (!is.list(coords) || is.null(coords$lh) || is.null(coords$rh)) {
    cli::cli_abort(paste(
      "{.arg coords} must be a list with",
      "{.field $lh} and {.field $rh} matrices."
    ))
  }
  if (!is.matrix(coords$lh) || ncol(coords$lh) != 3) {
    cli::cli_abort("{.field coords$lh} must be a matrix with 3 columns.")
  }
  if (!is.matrix(coords$rh) || ncol(coords$rh) != 3) {
    cli::cli_abort("{.field coords$rh} must be a matrix with 3 columns.")
  }
  invisible(coords)
}

#' @noRd
#' @keywords internal
validate_parcellation <- function(labels, n) {
  if (!is.numeric(labels) && !is.integer(labels)) {
    cli::cli_abort("{.arg parcellation} must be an integer or numeric vector.")
  }
  if (length(labels) != n) {
    cli::cli_abort(paste(
      "{.arg parcellation} length ({length(labels)})",
      "must match data length ({n})."
    ))
  }
  invisible(labels)
}

#' @noRd
#' @keywords internal
neuromaps_cache_dir <- function() {
  dir <- getOption(
    "neuromapr.data_dir",
    Sys.getenv("NEUROMAPR_DATA_DIR", unset = "")
  )
  if (nzchar(dir)) return(dir)
  tools::R_user_dir("neuromapr", "cache")
}

#' @noRd
#' @keywords internal
validate_checksum <- function(file, expected) {
  if (is.null(expected) || !nzchar(expected)) return(TRUE)
  actual <- unname(tools::md5sum(file))
  identical(actual, expected)
}

#' @noRd
#' @keywords internal
parse_neuromaps_filename <- function(fname) {
  base <- sub("_feature\\..*$", "", fname)
  ext <- sub("^.*_feature", "", fname)

  pairs <- strsplit(base, "_")[[1]]
  result <- list()
  for (pair in pairs) {
    parts <- strsplit(pair, "-", fixed = TRUE)[[1]]
    if (length(parts) == 2) {
      result[[parts[1]]] <- parts[2]
    }
  }
  result$ext <- ext
  result
}

describe("generate_nulls", {
  it("errors for non-numeric data", {
    expect_error(
      generate_nulls("abc", method = "burt2020"),
      "numeric"
    )
  })

  it("errors for NA in data", {
    expect_error(
      generate_nulls(c(1, NA, 3), method = "burt2020"),
      "NA"
    )
  })

  it("errors for n_perm < 1", {
    expect_error(
      generate_nulls(1:5, method = "burt2020", n_perm = 0, distmat = diag(5)),
      "n_perm"
    )
  })

  it("errors when distmat missing for burt2020", {
    expect_error(
      generate_nulls(1:5, method = "burt2020"),
      "distmat"
    )
  })

  it("errors when distmat missing for moran", {
    expect_error(
      generate_nulls(1:5, method = "moran"),
      "distmat"
    )
  })

  it("errors when coords missing for spin_vasa", {
    expect_error(
      generate_nulls(1:5, method = "spin_vasa"),
      "coords"
    )
  })

  it("errors when coords missing for spin_hungarian", {
    expect_error(
      generate_nulls(1:5, method = "spin_hungarian"),
      "coords"
    )
  })

  it("dispatches to moran", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- generate_nulls(
      data, method = "moran",
      n_perm = 5L, distmat = distmat, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "moran")
    expect_equal(result$n_perm, 5)
  })

  it("dispatches to spin_vasa", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    data <- rnorm(n_lh + n_rh)
    result <- generate_nulls(
      data, method = "spin_vasa",
      n_perm = 5L, coords = coords, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "spin_vasa")
  })

  it("errors when coords missing for alexander_bloch", {
    expect_error(
      generate_nulls(1:5, method = "alexander_bloch"),
      "coords"
    )
  })

  it("errors when coords missing for baum", {
    expect_error(
      generate_nulls(1:2, method = "baum"),
      "coords"
    )
  })

  it("errors when parcellation missing for baum", {
    coords <- list(
      lh = matrix(rnorm(15), ncol = 3),
      rh = matrix(rnorm(15), ncol = 3)
    )
    expect_error(
      generate_nulls(1:2, method = "baum", coords = coords),
      "parcellation"
    )
  })

  it("errors when coords missing for cornblath", {
    expect_error(
      generate_nulls(1:2, method = "cornblath"),
      "coords"
    )
  })

  it("errors when parcellation missing for cornblath", {
    coords <- list(
      lh = matrix(rnorm(15), ncol = 3),
      rh = matrix(rnorm(15), ncol = 3)
    )
    expect_error(
      generate_nulls(1:2, method = "cornblath", coords = coords),
      "parcellation"
    )
  })

  it("errors when distmat missing for burt2018", {
    expect_error(
      generate_nulls(1:5, method = "burt2018"),
      "distmat"
    )
  })

  it("dispatches to alexander_bloch", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    data <- rnorm(n_lh + n_rh)
    result <- generate_nulls(
      data, method = "alexander_bloch",
      n_perm = 5L, coords = coords, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "alexander_bloch")
  })

  it("dispatches to burt2018", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- generate_nulls(
      data, method = "burt2018",
      n_perm = 5L, distmat = distmat, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "burt2018")
  })

  it("dispatches to burt2020", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- generate_nulls(
      data, method = "burt2020",
      n_perm = 3L, distmat = distmat, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "burt2020")
  })

  it("dispatches to spin_hungarian", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    data <- rnorm(n_lh + n_rh)
    result <- generate_nulls(
      data, method = "spin_hungarian",
      n_perm = 3L, coords = coords, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "spin_hungarian")
  })

  it("dispatches to baum", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    parcellation <- rep(1:4, each = 5)
    data <- rnorm(4)
    result <- generate_nulls(
      data, method = "baum",
      n_perm = 3L, coords = coords,
      parcellation = parcellation, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "baum")
  })

  it("dispatches to cornblath", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    parcellation <- rep(1:4, each = 5)
    data <- rnorm(4)
    result <- generate_nulls(
      data, method = "cornblath",
      n_perm = 3L, coords = coords,
      parcellation = parcellation, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "cornblath")
  })
})

describe("compute_distance_matrix", {
  it("computes from a coordinate matrix", {
    coords <- matrix(c(0, 1, 0, 0, 0, 1), nrow = 2, ncol = 3)
    dm <- compute_distance_matrix(coords)
    expect_equal(nrow(dm), 2)
    expect_equal(ncol(dm), 2)
    expect_equal(dm[1, 1], 0)
    expect_true(dm[1, 2] > 0)
  })

  it("computes from a list with lh/rh", {
    coords <- list(
      lh = matrix(c(0, 1, 0, 0, 0, 0), nrow = 2, ncol = 3),
      rh = matrix(c(2, 3, 0, 0, 0, 0), nrow = 2, ncol = 3)
    )
    dm <- compute_distance_matrix(coords)
    expect_equal(nrow(dm), 4)
  })

  it("errors for non-3-column matrix", {
    expect_error(
      compute_distance_matrix(matrix(1:4, ncol = 2)),
      "3 columns"
    )
  })
})

describe("neuromaps_cache_dir", {
  it("uses option when set", {
    withr::local_options(neuromapr.data_dir = "/tmp/custom-cache")
    expect_equal(neuromaps_cache_dir(), "/tmp/custom-cache")
  })

  it("uses env var when option is not set", {
    withr::local_options(neuromapr.data_dir = NULL)
    withr::local_envvar(NEUROMAPR_DATA_DIR = "/tmp/env-cache")
    expect_equal(neuromaps_cache_dir(), "/tmp/env-cache")
  })

  it("falls back to R_user_dir", {
    withr::local_options(neuromapr.data_dir = NULL)
    withr::local_envvar(NEUROMAPR_DATA_DIR = "")
    result <- neuromaps_cache_dir()
    expect_true(grepl("neuromapr", result))
  })
})


describe("validate_checksum", {
  it("returns TRUE for matching checksum", {
    tmp <- withr::local_tempfile()
    writeLines("test content", tmp)
    md5 <- unname(tools::md5sum(tmp))
    expect_true(validate_checksum(tmp, md5))
  })

  it("returns FALSE for mismatched checksum", {
    tmp <- withr::local_tempfile()
    writeLines("test content", tmp)
    expect_false(validate_checksum(tmp, "0000000000000000"))
  })

  it("returns TRUE when expected is NULL", {
    tmp <- withr::local_tempfile()
    writeLines("test content", tmp)
    expect_true(validate_checksum(tmp, NULL))
  })

  it("returns TRUE when expected is empty string", {
    tmp <- withr::local_tempfile()
    writeLines("test content", tmp)
    expect_true(validate_checksum(tmp, ""))
  })
})


describe("parse_neuromaps_filename", {
  it("parses surface GIFTI filename", {
    fname <- paste0(
      "source-abagen_desc-genepc1_space-fsaverage",
      "_den-10k_hemi-L_feature.func.gii"
    )
    result <- parse_neuromaps_filename(fname)
    expect_equal(result$source, "abagen")
    expect_equal(result$desc, "genepc1")
    expect_equal(result$space, "fsaverage")
    expect_equal(result$den, "10k")
    expect_equal(result$hemi, "L")
    expect_equal(result$ext, ".func.gii")
  })

  it("parses volume NIfTI filename", {
    result <- parse_neuromaps_filename(
      "source-aghourian2017_desc-feobv_space-MNI152_res-1mm_feature.nii.gz"
    )
    expect_equal(result$source, "aghourian2017")
    expect_equal(result$desc, "feobv")
    expect_equal(result$space, "MNI152")
    expect_equal(result$res, "1mm")
    expect_null(result$hemi)
    expect_equal(result$ext, ".nii.gz")
  })
})

describe("compute_mem RSpectra branch (coverage)", {
  it("uses RSpectra for large matrices when available", {
    skip_if_not_installed("RSpectra")
    n <- 5001
    w <- matrix(0, n, n)
    idx <- cbind(seq_len(n - 1), 2:n)
    w[idx] <- 1
    w <- w + t(w)
    rs <- rowSums(w)
    rs[rs == 0] <- 1
    w <- w / rs

    mem <- compute_mem(w, n)
    expect_true(is.matrix(mem$vectors))
    expect_true(is.numeric(mem$values))
    expect_true(length(mem$values) <= 500)
  })
})

describe("null_moran pair rotation (coverage)", {
  it("exercises pair rotation with degenerate eigenvalues", {
    n <- 6
    angles <- seq(0, 2 * pi, length.out = n + 1)[seq_len(n)]
    coords <- cbind(cos(angles), sin(angles), 0)
    distmat <- as.matrix(dist(coords))
    data <- rnorm(n)

    w <- compute_weight_matrix(distmat)
    mem <- compute_mem(w, n)
    pairs <- make_pairs(mem$values, tol = 1e-6)
    has_pair <- any(vapply(pairs, length, integer(1)) == 2)
    expect_true(has_pair)

    result <- null_moran(
      data, distmat,
      n_perm = 5L, seed = 1, procedure = "pair"
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$n_perm, 5)
  })
})

describe("null_spin_hungarian data mismatch (coverage)", {
  it("errors when data length mismatches coords", {
    coords <- list(
      lh = matrix(rnorm(15), ncol = 3),
      rh = matrix(rnorm(15), ncol = 3)
    )
    expect_error(
      null_spin_hungarian(1:5, coords, n_perm = 3L),
      "total parcels"
    )
  })
})

describe("rodrigues rotation (coverage)", {
  it("produces valid rotation matrices", {
    set.seed(42)
    n_lh <- 5
    n_rh <- 5
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    data <- rnorm(n_lh + n_rh)

    result <- null_spin_vasa(
      data, coords,
      n_perm = 3L, seed = 1,
      rotation = "rodrigues"
    )
    expect_s3_class(result, "null_distribution")
    expect_equal(result$n_perm, 3)
  })
})

describe("null_burt2020 via generate_nulls with fixed_idx (coverage)", {
  it("exercises fixed_idx path when resample=FALSE", {
    set.seed(42)
    n <- 30
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- null_burt2020(
      data, distmat,
      n_perm = 3L, seed = 1, ns = 15L, resample = FALSE
    )
    expect_s3_class(result, "null_distribution")
  })
})

describe("neuromaps-registry refresh and cache (coverage)", {
  it("build_neuromaps_registry with refresh=TRUE clears cache", {
    the$registry <- tibble::tibble(source = "stale")
    the$osf_json <- "stale"
    the$meta_json <- "stale"

    mock_fname <- paste0(
      "source-t_desc-m_space-fs",
      "_den-10k_hemi-L_feature.func.gii"
    )

    local_mocked_bindings(
      fetch_neuromaps_osf_json = function() {
        list(annotations = list(
          list(
            source = "t", desc = "m",
            space = "fs", den = "10k", res = NULL,
            hemi = "L", format = "surface",
            fname = mock_fname,
            rel_path = "t/m/fs/",
            checksum = "aaa", tags = list("pet"),
            url = list("4mw3a", "f1")
          )
        ))
      },
      fetch_neuromaps_meta_json = function() {
        list(annotations = list(
          list(
            annot = list(
              source = "t", desc = "m",
              space = "fs", den = "10k"
            ),
            full_desc = "Test",
            demographics = list(N = 10, age = "30")
          )
        ))
      }
    )

    result <- build_neuromaps_registry(refresh = TRUE)
    expect_equal(result$source, "t")

    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL
  })

  it("clear_neuromaps_cache resets all cached data", {
    the$osf_json <- "data"
    the$meta_json <- "data"
    the$registry <- tibble::tibble(source = "cached")

    result <- clear_neuromaps_cache()
    expect_null(the$osf_json)
    expect_null(the$meta_json)
    expect_null(the$registry)
    expect_null(result)
  })
})

describe("read_surface_coordinates (coverage)", {
  it("errors for unsupported format", {
    expect_error(
      read_surface_coordinates("file.obj"),
      "Unsupported"
    )
  })
})

describe("validate_distmat (coverage)", {
  it("errors for non-matrix input", {
    expect_error(
      validate_distmat(1:5, 5),
      "matrix"
    )
  })
})

describe("read_brain_map_values branches (coverage)", {
  it("errors for .surf.gii files", {
    expect_error(
      read_brain_map_values("test.surf.gii"),
      "surface geometry"
    )
  })

  it("errors for unsupported format", {
    expect_error(
      read_brain_map_values("test.xyz"),
      "Unsupported"
    )
  })
})

describe("validate_coords (coverage)", {
  it("errors for non-list input", {
    expect_error(validate_coords("not a list"), "list")
  })

  it("errors for bad lh matrix", {
    coords <- list(
      lh = matrix(1:4, ncol = 2),
      rh = matrix(1:6, ncol = 3)
    )
    expect_error(validate_coords(coords), "3 columns")
  })

  it("errors for bad rh matrix", {
    coords <- list(
      lh = matrix(1:6, ncol = 3),
      rh = matrix(1:4, ncol = 2)
    )
    expect_error(validate_coords(coords), "3 columns")
  })
})

describe("density_to_n unknown density (coverage)", {
  it("errors for unknown density string", {
    expect_error(density_to_n("99k"), "Unknown density")
  })
})

describe("parcellate with file-like args (coverage)", {
  it("parcellate reads brain map values from path", {
    local_mocked_bindings(
      read_brain_map_values = function(path) c(1, 2, 3, 4)
    )
    local_mocked_bindings(
      read_parcellation_labels = function(path) c(1L, 1L, 2L, 2L)
    )

    result <- parcellate("fake_data.func.gii", "fake_parc.label.gii")
    expect_equal(unname(result), c(1.5, 3.5))
  })

  it("unparcellate reads parcellation from path", {
    local_mocked_bindings(
      read_parcellation_labels = function(path) c(1L, 1L, 2L, 2L)
    )

    parcel_data <- c("1" = 10, "2" = 20)
    result <- unparcellate(parcel_data, "fake_parc.label.gii")
    expect_equal(result, c(10, 10, 20, 20))
  })
})

describe("get_parcel_centroids surface and geodesic (coverage)", {
  it("computes geodesic centroids", {
    vertices <- matrix(c(
      0, 0, 0,
      1, 0, 0,
      0, 1, 0,
      1, 1, 0
    ), ncol = 3, byrow = TRUE)
    faces <- matrix(c(1L, 2L, 3L, 2L, 3L, 4L), ncol = 3, byrow = TRUE)
    labels <- c(1, 1, 1, 1)

    centroids <- get_parcel_centroids(
      vertices, labels,
      method = "geodesic", faces = faces
    )
    expect_equal(nrow(centroids), 1)
    expect_equal(ncol(centroids), 3)
  })

  it("errors for invalid vertices", {
    expect_error(
      get_parcel_centroids(matrix(1:4, ncol = 2), c(1, 1)),
      "3 columns"
    )
  })
})

describe("resample_images strategies (coverage)", {
  it("errors for downsample_only with different spaces", {
    src <- withr::local_tempfile(fileext = ".gii")
    trg <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    file.create(trg)
    expect_error(
      resample_images(
        src, trg,
        src_space = "fsaverage",
        trg_space = "fsLR",
        strategy = "downsample_only"
      ),
      "same space"
    )
  })

  it("errors for missing target file", {
    src <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    expect_error(
      resample_images(
        src, "nonexistent.gii",
        src_space = "fsaverage",
        trg_space = "fsaverage"
      ),
      "not found"
    )
  })

  it("downsample_only returns same files when equal density", {
    src <- withr::local_tempfile(fileext = ".gii")
    trg <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    file.create(trg)

    local_mocked_bindings(
      get_gifti_density = function(path) "10k"
    )

    result <- resample_images(
      src, trg,
      src_space = "fsaverage",
      trg_space = "fsaverage",
      strategy = "downsample_only"
    )
    expect_equal(result$src, src)
    expect_equal(result$trg, trg)
  })

  it("downsample_only downsamples source when higher", {
    src <- withr::local_tempfile(fileext = ".gii")
    trg <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    file.create(trg)

    density_call <- 0L
    local_mocked_bindings(
      get_gifti_density = function(path) {
        density_call <<- density_call + 1L
        if (density_call == 1L) "32k" else "10k"
      },
      transform_to_space = function(paths, ...) "resampled.gii"
    )

    result <- resample_images(
      src, trg,
      src_space = "fsaverage",
      trg_space = "fsaverage",
      strategy = "downsample_only"
    )
    expect_equal(result$src, "resampled.gii")
    expect_equal(result$trg, trg)
  })

  it("downsample_only downsamples target when higher", {
    src <- withr::local_tempfile(fileext = ".gii")
    trg <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    file.create(trg)

    density_call <- 0L
    local_mocked_bindings(
      get_gifti_density = function(path) {
        density_call <<- density_call + 1L
        if (density_call == 1L) "10k" else "32k"
      },
      transform_to_space = function(paths, ...) "resampled.gii"
    )

    result <- resample_images(
      src, trg,
      src_space = "fsaverage",
      trg_space = "fsaverage",
      strategy = "downsample_only"
    )
    expect_equal(result$src, src)
    expect_equal(result$trg, "resampled.gii")
  })

  it("transform_to_src transforms target to source space", {
    src <- withr::local_tempfile(fileext = ".gii")
    trg <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    file.create(trg)

    local_mocked_bindings(
      get_gifti_density = function(path) "10k",
      transform_to_space = function(...) "resampled.gii"
    )

    result <- resample_images(
      src, trg,
      src_space = "fsaverage",
      trg_space = "fsLR",
      strategy = "transform_to_src"
    )
    expect_equal(result$src, src)
    expect_equal(result$trg, "resampled.gii")
  })

  it("transform_to_trg transforms source to target space", {
    src <- withr::local_tempfile(fileext = ".gii")
    trg <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    file.create(trg)

    local_mocked_bindings(
      get_gifti_density = function(path) "10k",
      transform_to_space = function(...) "resampled.gii"
    )

    result <- resample_images(
      src, trg,
      src_space = "fsaverage",
      trg_space = "fsLR",
      strategy = "transform_to_trg"
    )
    expect_equal(result$src, "resampled.gii")
    expect_equal(result$trg, trg)
  })

  it("transform_to_alt transforms both to alt space", {
    src <- withr::local_tempfile(fileext = ".gii")
    trg <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    file.create(trg)

    local_mocked_bindings(
      get_gifti_density = function(path) "10k",
      transform_to_space = function(...) "resampled.gii"
    )

    result <- resample_images(
      src, trg,
      src_space = "fsaverage",
      trg_space = "fsLR",
      strategy = "transform_to_alt",
      alt_space = "fsaverage",
      alt_density = "32k"
    )
    expect_equal(result$src, "resampled.gii")
    expect_equal(result$trg, "resampled.gii")
  })
})

describe("transform_to_space full path (coverage)", {
  it("transforms a file via ciftiTools mock", {
    skip_if_not_installed("ciftiTools")

    tmp <- withr::local_tempfile(fileext = ".func.gii")
    file.create(tmp)

    local_mocked_bindings(
      check_wb_command = function(...) "/mock/wb_command"
    )
    local_mocked_bindings(
      ciftiTools.setOption = function(...) NULL,
      resample_gifti = function(...) NULL,
      .package = "ciftiTools"
    )

    result <- transform_to_space(
      tmp,
      target_space = "fsLR",
      target_density = "32k",
      method = "barycentric",
      verbose = TRUE
    )
    expect_length(result, 1)
    expect_true(grepl("fsLR_32k", result))
  })

  it("uses adaptive method with area surfaces", {
    skip_if_not_installed("ciftiTools")

    tmp <- withr::local_tempfile(fileext = ".func.gii")
    area_cur <- withr::local_tempfile(fileext = ".surf.gii")
    area_new <- withr::local_tempfile(fileext = ".surf.gii")
    file.create(tmp)
    file.create(area_cur)
    file.create(area_new)

    captured_args <- NULL
    local_mocked_bindings(
      check_wb_command = function(...) "/mock/wb_command"
    )
    local_mocked_bindings(
      ciftiTools.setOption = function(...) NULL,
      resample_gifti = function(...) {
        captured_args <<- list(...)
        NULL
      },
      .package = "ciftiTools"
    )

    result <- transform_to_space(
      tmp,
      target_space = "fsaverage",
      target_density = "10k",
      method = "adaptive",
      area_surf_current = area_cur,
      area_surf_new = area_new,
      verbose = FALSE
    )
    expect_true(grepl("fsaverage_10k", result))
  })
})

describe("check_wb_command ciftiTools path (coverage)", {
  it("returns ciftiTools configured path", {
    skip_if_not_installed("ciftiTools")

    tmp <- withr::local_tempfile()
    file.create(tmp)

    local_mocked_bindings(
      ciftiTools.getOption = function(...) tmp,
      .package = "ciftiTools"
    )

    result <- check_wb_command()
    expect_equal(result, tmp)
  })

  it("falls back to system PATH", {
    skip_if_not_installed("ciftiTools")

    local_mocked_bindings(
      ciftiTools.getOption = function(...) NULL,
      .package = "ciftiTools"
    )

    sys_wb <- Sys.which("wb_command")
    if (nzchar(sys_wb)) {
      result <- check_wb_command()
      expect_true(nzchar(result))
    } else {
      withr::local_envvar(PATH = "")
      expect_error(check_wb_command(), "wb_command")
    }
  })
})

describe("annot_to_gifti full path (coverage)", {
  it("converts annotation file to GIFTI", {
    skip_if_not_installed("freesurferformats")

    tmp_annot <- withr::local_tempfile(fileext = ".annot")
    tmp_out <- withr::local_tempfile(fileext = ".label.gii")

    local_mocked_bindings(
      read.fs.annot = function(path) {
        list(label_codes = c(1L, 2L, 3L, 1L, 2L))
      },
      .package = "freesurferformats"
    )
    local_mocked_bindings(
      write_gifti = function(gii, path) {
        writeLines("mock", path)
      },
      .package = "gifti"
    )

    file.create(tmp_annot)
    result <- annot_to_gifti(tmp_annot, output_path = tmp_out)
    expect_equal(result, tmp_out)
    expect_true(file.exists(tmp_out))
  })

  it("auto-generates output path", {
    skip_if_not_installed("freesurferformats")

    tmp_dir <- withr::local_tempdir()
    tmp_annot <- file.path(tmp_dir, "test.annot")
    file.create(tmp_annot)

    local_mocked_bindings(
      read.fs.annot = function(path) {
        list(label_codes = c(1L, 2L))
      },
      .package = "freesurferformats"
    )
    local_mocked_bindings(
      write_gifti = function(gii, path) {
        writeLines("mock", path)
      },
      .package = "gifti"
    )

    result <- annot_to_gifti(tmp_annot)
    expect_true(grepl("\\.label\\.gii$", result))
  })
})

describe("fsmorph_to_gifti full path (coverage)", {
  it("converts morphometry file to GIFTI", {
    skip_if_not_installed("freesurferformats")

    tmp_morph <- withr::local_tempfile(fileext = ".curv")
    tmp_out <- withr::local_tempfile(fileext = ".func.gii")

    local_mocked_bindings(
      read.fs.morph = function(path) c(0.1, 0.2, 0.3),
      .package = "freesurferformats"
    )
    local_mocked_bindings(
      write_gifti = function(gii, path) {
        writeLines("mock", path)
      },
      .package = "gifti"
    )

    file.create(tmp_morph)
    result <- fsmorph_to_gifti(tmp_morph, output_path = tmp_out)
    expect_equal(result, tmp_out)
    expect_true(file.exists(tmp_out))
  })

  it("auto-generates output path", {
    skip_if_not_installed("freesurferformats")

    tmp_dir <- withr::local_tempdir()
    tmp_morph <- file.path(tmp_dir, "test.curv")
    file.create(tmp_morph)

    local_mocked_bindings(
      read.fs.morph = function(path) c(0.1, 0.2),
      .package = "freesurferformats"
    )
    local_mocked_bindings(
      write_gifti = function(gii, path) {
        writeLines("mock", path)
      },
      .package = "gifti"
    )

    result <- fsmorph_to_gifti(tmp_morph)
    expect_true(grepl("\\.func\\.gii$", result))
  })
})

describe("get_gifti_density with mock (coverage)", {
  it("detects density from vertex count", {
    local_mocked_bindings(
      read_gifti = function(path) {
        list(data = list(numeric(10242)))
      },
      .package = "gifti"
    )

    result <- get_gifti_density("fake.func.gii")
    expect_equal(result, "10k")
  })

  it("errors for unknown vertex count", {
    local_mocked_bindings(
      read_gifti = function(path) {
        list(data = list(numeric(999)))
      },
      .package = "gifti"
    )

    expect_error(
      get_gifti_density("fake.func.gii"),
      "Unknown vertex count"
    )
  })
})

describe("parse_neuromaps_filename (coverage)", {
  it("parses BIDS-style filename", {
    result <- parse_neuromaps_filename(
      "source-abagen_desc-genepc1_feature.func.gii"
    )
    expect_equal(result$source, "abagen")
    expect_equal(result$desc, "genepc1")
    expect_equal(result$ext, ".func.gii")
  })
})

describe("neuromaps_cache_dir (coverage)", {
  it("uses option when set", {
    withr::local_options(neuromapr.data_dir = "/tmp/test_cache")
    result <- neuromaps_cache_dir()
    expect_equal(result, "/tmp/test_cache")
  })

  it("uses env var when set", {
    withr::local_options(neuromapr.data_dir = NULL)
    withr::local_envvar(NEUROMAPR_DATA_DIR = "/tmp/env_cache")
    result <- neuromaps_cache_dir()
    expect_equal(result, "/tmp/env_cache")
  })

  it("falls back to R_user_dir", {
    withr::local_options(neuromapr.data_dir = NULL)
    withr::local_envvar(NEUROMAPR_DATA_DIR = "")
    result <- neuromaps_cache_dir()
    expect_true(grepl("neuromapr", result))
  })
})

describe("validate_checksum (coverage)", {
  it("returns TRUE for null expected", {
    expect_true(validate_checksum("anyfile", NULL))
  })

  it("returns TRUE for empty expected", {
    expect_true(validate_checksum("anyfile", ""))
  })
})

describe("compute_knn (coverage)", {
  it("finds k nearest neighbors", {
    dm <- as.matrix(dist(1:5))
    result <- compute_knn(dm, k = 2)
    expect_equal(nrow(result$indices), 5)
    expect_equal(ncol(result$indices), 2)
    expect_equal(result$indices[1, 1], 2L)
    expect_true(all(result$distances > 0))
  })
})

describe("read_surface_coordinates (coverage)", {
  it("reads .surf.gii via gifti", {
    local_mocked_bindings(
      read_gifti = function(path) {
        list(data = list(matrix(rnorm(12), ncol = 3)))
      },
      .package = "gifti"
    )
    result <- read_surface_coordinates("test.surf.gii")
    expect_true(is.matrix(result))
    expect_equal(ncol(result), 3)
  })

  it("reads plain .gii via gifti", {
    local_mocked_bindings(
      read_gifti = function(path) {
        list(data = list(matrix(rnorm(12), ncol = 3)))
      },
      .package = "gifti"
    )
    result <- read_surface_coordinates("test.gii")
    expect_true(is.matrix(result))
  })
})

describe("read_brain_map_values .label.gii (coverage)", {
  it("reads .label.gii files", {
    local_mocked_bindings(
      read_gifti = function(path) {
        list(data = list(c(1L, 2L, 3L)))
      },
      .package = "gifti"
    )
    result <- read_brain_map_values("test.label.gii")
    expect_equal(result, c(1, 2, 3))
  })

  it("reads generic .gii files", {
    local_mocked_bindings(
      read_gifti = function(path) {
        list(data = list(c(0.5, 1.5, 2.5)))
      },
      .package = "gifti"
    )
    result <- read_brain_map_values("test_data.gii")
    expect_equal(result, c(0.5, 1.5, 2.5))
  })
})

describe("read_brain_map_values NIfTI (coverage)", {
  it("reads .nii.gz files", {
    skip_if_not_installed("RNifti")
    local_mocked_bindings(
      readNifti = function(path) array(1:8, dim = c(2, 2, 2)),
      .package = "RNifti"
    )
    result <- read_brain_map_values("test.nii.gz")
    expect_equal(result, as.numeric(1:8))
  })

  it("reads .nii files", {
    skip_if_not_installed("RNifti")
    local_mocked_bindings(
      readNifti = function(path) array(1:4, dim = c(2, 2)),
      .package = "RNifti"
    )
    result <- read_brain_map_values("test.nii")
    expect_equal(result, as.numeric(1:4))
  })
})

describe("null_baum NA branch (coverage)", {
  it("produces NA when rotated labels are all zero", {
    n_lh <- 3
    n_rh <- 3
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    parcellation <- c(1, 0, 0, 0, 0, 2)
    data <- c("1" = 10, "2" = 20)

    local_mocked_bindings(
      rotate_coords = function(...) {
        lh_rot <- array(0, dim = c(n_lh, 3, 1))
        rh_rot <- array(0, dim = c(n_rh, 3, 1))
        list(lh = lh_rot, rh = rh_rot)
      }
    )

    result <- null_baum(
      data, coords, parcellation,
      n_perm = 1L, seed = 1
    )
    expect_s3_class(result, "null_distribution")
    expect_true(any(is.na(result$nulls)))
  })
})

describe("null_cornblath NA branch (coverage)", {
  it("produces NA when no valid nearest labels", {
    n_lh <- 5
    n_rh <- 5
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    parcellation <- c(1, 1, 1, 0, 0, 0, 0, 0, 2, 2)
    data <- c("1" = 10, "2" = 20)

    local_mocked_bindings(
      rotate_coords = function(...) {
        list(
          lh = array(0, dim = c(n_lh, 3, 1)),
          rh = array(0, dim = c(n_rh, 3, 1))
        )
      },
      nearest_valid_label = function(rotated, original, labels, valid) {
        rep(0L, length(labels))
      }
    )

    result <- null_cornblath(
      data, coords, parcellation,
      n_perm = 1L, seed = 1
    )
    expect_s3_class(result, "null_distribution")
  })
})

describe("read_parcellation_labels (coverage)", {
  it("reads labels from GIFTI file", {
    local_mocked_bindings(
      read_gifti = function(path) {
        list(data = list(c(1L, 2L, 3L, 1L)))
      },
      .package = "gifti"
    )
    result <- read_parcellation_labels("fake.label.gii")
    expect_equal(result, c(1L, 2L, 3L, 1L))
  })
})

#' Burt 2018 spatial autoregressive null model
#'
#' Generates surrogate brain maps using a spatial autoregressive (SAR) model.
#' Estimates spatial autocorrelation and distance decay parameters, then
#' generates surrogates by solving the SAR equation and rank-matching to the
#' original data.
#'
#' @template null-params
#' @param distmat Distance matrix between parcels/vertices.
#'
#' @return A [null_distribution] object.
#'
#' @references
#' Burt JB et al. (2018) Nature Neuroscience 21:1251-1259.
#' doi:10.1038/s41593-018-0195-0
#'
#' @examples
#' data <- rnorm(50)
#' distmat <- as.matrix(dist(matrix(rnorm(100), 50, 2)))
#' nd <- null_burt2018(data, distmat, n_perm = 10L, seed = 1L)
#' @export
null_burt2018 <- function(data, distmat, n_perm = 1000L, seed = NULL) {
  validate_data(data)
  n <- length(data)
  validate_distmat(distmat, n)

  params <- estimate_sar_params(data, distmat)
  weight_mat <- build_sar_weights(distmat, params$d0)
  sar_inv <- solve(diag(n) - params$rho * weight_mat)

  if (!is.null(seed)) withr::local_seed(seed)
  nulls <- matrix(0, nrow = n, ncol = n_perm)

  msg <- "Generating burt2018 nulls"
  for (i in cli::cli_progress_along(seq_len(n_perm), msg)) {
    nulls[, i] <- sar_surrogate(sar_inv, data)
  }

  new_null_distribution(nulls, "burt2018", data, list(
    n_perm = n_perm,
    rho = params$rho,
    d0 = params$d0
  ))
}

#' @noRd
#' @keywords internal
estimate_sar_params <- function(data, distmat) {
  d_vals <- distmat[upper.tri(distmat)]
  d_candidates <- stats::quantile(
    d_vals[d_vals > 0], probs = seq(0.1, 0.9, by = 0.1)
  )

  best_sse <- Inf
  best_rho <- 0
  best_d0 <- stats::median(d_vals[d_vals > 0])

  for (d0 in d_candidates) {
    w <- build_sar_weights(distmat, d0)
    wx <- drop(w %*% data)
    fit <- stats::lm.fit(cbind(1, wx), data)
    rho <- fit$coefficients[2]
    sse <- sum(fit$residuals^2)
    if (sse < best_sse) {
      best_sse <- sse
      best_rho <- rho
      best_d0 <- d0
    }
  }

  list(rho = best_rho, d0 = best_d0)
}

#' @noRd
#' @keywords internal
build_sar_weights <- function(distmat, d0) {
  w <- exp(-distmat / d0)
  diag(w) <- 0
  rs <- rowSums(w)
  rs[rs == 0] <- 1
  w / rs
}

#' @noRd
#' @keywords internal
sar_surrogate <- function(sar_inv, data) {
  z <- stats::rnorm(nrow(sar_inv))
  surrogate <- drop(sar_inv %*% z)
  rank_match(surrogate, data)
}

describe("null_burt2018", {
  it("returns null_distribution with correct dims", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))

    result <- null_burt2018(data, distmat, n_perm = 5L, seed = 1)
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "burt2018")
    expect_equal(nrow(result$nulls), 20)
    expect_equal(ncol(result$nulls), 5)
  })

  it("preserves the rank distribution of original data", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))

    result <- null_burt2018(data, distmat, n_perm = 3L, seed = 1)
    for (i in seq_len(3)) {
      expect_equal(sort(result$nulls[, i]), sort(data))
    }
  })

  it("stores estimated parameters", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))

    result <- null_burt2018(data, distmat, n_perm = 3L, seed = 1)
    expect_true("rho" %in% names(result$params))
    expect_true("d0" %in% names(result$params))
    expect_true(is.numeric(result$params$rho))
    expect_true(is.numeric(result$params$d0))
  })

  it("errors for distmat dimension mismatch", {
    expect_error(
      null_burt2018(1:5, diag(3), n_perm = 2L),
      "5 x 5"
    )
  })
})

#' Compute per-vertex surface areas
#'
#' Each triangle distributes one-third of its area to each of its three
#' vertices. Triangle area is computed via the cross-product formula.
#'
#' @param vertices Numeric matrix (n x 3) of vertex coordinates.
#' @param faces Integer matrix (m x 3) of face indices (1-indexed).
#'
#' @return Numeric vector of length `nrow(vertices)`.
#'
#' @examples
#' vertices <- matrix(c(0, 1, 0, 0, 0, 1, 0, 0, 0), nrow = 3, byrow = TRUE)
#' faces <- matrix(c(1L, 2L, 3L), nrow = 1)
#' vertex_areas(vertices, faces)
#' @export
vertex_areas <- function(vertices, faces) {
  if (!is.matrix(vertices) || ncol(vertices) != 3) {
    cli::cli_abort("{.arg vertices} must be a matrix with 3 columns.")
  }
  if (!is.matrix(faces) || ncol(faces) != 3) {
    cli::cli_abort("{.arg faces} must be a matrix with 3 columns.")
  }

  areas <- numeric(nrow(vertices))

  a <- vertices[faces[, 1], , drop = FALSE]
  b <- vertices[faces[, 2], , drop = FALSE]
  c <- vertices[faces[, 3], , drop = FALSE]
  ab <- b - a
  ac <- c - a

  cross <- cbind(
    ab[, 2] * ac[, 3] - ab[, 3] * ac[, 2],
    ab[, 3] * ac[, 1] - ab[, 1] * ac[, 3],
    ab[, 1] * ac[, 2] - ab[, 2] * ac[, 1]
  )
  tri_areas <- 0.5 * sqrt(rowSums(cross^2))
  third_area <- tri_areas / 3

  all_idx <- as.vector(faces)
  all_areas <- rep(third_area, 3)
  accumulated <- rowsum(all_areas, all_idx, reorder = TRUE)
  areas[as.integer(rownames(accumulated))] <- as.vector(accumulated)

  areas
}

#' Convert FreeSurfer annotation to GIFTI
#'
#' Reads a FreeSurfer `.annot` file and writes a GIFTI label file.
#'
#' @param annot_path Path to FreeSurfer `.annot` file.
#' @param output_path Output GIFTI path. If `NULL`, replaces the extension
#'   with `.label.gii`.
#'
#' @return The output file path (invisibly).
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' \dontrun{
#' annot_to_gifti("lh.aparc.annot")
#' annot_to_gifti("lh.aparc.annot", "lh.aparc.label.gii")
#' }
#' @export
annot_to_gifti <- function(annot_path, output_path = NULL) {
  rlang::check_installed(
    "freesurferformats",
    reason = "to read FreeSurfer annotation files"
  )

  if (!file.exists(annot_path)) {
    cli::cli_abort("File not found: {.file {annot_path}}")
  }

  annot <- freesurferformats::read.fs.annot(annot_path)
  labels <- annot$label_codes

  if (is.null(output_path)) {
    output_path <- sub("\\.[^.]+$", ".label.gii", annot_path)
  }

  gii <- structure(
    list(
      data = list(as.integer(labels)),
      data_info = data.frame(
        Intent = "NIFTI_INTENT_LABEL",
        DataType = "NIFTI_TYPE_INT32",
        n = length(labels),
        Dim0 = length(labels),
        stringsAsFactors = FALSE
      )
    ),
    class = "gifti"
  )
  gifti::write_gifti(gii, output_path)

  invisible(output_path)
}

#' Convert FreeSurfer morphometry to GIFTI
#'
#' Reads a FreeSurfer morphometry file (`.curv`, `.thickness`, `.sulc`) and
#' writes a GIFTI func file.
#'
#' @param morph_path Path to FreeSurfer morphometry file.
#' @param output_path Output GIFTI path. If `NULL`, replaces the extension
#'   with `.func.gii`.
#'
#' @return The output file path (invisibly).
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' \dontrun{
#' fsmorph_to_gifti("lh.thickness")
#' fsmorph_to_gifti("lh.thickness", "lh.thickness.func.gii")
#' }
#' @export
fsmorph_to_gifti <- function(morph_path, output_path = NULL) {
  rlang::check_installed(
    "freesurferformats",
    reason = "to read FreeSurfer morphometry files"
  )

  if (!file.exists(morph_path)) {
    cli::cli_abort("File not found: {.file {morph_path}}")
  }

  morph <- freesurferformats::read.fs.morph(morph_path)

  if (is.null(output_path)) {
    output_path <- sub("\\.[^.]+$", ".func.gii", morph_path)
  }

  gii <- structure(
    list(
      data = list(as.numeric(morph)),
      data_info = data.frame(
        Intent = "NIFTI_INTENT_SHAPE",
        DataType = "NIFTI_TYPE_FLOAT32",
        n = length(morph),
        Dim0 = length(morph),
        stringsAsFactors = FALSE
      )
    ),
    class = "gifti"
  )
  gifti::write_gifti(gii, output_path)

  invisible(output_path)
}

describe("vertex_areas", {
  it("computes areas for a single triangle", {
    vertices <- matrix(c(
      0, 0, 0,
      1, 0, 0,
      0, 1, 0
    ), ncol = 3, byrow = TRUE)
    faces <- matrix(c(1, 2, 3), ncol = 3)

    areas <- vertex_areas(vertices, faces)
    expect_equal(length(areas), 3)
    expect_equal(sum(areas), 0.5)
    expect_true(all(areas > 0))
    expect_equal(areas[1], areas[2])
    expect_equal(areas[2], areas[3])
  })

  it("distributes area equally across face vertices", {
    vertices <- matrix(c(
      0, 0, 0,
      2, 0, 0,
      0, 2, 0
    ), ncol = 3, byrow = TRUE)
    faces <- matrix(c(1, 2, 3), ncol = 3)

    areas <- vertex_areas(vertices, faces)
    expect_equal(areas, rep(2 / 3, 3), tolerance = 1e-10)
  })

  it("accumulates from multiple faces", {
    vertices <- matrix(c(
      0, 0, 0,
      1, 0, 0,
      0, 1, 0,
      1, 1, 0
    ), ncol = 3, byrow = TRUE)
    faces <- matrix(c(1, 2, 3, 2, 3, 4), ncol = 3, byrow = TRUE)

    areas <- vertex_areas(vertices, faces)
    expect_equal(sum(areas), 1.0, tolerance = 1e-10)
    expect_true(areas[2] > areas[1])
    expect_true(areas[3] > areas[4])
  })

  it("errors for invalid vertices", {
    expect_error(
      vertex_areas(matrix(1:4, ncol = 2), matrix(1:3, ncol = 3)),
      "3 columns"
    )
  })

  it("errors for invalid faces", {
    expect_error(
      vertex_areas(matrix(1:6, ncol = 3), matrix(1:4, ncol = 2)),
      "3 columns"
    )
  })
})

describe("annot_to_gifti", {
  it("errors for missing file", {
    skip_if_not_installed("freesurferformats")
    skip_if_not_installed("gifti")
    expect_error(annot_to_gifti("nonexistent.annot"), "not found")
  })
})

describe("fsmorph_to_gifti", {
  it("errors for missing file", {
    skip_if_not_installed("freesurferformats")
    skip_if_not_installed("gifti")
    expect_error(fsmorph_to_gifti("nonexistent.curv"), "not found")
  })
})

#' Compare brain maps with spatial null model significance testing
#'
#' Computes the correlation between two brain maps and optionally tests
#' significance using a spatial null model to account for spatial
#' autocorrelation.
#'
#' @param x,y Numeric vectors of brain map values, or file paths to
#'   GIFTI/NIfTI files.
#' @param method Correlation method: `"pearson"` or `"spearman"`.
#' @param null_method Optional null model method for empirical p-values.
#'   One of `"burt2020"`, `"moran"`, `"spin_vasa"`, `"spin_hungarian"`,
#'   `"alexander_bloch"`, `"baum"`, `"cornblath"`, `"burt2018"`,
#'   or `NULL` for parametric only.
#' @param n_perm Integer number of null permutations.
#' @param nulls Pre-computed [null_distribution] object for `x`.
#' @param distmat Distance matrix (passed to null model if needed).
#' @param coords Coordinate list (passed to spin null models if needed).
#' @param seed Optional integer seed for reproducibility.
#' @param na.rm Logical, remove NA values before computing correlation.
#' @param verbose Logical, print progress messages.
#' @param ... Additional arguments passed to [generate_nulls()].
#'
#' @return A `neuromaps_enhanced_comparison` object (inherits
#'   `neuromaps_comparison`) with additional fields `p_null`, `null_method`,
#'   `null_r`, and `n_perm`.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' x <- rnorm(50)
#' y <- x + rnorm(50)
#' compare_maps(x, y, verbose = FALSE)
#' @export
compare_maps <- function(x,
                         y,
                         method = c("pearson", "spearman"),
                         null_method = NULL,
                         n_perm = 1000L,
                         nulls = NULL,
                         distmat = NULL,
                         coords = NULL,
                         seed = NULL,
                         na.rm = TRUE, # nolint: object_name_linter.
                         verbose = TRUE,
                         ...) {
  method <- match.arg(method)

  if (is.character(x) && length(x) == 1 && file.exists(x)) {
    x <- read_brain_map_values(x)
  }

  if (is.character(y) && length(y) == 1 && file.exists(y)) {
    y <- read_brain_map_values(y)
  }

  if (length(x) != length(y)) {
    cli::cli_abort(paste(
      "{.arg x} and {.arg y} must have the same length",
      "({length(x)} vs {length(y)})."
    ))
  }

  if (na.rm) {
    keep <- !is.na(x) & !is.na(y)
    x <- x[keep]
    y <- y[keep]
  }

  test <- stats::cor.test(x, y, method = method)
  r <- unname(test$estimate)
  p <- test$p.value
  n <- length(x)

  p_null <- NULL
  null_r <- NULL
  null_method_used <- null_method

  if (!is.null(nulls)) {
    validate_null_distribution(nulls)
    if (nrow(nulls$nulls) != n) {
      cli::cli_abort(paste(
        "Pre-computed nulls have {nrow(nulls$nulls)} rows",
        "but data has {n} observations."
      ))
    }
    null_method_used <- nulls$method
    n_perm <- nulls$n_perm
    if (verbose) {
      cli::cli_alert_info(
        "Using pre-computed {.val {null_method_used}} null distribution"
      )
    }
    null_r <- compute_null_correlations(nulls$nulls, y, method)
    p_null <- (sum(abs(null_r) >= abs(r)) + 1) / (n_perm + 1)
  } else if (!is.null(null_method)) {
    null_method <- match.arg(
      null_method,
      c("burt2020", "moran", "spin_vasa", "spin_hungarian",
        "alexander_bloch", "baum", "cornblath", "burt2018")
    )
    if (verbose) {
      cli::cli_alert_info(
        "Generating {.val {null_method}} nulls ({n_perm} permutations)"
      )
    }
    null_dist <- generate_nulls(
      x,
      method = null_method,
      n_perm = n_perm,
      distmat = distmat,
      coords = coords,
      seed = seed,
      ...
    )
    null_r <- compute_null_correlations(null_dist$nulls, y, method)
    p_null <- (sum(abs(null_r) >= abs(r)) + 1) / (n_perm + 1)
  }

  structure(
    list(
      r = r,
      p = p,
      p_null = p_null,
      method = method,
      n = n,
      null_method = null_method_used,
      null_r = null_r,
      n_perm = if (!is.null(null_r)) n_perm else NULL
    ),
    class = c("neuromaps_enhanced_comparison", "neuromaps_comparison")
  )
}

#' @noRd
#' @keywords internal
compute_null_correlations <- function(null_matrix, y, method) {
  apply(null_matrix, 2, function(null_x) {
    stats::cor(null_x, y, method = method, use = "complete.obs")
  })
}

#' @rdname compare_maps
#' @export
print.neuromaps_enhanced_comparison <- function(x, ...) {
  cli::cli_h3("Brain Map Comparison")
  cli::cli_text("{.field Method}: {x$method}")
  cli::cli_text(
    "{.field r} = {round(x$r, 4)}, {.field p} = {format.pval(x$p, digits = 3)}"
  )
  cli::cli_text("{.field n} = {x$n}")

  if (!is.null(x$p_null)) {
    cli::cli_rule()
    cli::cli_text(
      "{.field Null model}: {x$null_method} ({x$n_perm} permutations)"
    )
    cli::cli_text("{.field p_null} = {format.pval(x$p_null, digits = 3)}")
  }

  invisible(x)
}

describe("compare_maps", {
  it("computes parametric correlation", {
    set.seed(42)
    x <- rnorm(50)
    y <- x + rnorm(50, sd = 0.5)
    result <- compare_maps(x, y, verbose = FALSE)

    expect_s3_class(result, "neuromaps_enhanced_comparison")
    expect_s3_class(result, "neuromaps_comparison")
    expect_true(result$r > 0)
    expect_true(result$p < 0.05)
    expect_equal(result$n, 50)
    expect_null(result$p_null)
    expect_null(result$null_method)
  })

  it("supports spearman method", {
    set.seed(42)
    x <- rnorm(50)
    y <- x + rnorm(50, sd = 0.5)
    result <- compare_maps(x, y, method = "spearman", verbose = FALSE)
    expect_equal(result$method, "spearman")
  })

  it("handles NA removal", {
    x <- c(1, 2, NA, 4, 5)
    y <- c(5, 4, 3, NA, 1)
    result <- compare_maps(x, y, na.rm = TRUE, verbose = FALSE)
    expect_equal(result$n, 3)
  })

  it("errors for length mismatch", {
    expect_error(compare_maps(1:5, 1:3, verbose = FALSE), "same length")
  })

  it("computes empirical p with moran null", {
    set.seed(42)
    n <- 20
    x <- rnorm(n)
    y <- x + rnorm(n, sd = 0.5)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- compare_maps(
      x, y,
      null_method = "moran",
      n_perm = 10L,
      distmat = distmat,
      seed = 1,
      verbose = FALSE
    )

    expect_s3_class(result, "neuromaps_enhanced_comparison")
    expect_true(!is.null(result$p_null))
    expect_true(result$p_null >= 0 && result$p_null <= 1)
    expect_equal(result$null_method, "moran")
    expect_length(result$null_r, 10)
    expect_equal(result$n_perm, 10)
  })

  it("accepts pre-computed null distribution", {
    set.seed(42)
    n <- 20
    x <- rnorm(n)
    y <- x + rnorm(n, sd = 0.5)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    nd <- null_moran(x, distmat, n_perm = 10L, seed = 1)

    result <- compare_maps(x, y, nulls = nd, verbose = FALSE)
    expect_true(!is.null(result$p_null))
    expect_equal(result$null_method, "moran")
  })

  it("verbose messages for pre-computed nulls", {
    set.seed(42)
    n <- 20
    x <- rnorm(n)
    y <- x + rnorm(n, sd = 0.5)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    nd <- null_moran(x, distmat, n_perm = 10L, seed = 1)

    expect_message(
      compare_maps(x, y, nulls = nd, verbose = TRUE),
      "pre-computed"
    )
  })

  it("verbose messages for generated nulls", {
    set.seed(42)
    n <- 20
    x <- rnorm(n)
    y <- x + rnorm(n, sd = 0.5)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))

    expect_message(
      compare_maps(
        x, y,
        null_method = "moran",
        n_perm = 5L,
        distmat = distmat,
        seed = 1,
        verbose = TRUE
      ),
      "Generating"
    )
  })

  it("errors for mismatched pre-computed nulls", {
    set.seed(42)
    nd <- new_null_distribution(
      matrix(rnorm(50), nrow = 5, ncol = 10),
      "moran", rnorm(5)
    )
    expect_error(
      compare_maps(1:10, 1:10, nulls = nd, verbose = FALSE),
      "rows"
    )
  })
})

describe("compute_null_correlations", {
  it("computes correlation between each null column and y", {
    null_mat <- matrix(rnorm(30), nrow = 10, ncol = 3)
    y <- rnorm(10)
    result <- compute_null_correlations(null_mat, y, "pearson")
    expect_length(result, 3)
    expect_true(all(abs(result) <= 1))
  })
})

describe("print.neuromaps_enhanced_comparison", {
  it("prints without null model info when no nulls used", {
    result <- structure(
      list(r = 0.5, p = 0.001, p_null = NULL, method = "pearson", n = 100,
           null_method = NULL, null_r = NULL, n_perm = NULL),
      class = c("neuromaps_enhanced_comparison", "neuromaps_comparison")
    )
    expect_snapshot(res <- print(result))
    expect_identical(res, result)
  })

  it("prints null model info when nulls used", {
    result <- structure(
      list(r = 0.5, p = 0.001, p_null = 0.05, method = "pearson", n = 100,
           null_method = "burt2020", null_r = rnorm(100), n_perm = 100L),
      class = c("neuromaps_enhanced_comparison", "neuromaps_comparison")
    )
    expect_snapshot(print(result))
  })
})

#' Transform brain maps between coordinate spaces
#'
#' Resamples GIFTI surface files between `fsaverage` and `fsLR` coordinate
#' spaces using Connectome Workbench via `ciftiTools`.
#'
#' @param paths Character vector of GIFTI file paths to transform.
#' @param target_space Target coordinate space: `"fsLR"` or `"fsaverage"`.
#' @param target_density Target mesh density (e.g., `"32k"`, `"164k"`).
#' @param hemisphere Which hemispheres to transform: `"left"`, `"right"`,
#'   or both.
#' @param method Resampling method: `"adaptive"` (default) or
#'   `"barycentric"`.
#' @param area_surf_current Path to the current-resolution area-correction
#'   surface (e.g. midthickness). Recommended when `method = "adaptive"` for
#'   proper vertex-area correction (matches neuromaps Python behaviour).
#' @param area_surf_new Path to the target-resolution area-correction surface.
#'   If `NULL` and `area_surf_current` is provided, ciftiTools resamples it
#'   barycentrically.
#' @param wb_path Path to `wb_command` executable. If `NULL`, auto-detected
#'   via `ciftiTools`.
#' @param verbose Logical, print progress messages.
#'
#' @return Character vector of transformed file paths.
#'
#' @references
#' Robinson EC et al. (2014) NeuroImage 100:414-426.
#' doi:10.1016/j.neuroimage.2014.07.033
#'
#' Robinson EC et al. (2018) NeuroImage 167:150-165.
#' doi:10.1016/j.neuroimage.2017.10.037
#'
#' @examples
#' \dontrun{
#' transform_to_space("map.func.gii", target_space = "fsLR")
#' }
#' @export
transform_to_space <- function(paths,
                               target_space = c("fsLR", "fsaverage"),
                               target_density = "32k",
                               hemisphere = c("left", "right"),
                               method = c("adaptive", "barycentric"),
                               area_surf_current = NULL,
                               area_surf_new = NULL,
                               wb_path = NULL,
                               verbose = TRUE) {
  rlang::check_installed(
    "ciftiTools",
    reason = "to transform brain maps between spaces"
  )
  target_space <- match.arg(target_space)
  method <- match.arg(method)
  hemisphere <- match.arg(hemisphere, several.ok = TRUE)
  wb_path <- check_wb_command(wb_path)
  ciftiTools::ciftiTools.setOption("wb_path", wb_path)

  out_paths <- character(length(paths))

  for (i in seq_along(paths)) {
    path <- paths[i]
    if (!file.exists(path)) {
      cli::cli_abort("File not found: {.file {path}}")
    }

    out_dir <- dirname(path)
    base <- tools::file_path_sans_ext(basename(path))
    out_file <- file.path(
      out_dir,
      paste0(base, "_", target_space, "_", target_density, ".func.gii")
    )

    if (verbose) {
      cli::cli_alert_info(paste(
        "Transforming {.file {basename(path)}}",
        "to {target_space} {target_density}"
      ))
    }

    resample_args <- list(
      original_fname = path,
      target_fname = out_file,
      hemisphere = if (length(hemisphere) == 1) hemisphere else "both",
      file_type = "metric",
      original_res = NULL,
      target_res = as.integer(gsub("k", "000", target_density)),
      resamp_method = if (method == "adaptive") {
        "ADAP_BARY_AREA"
      } else {
        "BARYCENTRIC"
      }
    )
    if (!is.null(area_surf_current)) {
      resample_args$area_original_fname <- area_surf_current
    }
    if (!is.null(area_surf_new)) {
      resample_args$area_target_fname <- area_surf_new
    }
    do.call(ciftiTools::resample_gifti, resample_args)

    out_paths[i] <- out_file
  }

  out_paths
}

#' Check for Connectome Workbench
#'
#' Verifies that `wb_command` is available. If `wb_path` is `NULL`,
#' checks `ciftiTools` default and system PATH.
#'
#' @param wb_path Optional explicit path to `wb_command`.
#'
#' @return Path to `wb_command` executable.
#'
#' @examples
#' \dontrun{
#' check_wb_command()
#' }
#' @export
check_wb_command <- function(wb_path = NULL) {
  rlang::check_installed("ciftiTools", reason = "to locate wb_command")

  if (!is.null(wb_path)) {
    if (!file.exists(wb_path)) {
      cli::cli_abort("wb_command not found at {.file {wb_path}}")
    }
    return(wb_path)
  }

  ct_path <- tryCatch(
    ciftiTools::ciftiTools.getOption("wb_path"),
    error = function(e) NULL
  )
  if (!is.null(ct_path) && nzchar(ct_path) && file.exists(ct_path)) {
    return(ct_path)
  }

  sys_path <- Sys.which("wb_command")
  if (nzchar(sys_path)) {
    return(unname(sys_path))
  }

  cli::cli_abort(c(
    "Connectome Workbench {.code wb_command} not found.",
    i = paste(
      "Install from",
      "{.url https://www.humanconnectome.org/software/get-connectome-workbench}"
    ),
    i = paste(
      "Or set path with",
      "{.code ciftiTools::ciftiTools.setOption(",
      "'wb_path', '/path/to/wb_command')}"
    )
  ))
}

describe("check_wb_command", {
  it("returns provided path if it exists", {
    skip_if_not_installed("ciftiTools")
    tmp <- withr::local_tempfile()
    file.create(tmp)
    expect_equal(check_wb_command(tmp), tmp)
  })

  it("errors for non-existent explicit path", {
    skip_if_not_installed("ciftiTools")
    expect_error(
      check_wb_command("/nonexistent/wb_command"),
      "not found"
    )
  })

  it("errors when wb_command not available anywhere", {
    skip_if_not_installed("ciftiTools")

    local_mocked_bindings(
      ciftiTools.getOption = function(...) NULL,
      .package = "ciftiTools"
    )
    withr::local_envvar(PATH = "")

    expect_error(check_wb_command(), "wb_command")
  })
})

describe("transform_to_space", {
  it("errors for missing ciftiTools", {
    local_mocked_bindings(
      check_installed = function(pkg, ...) {
        rlang::abort(paste0("Package {.pkg ", pkg, "} is required."))
      },
      .package = "rlang"
    )
    expect_error(
      transform_to_space("test.gii"),
      "ciftiTools"
    )
  })

  it("errors for non-existent file", {
    skip_if_not_installed("ciftiTools")

    local_mocked_bindings(
      check_wb_command = function(...) "/mock/wb_command"
    )
    local_mocked_bindings(
      ciftiTools.setOption = function(...) NULL,
      .package = "ciftiTools"
    )

    expect_error(
      transform_to_space("/nonexistent/file.gii", verbose = FALSE),
      "not found"
    )
  })
})

#' Cornblath spin test null model
#'
#' Spin-based null model where each rotated vertex receives the label of its
#' nearest non-medial-wall original vertex, then parcels are reassigned by
#' majority vote among the resulting vertex labels.
#'
#' @template null-params
#' @param coords List with `$lh` and `$rh` matrices of spherical coordinates
#'   (n x 3 each).
#' @param parcellation Integer vector of parcel labels for all vertices.
#'   `0` and `NA` indicate medial wall.
#' @param rotation Rotation generation method: `"euler"` (ZYZ Euler angles,
#'   default, matches neuromaps Python) or `"rodrigues"` (Rodrigues axis-angle
#'   formula).
#'
#' @return A [null_distribution] object.
#'
#' @references
#' Cornblath EJ et al. (2020) Communications Biology 3:590.
#' doi:10.1038/s42003-020-01296-5
#'
#' @examples
#' coords <- list(lh = matrix(rnorm(30), 10, 3), rh = matrix(rnorm(30), 10, 3))
#' parcellation <- c(rep(1L, 5), rep(2L, 5), rep(3L, 5), rep(4L, 5))
#' data <- c(1.0, 2.0, 3.0, 4.0)
#' nd <- null_cornblath(data, coords, parcellation, n_perm = 10L, seed = 1L)
#' @export
null_cornblath <- function(
    data, coords, parcellation,
    n_perm = 1000L, seed = NULL,
    rotation = c("euler", "rodrigues")) {
  validate_data(data)
  validate_coords(coords)
  rotation <- match.arg(rotation)
  n_lh <- nrow(coords$lh)
  n_rh <- nrow(coords$rh)
  n <- n_lh + n_rh
  validate_parcellation(parcellation, n)

  valid <- !is.na(parcellation) & parcellation != 0
  ulabels <- sort(unique(parcellation[valid]))
  n_parcels <- length(ulabels)
  if (length(data) != n_parcels) {
    cli::cli_abort(paste(
      "Length of {.arg data} ({length(data)})",
      "must match number of parcels ({n_parcels})."
    ))
  }
  names(data) <- ulabels

  valid_lh <- which(valid[seq_len(n_lh)])
  valid_rh <- which(valid[n_lh + seq_len(n_rh)])

  rotated <- rotate_coords(
    coords$lh, coords$rh, n_perm, seed,
    rotation = rotation
  )
  nulls <- matrix(0, nrow = n_parcels, ncol = n_perm)

  msg <- "Generating cornblath nulls"
  for (i in cli::cli_progress_along(seq_len(n_perm), msg)) {
    rot_lh <- rotated$lh[, , i]
    rot_rh <- rotated$rh[, , i]

    new_labels_lh <- nearest_valid_label(
      rot_lh, coords$lh, parcellation[seq_len(n_lh)], valid_lh
    )
    new_labels_rh <- nearest_valid_label(
      rot_rh, coords$rh, parcellation[n_lh + seq_len(n_rh)], valid_rh
    )

    new_labels <- c(new_labels_lh, new_labels_rh)

    for (j in seq_along(ulabels)) {
      orig_mask <- which(parcellation == ulabels[j])
      votes <- new_labels[orig_mask]
      votes <- votes[!is.na(votes) & votes != 0]
      if (length(votes) > 0) {
        tbl <- table(votes)
        best_label <- names(tbl)[which.max(tbl)]
        nulls[j, i] <- data[best_label]
      } else {
        nulls[j, i] <- NA_real_
      }
    }
  }

  new_null_distribution(nulls, "cornblath", data, list(n_perm = n_perm))
}

#' @noRd
#' @keywords internal
nearest_valid_label <- function(
    rotated_coords, original_coords,
    labels, valid_idx) {
  valid_coords <- original_coords[valid_idx, , drop = FALSE]
  valid_labels <- labels[valid_idx]
  a <- rowSums(rotated_coords^2)
  b <- rowSums(valid_coords^2)
  dist_sq <- outer(a, b, "+") - 2 * tcrossprod(rotated_coords, valid_coords)
  nearest <- apply(dist_sq, 1, which.min)
  valid_labels[nearest]
}

the <- new.env(parent = emptyenv())

neuromaps_osf_url <- paste0(
  "https://raw.githubusercontent.com/netneurolab/neuromaps/",
  "main/neuromaps/datasets/data/osf.json"
)

neuromaps_meta_url <- paste0(
  "https://raw.githubusercontent.com/netneurolab/neuromaps/",
  "main/neuromaps/datasets/data/meta.json"
)

#' @noRd
#' @keywords internal
fetch_neuromaps_osf_json <- function() {
  if (!is.null(the$osf_json)) return(the$osf_json)
  rlang::check_installed("httr2", reason = "to fetch neuromaps registry")
  resp <- httr2::request(neuromaps_osf_url) |>
    httr2::req_perform()
  the$osf_json <- httr2::resp_body_json(
    resp,
    check_type = FALSE
  )
  the$osf_json
}

#' @noRd
#' @keywords internal
fetch_neuromaps_meta_json <- function() {
  if (!is.null(the$meta_json)) return(the$meta_json)
  rlang::check_installed("httr2", reason = "to fetch neuromaps registry")
  resp <- httr2::request(neuromaps_meta_url) |>
    httr2::req_perform()
  the$meta_json <- httr2::resp_body_json(
    resp,
    check_type = FALSE
  )
  the$meta_json
}

#' @noRd
#' @keywords internal
parse_osf_annotations <- function(osf_data) {
  annotations <- osf_data[["annotations"]]
  rows <- lapply(annotations, function(a) {
    tibble::tibble(
      source   = a[["source"]] %||% NA_character_,
      desc     = a[["desc"]] %||% NA_character_,
      space    = a[["space"]] %||% NA_character_,
      den      = a[["den"]] %||% NA_character_,
      res      = a[["res"]] %||% NA_character_,
      hemi     = a[["hemi"]] %||% NA_character_,
      format   = a[["format"]] %||% NA_character_,
      fname    = a[["fname"]] %||% NA_character_,
      rel_path = a[["rel_path"]] %||% NA_character_,
      checksum = a[["checksum"]] %||% NA_character_,
      tags     = list(a[["tags"]] %||% character()),
      osf_project = a[["url"]][[1]] %||% NA_character_,
      osf_file_id = a[["url"]][[2]] %||% NA_character_
    )
  })
  dplyr::bind_rows(rows)
}

#' @noRd
#' @keywords internal
parse_meta_annotations <- function(meta_data) {
  annotations <- meta_data[["annotations"]]
  rows <- lapply(annotations, function(a) {
    annot <- a[["annot"]]
    demo <- a[["demographics"]] %||% list()
    tibble::tibble(
      source    = annot[["source"]] %||% NA_character_,
      desc      = annot[["desc"]] %||% NA_character_,
      space     = annot[["space"]] %||% NA_character_,
      den       = annot[["den"]] %||% NA_character_,
      res       = annot[["res"]] %||% NA_character_,
      full_desc = a[["full_desc"]] %||% NA_character_,
      N         = demo[["N"]] %||% NA_integer_,
      age       = as.character(demo[["age"]] %||% NA_character_)
    )
  })
  dplyr::bind_rows(rows)
}

#' @noRd
#' @keywords internal
build_neuromaps_registry <- function(refresh = FALSE) {
  if (!is.null(the$registry) && !refresh) return(the$registry)

  if (refresh) {
    the$osf_json <- NULL
    the$meta_json <- NULL
  }

  osf_data <- fetch_neuromaps_osf_json()
  meta_data <- fetch_neuromaps_meta_json()

  osf_tbl <- parse_osf_annotations(osf_data)
  meta_tbl <- parse_meta_annotations(meta_data)

  join_by <- intersect(
    c("source", "desc", "space", "den", "res"),
    intersect(names(osf_tbl), names(meta_tbl))
  )

  the$registry <- dplyr::left_join(osf_tbl, meta_tbl, by = join_by)
  the$registry
}

#' @noRd
#' @keywords internal
filter_neuromaps_registry <- function(
  registry,
  source = NULL,
  desc = NULL,
  space = NULL,
  density = NULL,
  resolution = NULL,
  hemisphere = NULL,
  tags = NULL,
  format = NULL,
  fixed = FALSE
) {
  if (!is.null(source)) {
    registry <- registry[grepl(source, registry$source, fixed = fixed), ]
  }
  if (!is.null(desc)) {
    registry <- registry[grepl(desc, registry$desc, fixed = fixed), ]
  }
  if (!is.null(space)) {
    registry <- registry[grepl(space, registry$space, fixed = fixed), ]
  }
  if (!is.null(density)) {
    registry <- registry[grepl(density, registry$den, fixed = fixed), ]
  }
  if (!is.null(resolution)) {
    registry <- registry[grepl(resolution, registry$res, fixed = fixed), ]
  }
  if (!is.null(hemisphere)) {
    registry <- registry[grepl(hemisphere, registry$hemi, fixed = fixed), ]
  }
  if (!is.null(format)) {
    registry <- registry[grepl(format, registry$format, fixed = fixed), ]
  }

  if (!is.null(tags)) {
    has_all_tags <- vapply(registry$tags, function(t) {
      all(tags %in% t)
    }, logical(1))
    registry <- registry[has_all_tags, ]
  }

  registry
}

#' Clear cached neuromaps registry data
#'
#' Removes the session-level cache of the neuromaps annotation registry,
#' forcing a fresh download on the next call to [neuromaps_available()] or
#' [fetch_neuromaps_annotation()].
#'
#' @return `NULL`, invisibly.
#' @examples
#' clear_neuromaps_cache()
#' @export
clear_neuromaps_cache <- function() {
  the$osf_json <- NULL
  the$meta_json <- NULL
  the$registry <- NULL
  invisible(NULL)
}

#' @noRd
#' @keywords internal
resolve_neuromaps_entries <- function(
  source,
  desc,
  space,
  density = NULL,
  resolution = NULL,
  hemisphere = NULL
) {
  registry <- build_neuromaps_registry()

  matches <- registry[
    registry$source == source &
      registry$desc == desc &
      registry$space == space,
  ]

  if (!is.null(density)) {
    matches <- matches[
      !is.na(matches$den) & matches$den == density,
    ]
  }
  if (!is.null(resolution)) {
    matches <- matches[
      !is.na(matches$res) & matches$res == resolution,
    ]
  }
  if (!is.null(hemisphere)) {
    matches <- matches[
      !is.na(matches$hemi) & matches$hemi == hemisphere,
    ]
  }

  if (nrow(matches) == 0) {
    cli::cli_abort(c(
      "No matching neuromaps annotation found.",
      "i" = "source={.val {source}}, desc={.val {desc}}, space={.val {space}}"
    ))
  }

  matches
}

describe("fetch_neuromaps_osf_json", {
  it("fetches and caches osf.json", {
    the$osf_json <- NULL
    fetch_count <- 0

    local_mocked_bindings(
      req_perform = function(req) "mock_response",
      resp_body_json = function(resp, ...) {
        fetch_count <<- fetch_count + 1
        list(annotations = list())
      },
      .package = "httr2"
    )

    result <- fetch_neuromaps_osf_json()
    expect_type(result, "list")

    fetch_neuromaps_osf_json()
    expect_equal(fetch_count, 1)

    the$osf_json <- NULL
  })
})


describe("fetch_neuromaps_meta_json", {
  it("fetches and caches meta.json", {
    the$meta_json <- NULL
    fetch_count <- 0

    local_mocked_bindings(
      req_perform = function(req) "mock_response",
      resp_body_json = function(resp, ...) {
        fetch_count <<- fetch_count + 1
        list(annotations = list())
      },
      .package = "httr2"
    )

    result <- fetch_neuromaps_meta_json()
    expect_type(result, "list")

    fetch_neuromaps_meta_json()
    expect_equal(fetch_count, 1)

    the$meta_json <- NULL
  })
})


describe("parse_osf_annotations", {
  it("parses annotation entries into tibble", {
    long_fname_l <- paste0(
      "source-abagen_desc-genepc1_space-fsaverage",
      "_den-10k_hemi-L_feature.func.gii"
    )
    long_fname_r <- paste0(
      "source-abagen_desc-genepc1_space-fsaverage",
      "_den-10k_hemi-R_feature.func.gii"
    )

    osf_data <- list(
      fsaverage = list(),
      annotations = list(
        list(
          source = "abagen", desc = "genepc1",
          space = "fsaverage",
          den = "10k", res = NULL, hemi = "L",
          format = "surface",
          fname = long_fname_l,
          rel_path = "abagen/genepc1/fsaverage/",
          checksum = "abc123",
          tags = list("genetics"),
          url = list("4mw3a", "file123")
        ),
        list(
          source = "abagen", desc = "genepc1",
          space = "fsaverage",
          den = "10k", res = NULL, hemi = "R",
          format = "surface",
          fname = long_fname_r,
          rel_path = "abagen/genepc1/fsaverage/",
          checksum = "def456",
          tags = list("genetics"),
          url = list("4mw3a", "file456")
        )
      )
    )

    result <- parse_osf_annotations(osf_data)
    expect_s3_class(result, "tbl_df")
    expect_equal(nrow(result), 2)
    expect_equal(result$osf_project, c("4mw3a", "4mw3a"))
    expect_equal(result$osf_file_id, c("file123", "file456"))
    expect_equal(result$hemi, c("L", "R"))
  })

  it("skips template entries", {
    osf_data <- list(
      fsaverage = list("10k" = list(url = list("4mw3a", "xxx"))),
      annotations = list()
    )
    result <- parse_osf_annotations(osf_data)
    expect_equal(nrow(result), 0)
  })
})


describe("parse_meta_annotations", {
  it("parses meta entries into tibble", {
    meta_data <- list(annotations = list(
      list(
        annot = list(
          source = "abagen", desc = "genepc1",
          space = "fsaverage", den = "10k"
        ),
        full_desc = "Gene expression PC1",
        demographics = list(N = 6, age = "24-55")
      )
    ))

    result <- parse_meta_annotations(meta_data)
    expect_s3_class(result, "tbl_df")
    expect_equal(nrow(result), 1)
    expect_equal(result$source, "abagen")
    expect_equal(result$full_desc, "Gene expression PC1")
    expect_equal(result$N, 6)
    expect_equal(result$age, "24-55")
  })
})


describe("build_neuromaps_registry", {
  it("merges osf and meta data", {
    the$registry <- NULL
    the$osf_json <- NULL
    the$meta_json <- NULL

    mock_fname <- paste0(
      "source-test_desc-map1_space-fsaverage",
      "_den-10k_hemi-L_feature.func.gii"
    )

    local_mocked_bindings(
      fetch_neuromaps_osf_json = function() {
        list(annotations = list(
          list(
            source = "test", desc = "map1",
            space = "fsaverage",
            den = "10k", res = NULL, hemi = "L",
            format = "surface",
            fname = mock_fname,
            rel_path = "test/map1/fsaverage/",
            checksum = "aaa", tags = list("pet"),
            url = list("4mw3a", "f1")
          )
        ))
      },
      fetch_neuromaps_meta_json = function() {
        list(annotations = list(
          list(
            annot = list(
              source = "test", desc = "map1",
              space = "fsaverage", den = "10k"
            ),
            full_desc = "Test map",
            demographics = list(N = 10, age = "30")
          )
        ))
      }
    )

    result <- build_neuromaps_registry()
    expect_s3_class(result, "tbl_df")
    expect_equal(result$full_desc, "Test map")
    expect_equal(result$N, 10)

    the$registry <- NULL
  })

  it("returns cached registry on second call", {
    the$registry <- tibble::tibble(
      source = "cached", desc = "x", space = "fs"
    )

    result <- build_neuromaps_registry()
    expect_equal(result$source, "cached")

    the$registry <- NULL
  })
})


describe("filter_neuromaps_registry", {
  reg <- tibble::tibble(
    source = c("abagen", "beliveau"),
    desc = c("genepc1", "az"),
    space = c("fsaverage", "MNI152"),
    den = c("10k", NA),
    res = c(NA, "1mm"),
    hemi = c("L", NA),
    format = c("surface", "volume"),
    tags = list(c("genetics"), c("pet"))
  )

  it("filters by source", {
    result <- filter_neuromaps_registry(reg, source = "abagen")
    expect_equal(nrow(result), 1)
    expect_equal(result$source, "abagen")
  })

  it("filters by desc", {
    result <- filter_neuromaps_registry(reg, desc = "genepc1")
    expect_equal(nrow(result), 1)
  })

  it("filters by space", {
    result <- filter_neuromaps_registry(reg, space = "MNI152")
    expect_equal(nrow(result), 1)
  })

  it("filters by density", {
    result <- filter_neuromaps_registry(reg, density = "10k")
    expect_equal(nrow(result), 1)
  })

  it("filters by resolution", {
    result <- filter_neuromaps_registry(
      reg, resolution = "1mm"
    )
    expect_equal(nrow(result), 1)
  })

  it("filters by hemisphere", {
    result <- filter_neuromaps_registry(
      reg, hemisphere = "L"
    )
    expect_equal(nrow(result), 1)
  })

  it("filters by format", {
    result <- filter_neuromaps_registry(
      reg, format = "volume"
    )
    expect_equal(nrow(result), 1)
  })

  it("filters by tags with AND logic", {
    reg2 <- tibble::tibble(
      source = c("a", "b", "c"),
      desc = c("x", "y", "z"),
      space = "fs", den = "10k", res = NA,
      hemi = "L", format = "surface",
      tags = list(
        c("pet", "receptor"), c("pet"),
        c("receptor", "pet")
      )
    )
    result <- filter_neuromaps_registry(
      reg2, tags = c("pet", "receptor")
    )
    expect_equal(nrow(result), 2)
  })
})


describe("resolve_neuromaps_entries", {
  it("errors on no match", {
    the$registry <- NULL

    local_mocked_bindings(
      build_neuromaps_registry = function() {
        tibble::tibble(
          source = character(), desc = character(),
          space = character(), den = character(),
          res = character(), hemi = character()
        )
      }
    )

    expect_error(
      resolve_neuromaps_entries(
        "nonexistent", "none", "nowhere"
      ),
      "No matching"
    )

    the$registry <- NULL
  })

  it("returns matching entries", {
    the$registry <- NULL

    local_mocked_bindings(
      build_neuromaps_registry = function() {
        tibble::tibble(
          source = c("test", "test"),
          desc = c("m", "m"),
          space = c("fs", "fs"),
          den = c("10k", "10k"),
          res = c(NA, NA), hemi = c("L", "R"),
          fname = c("f1.gii", "f2.gii")
        )
      }
    )

    result <- resolve_neuromaps_entries(
      "test", "m", "fs", density = "10k"
    )
    expect_equal(nrow(result), 2)

    the$registry <- NULL
  })

  it("filters by resolution", {
    the$registry <- NULL

    local_mocked_bindings(
      build_neuromaps_registry = function() {
        tibble::tibble(
          source = c("test", "test"),
          desc = c("m", "m"),
          space = c("mni", "mni"),
          den = c(NA, NA),
          res = c("1mm", "2mm"),
          hemi = c(NA, NA),
          fname = c("f1.nii.gz", "f2.nii.gz")
        )
      }
    )

    result <- resolve_neuromaps_entries(
      "test", "m", "mni", resolution = "1mm"
    )
    expect_equal(nrow(result), 1)

    the$registry <- NULL
  })

  it("filters by hemisphere", {
    the$registry <- NULL

    local_mocked_bindings(
      build_neuromaps_registry = function() {
        tibble::tibble(
          source = c("test", "test"),
          desc = c("m", "m"),
          space = c("fs", "fs"),
          den = c("10k", "10k"),
          res = c(NA, NA), hemi = c("L", "R"),
          fname = c("f1.gii", "f2.gii")
        )
      }
    )

    result <- resolve_neuromaps_entries(
      "test", "m", "fs", hemisphere = "L"
    )
    expect_equal(nrow(result), 1)
    expect_equal(result$hemi, "L")

    the$registry <- NULL
  })
})

#' Moran spectral randomization null model
#'
#' Generates spatially-constrained surrogate brain maps using Moran's
#' eigenvector maps (MEMs) for spectral randomization.
#'
#' @template null-params
#' @param distmat Distance matrix between parcels/vertices.
#' @param procedure Character, either `"pair"` (default, random 2D rotations
#'   of near-degenerate eigenvector pairs, matches neuromaps Python) or
#'   `"singleton"` (random sign flips of individual eigenvectors).
#' @param kernel Weight matrix kernel: `"inverse_distance"` (default, matches
#'   neuromaps Python), `"exponential"`, `"gaussian"`, or `"bisquare"`.
#' @param tol Numeric tolerance for eigenvalue comparison.
#'
#' @return A [null_distribution] object.
#'
#' @references
#' Wagner HH, Dray S (2015) Methods in Ecology and Evolution 6:1169-1178.
#' doi:10.1111/2041-210X.12407
#'
#' @examples
#' data <- rnorm(50)
#' distmat <- as.matrix(dist(matrix(rnorm(100), 50, 2)))
#' nd <- null_moran(data, distmat, n_perm = 10L, seed = 1L)
#' @export
null_moran <- function(data,
                       distmat,
                       n_perm = 1000L,
                       seed = NULL,
                       procedure = c("pair", "singleton"),
                       kernel = c("inverse_distance", "exponential",
                                  "gaussian", "bisquare"),
                       tol = 1e-6) {
  validate_data(data)
  procedure <- match.arg(procedure)
  kernel <- match.arg(kernel)
  n <- length(data)
  validate_distmat(distmat, n)

  weight_mat <- compute_weight_matrix(distmat, kernel = kernel)
  mem <- compute_mem(weight_mat, n)

  eigvecs <- mem$vectors
  eigvals <- mem$values
  n_mem <- ncol(eigvecs)

  coeffs <- drop(crossprod(eigvecs, data - mean(data)))

  if (!is.null(seed)) withr::local_seed(seed)
  nulls <- matrix(0, nrow = n, ncol = n_perm)

  if (procedure == "singleton") {
    msg <- "Generating moran nulls"
    for (i in cli::cli_progress_along(seq_len(n_perm), msg)) {
      signs <- sample(c(-1, 1), n_mem, replace = TRUE)
      nulls[, i] <- eigvecs %*% (coeffs * signs) + mean(data)
    }
  } else {
    pairs <- make_pairs(eigvals, tol)
    msg <- "Generating moran nulls"
    for (i in cli::cli_progress_along(seq_len(n_perm), msg)) {
      new_coeffs <- coeffs
      for (pair in pairs) {
        if (length(pair) == 1) {
          new_coeffs[pair] <- new_coeffs[pair] * sample(c(-1, 1), 1)
        } else {
          theta <- stats::runif(1, 0, 2 * pi)
          ct <- cos(theta)
          st <- sin(theta)
          a <- new_coeffs[pair[1]]
          b <- new_coeffs[pair[2]]
          new_coeffs[pair[1]] <- a * ct - b * st
          new_coeffs[pair[2]] <- a * st + b * ct
        }
      }
      nulls[, i] <- eigvecs %*% new_coeffs + mean(data)
    }
  }

  new_null_distribution(nulls, "moran", data, list(
    procedure = procedure,
    kernel = kernel,
    n_mem = n_mem,
    tol = tol
  ))
}

#' @noRd
#' @keywords internal
compute_weight_matrix <- function(
    distmat,
    kernel = c("inverse_distance", "exponential",
               "gaussian", "bisquare")) {
  kernel <- match.arg(kernel)

  w <- switch(kernel,
    inverse_distance = {
      inv <- 1 / (distmat + .Machine$double.eps)
      diag(inv) <- 0
      inv
    },
    exponential = {
      bw <- stats::median(distmat[distmat > 0])
      exp(-distmat / bw)
    },
    gaussian = {
      bw <- stats::median(distmat[distmat > 0])
      exp(-0.5 * (distmat / bw)^2)
    },
    bisquare = {
      bw <- stats::median(distmat[distmat > 0])
      u <- distmat / bw
      ifelse(u < 1, (1 - u^2)^2, 0)
    }
  )
  diag(w) <- 0
  rs <- rowSums(w)
  rs[rs == 0] <- 1
  w / rs
}

#' @noRd
#' @keywords internal
compute_mem <- function(weight_mat, n) {
  sym_w <- (weight_mat + t(weight_mat)) / 2
  row_means <- rowMeans(sym_w)
  grand_mean <- mean(row_means)
  dbl_centered <- sym_w - outer(row_means, row_means, "+") + grand_mean

  if (n > 5000 && rlang::is_installed("RSpectra")) {
    k <- min(n - 1, 500)
    eig <- RSpectra::eigs_sym(dbl_centered, k = k)
  } else {
    eig <- eigen(dbl_centered, symmetric = TRUE)
  }

  keep <- abs(eig$values) > 1e-10
  list(
    vectors = eig$vectors[, keep, drop = FALSE],
    values = eig$values[keep]
  )
}

#' @noRd
#' @keywords internal
make_pairs <- function(eigvals, tol = 1e-6) {
  n <- length(eigvals)
  used <- logical(n)
  pairs <- list()

  for (i in seq_len(n)) {
    if (used[i]) next
    paired <- FALSE
    if (i < n) {
      for (j in (i + 1):n) {
        if (!used[j] && abs(eigvals[i] - eigvals[j]) < tol) {
          pairs <- c(pairs, list(c(i, j)))
          used[c(i, j)] <- TRUE
          paired <- TRUE
          break
        }
      }
    }
    if (!paired) {
      pairs <- c(pairs, list(i))
      used[i] <- TRUE
    }
  }
  pairs
}

describe("compute_weight_matrix", {
  it("produces row-standardized matrix", {
    dm <- as.matrix(dist(matrix(rnorm(30), ncol = 3)))
    w <- compute_weight_matrix(dm)
    rs <- rowSums(w)
    expect_equal(unname(rs), rep(1, 10), tolerance = 1e-10)
    expect_true(all(diag(w) == 0))
  })

  it("supports gaussian kernel", {
    dm <- as.matrix(dist(matrix(rnorm(30), ncol = 3)))
    w <- compute_weight_matrix(dm, kernel = "gaussian")
    expect_equal(unname(rowSums(w)), rep(1, 10), tolerance = 1e-10)
  })

  it("supports bisquare kernel", {
    dm <- as.matrix(dist(matrix(rnorm(30), ncol = 3)))
    w <- compute_weight_matrix(dm, kernel = "bisquare")
    expect_true(all(diag(w) == 0))
  })
})

describe("compute_mem", {
  it("returns eigenvectors and eigenvalues", {
    dm <- as.matrix(dist(matrix(rnorm(30), ncol = 3)))
    w <- compute_weight_matrix(dm)
    mem <- compute_mem(w, 10)
    expect_true(is.matrix(mem$vectors))
    expect_true(is.numeric(mem$values))
    expect_equal(nrow(mem$vectors), 10)
  })

  it("eigenvectors are orthogonal", {
    dm <- as.matrix(dist(matrix(rnorm(30), ncol = 3)))
    w <- compute_weight_matrix(dm)
    mem <- compute_mem(w, 10)
    cross <- crossprod(mem$vectors)
    expect_equal(cross, diag(ncol(mem$vectors)), tolerance = 1e-8)
  })
})

describe("make_pairs", {
  it("pairs eigenvalues within tolerance", {
    eigvals <- c(1.0, 1.0, 0.5, 0.2)
    pairs <- make_pairs(eigvals, tol = 1e-6)
    expect_equal(pairs[[1]], c(1, 2))
    expect_equal(pairs[[2]], 3)
    expect_equal(pairs[[3]], 4)
  })

  it("leaves singletons unpaired", {
    eigvals <- c(3, 2, 1)
    pairs <- make_pairs(eigvals)
    expect_length(pairs, 3)
    expect_equal(pairs[[1]], 1)
    expect_equal(pairs[[2]], 2)
    expect_equal(pairs[[3]], 3)
  })
})

describe("null_moran", {
  it("generates null distribution with singleton procedure", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- null_moran(
      data, distmat,
      n_perm = 5L, seed = 1, procedure = "singleton"
    )

    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "moran")
    expect_equal(result$n_perm, 5)
    expect_equal(result$n, 20)
    expect_equal(result$params$procedure, "singleton")
  })

  it("generates null distribution with pair procedure", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- null_moran(
      data, distmat,
      n_perm = 5L, seed = 1, procedure = "pair"
    )

    expect_s3_class(result, "null_distribution")
    expect_equal(result$params$procedure, "pair")
  })

  it("is reproducible with seed", {
    set.seed(42)
    n <- 15
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    r1 <- null_moran(data, distmat, n_perm = 5L, seed = 123)
    r2 <- null_moran(data, distmat, n_perm = 5L, seed = 123)
    expect_equal(r1$nulls, r2$nulls, tolerance = 1e-10)
  })

  it("preserves mean of data", {
    set.seed(42)
    n <- 20
    data <- rnorm(n, mean = 5)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- null_moran(data, distmat, n_perm = 10L, seed = 1)
    null_means <- colMeans(result$nulls)
    expect_equal(null_means, rep(mean(data), 10), tolerance = 1e-8)
  })

  it("errors for dimension mismatch", {
    expect_error(
      null_moran(1:5, diag(3)),
      "5 x 5"
    )
  })
})

#' Baum spin test null model
#'
#' Spin-based null model with maximum-overlap parcel reassignment. After
#' rotating vertex coordinates, each original parcel is assigned the value
#' of the rotated parcel with the most vertex overlap.
#'
#' @template null-params
#' @param coords List with `$lh` and `$rh` matrices of spherical coordinates
#'   (n x 3 each).
#' @param parcellation Integer vector of parcel labels for all vertices.
#'   `0` and `NA` indicate medial wall.
#' @param rotation Rotation generation method: `"euler"` (ZYZ Euler angles,
#'   default, matches neuromaps Python) or `"rodrigues"` (Rodrigues axis-angle
#'   formula).
#'
#' @return A [null_distribution] object.
#'
#' @references
#' Baum GL et al. (2020) PNAS 117:21854-21861.
#' doi:10.1073/pnas.2005518117
#'
#' @examples
#' coords <- list(lh = matrix(rnorm(30), 10, 3), rh = matrix(rnorm(30), 10, 3))
#' parcellation <- c(rep(1L, 5), rep(2L, 5), rep(3L, 5), rep(4L, 5))
#' data <- c(1.0, 2.0, 3.0, 4.0)
#' nd <- null_baum(data, coords, parcellation, n_perm = 10L, seed = 1L)
#' @export
null_baum <- function(data, coords, parcellation, n_perm = 1000L, seed = NULL,
                      rotation = c("euler", "rodrigues")) {
  validate_data(data)
  validate_coords(coords)
  rotation <- match.arg(rotation)
  n_lh <- nrow(coords$lh)
  n_rh <- nrow(coords$rh)
  n <- n_lh + n_rh
  validate_parcellation(parcellation, n)

  valid <- !is.na(parcellation) & parcellation != 0
  ulabels <- sort(unique(parcellation[valid]))
  n_parcels <- length(ulabels)
  if (length(data) != n_parcels) {
    cli::cli_abort(paste(
      "Length of {.arg data} ({length(data)})",
      "must match number of parcels ({n_parcels})."
    ))
  }
  names(data) <- ulabels

  rotated <- rotate_coords(
    coords$lh, coords$rh, n_perm, seed,
    rotation = rotation
  )
  nulls <- matrix(0, nrow = n_parcels, ncol = n_perm)

  for (i in cli::cli_progress_along(seq_len(n_perm), "Generating baum nulls")) {
    cost_lh <- compute_cost_matrix(coords$lh, rotated$lh[, , i])
    cost_rh <- compute_cost_matrix(coords$rh, rotated$rh[, , i])
    assign_lh <- apply(cost_lh, 1, which.min)
    assign_rh <- apply(cost_rh, 1, which.min)

    rotated_labels <- integer(n)
    rotated_labels[seq_len(n_lh)] <- parcellation[assign_lh]
    rotated_labels[n_lh + seq_len(n_rh)] <- parcellation[n_lh + assign_rh]

    for (j in seq_along(ulabels)) {
      orig_mask <- which(parcellation == ulabels[j])
      rot_at_orig <- rotated_labels[orig_mask]
      rot_at_orig <- rot_at_orig[rot_at_orig != 0 & !is.na(rot_at_orig)]
      if (length(rot_at_orig) > 0) {
        tbl <- table(rot_at_orig)
        best_label <- names(tbl)[which.max(tbl)]
        nulls[j, i] <- data[best_label]
      } else {
        nulls[j, i] <- NA_real_
      }
    }
  }

  new_null_distribution(nulls, "baum", data, list(n_perm = n_perm))
}

describe("null_baum", {
  it("returns null_distribution with correct dims", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    n <- n_lh + n_rh
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    parcellation <- rep(1:4, each = 5)
    data <- c(10, 20, 30, 40)

    result <- null_baum(data, coords, parcellation, n_perm = 5L, seed = 1)
    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "baum")
    expect_equal(nrow(result$nulls), 4)
    expect_equal(ncol(result$nulls), 5)
  })

  it("errors for wrong data length", {
    n_lh <- 5
    n_rh <- 5
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    parcellation <- rep(1:2, each = 5)
    expect_error(
      null_baum(1:5, coords, parcellation, n_perm = 2L),
      "must match"
    )
  })

  it("null values are drawn from original data", {
    set.seed(42)
    n_lh <- 10
    n_rh <- 10
    coords <- list(
      lh = matrix(rnorm(n_lh * 3), ncol = 3),
      rh = matrix(rnorm(n_rh * 3), ncol = 3)
    )
    parcellation <- rep(1:4, each = 5)
    data <- c(10, 20, 30, 40)

    result <- null_baum(data, coords, parcellation, n_perm = 5L, seed = 1)
    non_na <- result$nulls[!is.na(result$nulls)]
    expect_true(all(non_na %in% data))
  })
})

#' Generate null distributions for brain map data
#'
#' Dispatches to the appropriate null model method for generating
#' spatially-constrained surrogate brain maps.
#'
#' @template null-params
#' @param method Character string specifying the null model method.
#' @param distmat Distance matrix (required for `"burt2020"`, `"burt2018"`,
#'   and `"moran"`).
#' @param coords List with `$lh` and `$rh` coordinate matrices
#'   (required for spin methods).
#' @param parcellation Integer vector of parcel labels (required for `"baum"`
#'   and `"cornblath"`).
#' @param ... Additional arguments passed to the specific null method
#'   (e.g. `rotation` for spin methods, `kernel` for moran/burt2020).
#'
#' @return A [null_distribution] object.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' data <- rnorm(100)
#' distmat <- as.matrix(dist(seq_len(100)))
#' nd <- generate_nulls(data, method = "moran", distmat = distmat, n_perm = 10L)
#' @export
generate_nulls <- function(data,
                           method = c(
                             "burt2020", "moran", "spin_vasa", "spin_hungarian",
                             "alexander_bloch", "baum", "cornblath", "burt2018"
                           ),
                           n_perm = 1000L,
                           distmat = NULL,
                           coords = NULL,
                           parcellation = NULL,
                           seed = NULL,
                           ...) {
  method <- match.arg(method)
  validate_data(data)
  n_perm <- as.integer(n_perm)
  if (n_perm < 1L) {
    cli::cli_abort("{.arg n_perm} must be >= 1.")
  }

  switch(method,
    burt2020 = {
      if (is.null(distmat)) {
        cli::cli_abort("{.arg distmat} is required for method {.val burt2020}.")
      }
      null_burt2020(data, distmat, n_perm = n_perm, seed = seed, ...)
    },
    moran = {
      if (is.null(distmat)) {
        cli::cli_abort("{.arg distmat} is required for method {.val moran}.")
      }
      null_moran(data, distmat, n_perm = n_perm, seed = seed, ...)
    },
    spin_vasa = {
      if (is.null(coords)) {
        cli::cli_abort("{.arg coords} is required for method {.val spin_vasa}.")
      }
      null_spin_vasa(data, coords, n_perm = n_perm, seed = seed, ...)
    },
    spin_hungarian = {
      if (is.null(coords)) {
        cli::cli_abort(
          "{.arg coords} is required for method {.val spin_hungarian}."
        )
      }
      null_spin_hungarian(
        data, coords, n_perm = n_perm, seed = seed, ...
      )
    },
    alexander_bloch = {
      if (is.null(coords)) {
        cli::cli_abort(
          "{.arg coords} is required for method {.val alexander_bloch}."
        )
      }
      null_alexander_bloch(
        data, coords, n_perm = n_perm, seed = seed, ...
      )
    },
    baum = {
      if (is.null(coords)) {
        cli::cli_abort("{.arg coords} is required for method {.val baum}.")
      }
      if (is.null(parcellation)) {
        cli::cli_abort(
          "{.arg parcellation} is required for method {.val baum}."
        )
      }
      null_baum(
        data, coords, parcellation,
        n_perm = n_perm, seed = seed, ...
      )
    },
    cornblath = {
      if (is.null(coords)) {
        cli::cli_abort("{.arg coords} is required for method {.val cornblath}.")
      }
      if (is.null(parcellation)) {
        cli::cli_abort(
          "{.arg parcellation} is required for method {.val cornblath}."
        )
      }
      null_cornblath(
        data, coords, parcellation,
        n_perm = n_perm, seed = seed, ...
      )
    },
    burt2018 = {
      if (is.null(distmat)) {
        cli::cli_abort("{.arg distmat} is required for method {.val burt2018}.")
      }
      null_burt2018(data, distmat, n_perm = n_perm, seed = seed)
    }
  )
}

#' Permutation test for any metric between brain maps
#'
#' Computes a user-specified metric between two vectors and tests significance
#' using either spatially-constrained null surrogates or simple random
#' permutation.
#'
#' @param x,y Numeric vectors.
#' @param metric_func Function taking `(x, y)` and returning a scalar.
#' @param n_perm Integer number of permutations.
#' @param seed Optional integer seed for reproducibility.
#' @param null_method Optional null model method passed to [generate_nulls()].
#'   If `NULL`, uses simple random permutation.
#' @param distmat Distance matrix (passed to [generate_nulls()] if needed).
#' @param coords Coordinate list (passed to [generate_nulls()] if needed).
#' @param parcellation Integer vector (passed to [generate_nulls()] if needed).
#' @param ... Additional arguments passed to [generate_nulls()].
#'
#' @return List with `$observed`, `$null_values`, `$p_value`, and `$n_perm`.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' x <- rnorm(100)
#' y <- x + rnorm(100)
#' result <- permtest_metric(x, y, n_perm = 99L, seed = 1L)
#' result$observed
#' result$p_value
#' @export
permtest_metric <- function(x,
                            y,
                            metric_func = stats::cor,
                            n_perm = 1000L,
                            seed = NULL,
                            null_method = NULL,
                            distmat = NULL,
                            coords = NULL,
                            parcellation = NULL,
                            ...) {
  validate_data(x, "x")
  validate_data(y, "y")
  if (length(x) != length(y)) {
    cli::cli_abort(paste(
      "{.arg x} ({length(x)}) and {.arg y} ({length(y)})",
      "must have the same length."
    ))
  }

  observed <- metric_func(x, y)
  n_perm <- as.integer(n_perm)

  if (!is.null(null_method)) {
    null_dist <- generate_nulls(
      x,
      method = null_method,
      n_perm = n_perm,
      distmat = distmat,
      coords = coords,
      parcellation = parcellation,
      seed = seed,
      ...
    )
    null_values <- apply(null_dist$nulls, 2, function(x_perm) {
      metric_func(x_perm, y)
    })
  } else {
    if (!is.null(seed)) withr::local_seed(seed)
    null_values <- vapply(
      cli::cli_progress_along(seq_len(n_perm), "Permuting"),
      function(i) metric_func(sample(x), y),
      numeric(1)
    )
  }

  p_value <- (sum(abs(null_values) >= abs(observed)) + 1) / (n_perm + 1)

  list(
    observed = observed,
    null_values = null_values,
    p_value = p_value,
    n_perm = n_perm
  )
}

#' @noRd
#' @keywords internal
osf_download_url <- function(osf_project, osf_file_id) {
  paste0(
    "https://files.osf.io/v1/resources/",
    osf_project,
    "/providers/osfstorage/",
    osf_file_id
  )
}

#' @noRd
#' @keywords internal
download_neuromaps_file <- function(
  url,
  destfile,
  checksum = NULL,
  overwrite = FALSE,
  verbose = TRUE
) {
  if (file.exists(destfile) && !overwrite) {
    if (validate_checksum(destfile, checksum)) {
      if (verbose) {
        cli::cli_alert_info(
          "Using cached {.file {basename(destfile)}}"
        )
      }
      return(invisible(destfile))
    }
    if (verbose) {
      cli::cli_alert_warning(
        "Checksum mismatch, re-downloading {.file {basename(destfile)}}"
      )
    }
  }

  dir.create(dirname(destfile), recursive = TRUE, showWarnings = FALSE)

  if (verbose) {
    cli::cli_alert("Downloading {.file {basename(destfile)}}")
  }
  perform_download(url, destfile)

  if (!validate_checksum(destfile, checksum)) {
    unlink(destfile)
    cli::cli_abort(
      "Checksum validation failed for {.file {basename(destfile)}}"
    )
  }

  if (verbose) cli::cli_alert_success("Saved {.file {basename(destfile)}}")
  invisible(destfile)
}

#' @noRd
#' @keywords internal
perform_download <- function(url, destfile) {
  resp <- httr2::request(url) |>
    httr2::req_timeout(seconds = 120) |>
    httr2::req_retry(max_tries = 3) |>
    httr2::req_perform()
  writeBin(httr2::resp_body_raw(resp), destfile)
}

describe("osf_download_url", {
  it("constructs correct URL", {
    url <- osf_download_url("4mw3a", "abc123")
    expect_equal(
      url,
      "https://files.osf.io/v1/resources/4mw3a/providers/osfstorage/abc123"
    )
  })
})


describe("download_neuromaps_file", {
  it("skips download when cached file has valid checksum", {
    tmp <- withr::local_tempfile()
    writeLines("cached content", tmp)
    md5 <- unname(tools::md5sum(tmp))

    result <- download_neuromaps_file(
      url = "https://example.com/fake",
      destfile = tmp,
      checksum = md5,
      overwrite = FALSE,
      verbose = FALSE
    )
    expect_equal(result, tmp)
  })

  it("re-downloads when overwrite is TRUE", {
    tmp_dir <- withr::local_tempdir()
    destfile <- file.path(tmp_dir, "test.gii")
    writeLines("old content", destfile)

    download_called <- FALSE
    local_mocked_bindings(
      perform_download = function(url, destfile) {
        download_called <<- TRUE
        writeLines("new content", destfile)
      }
    )

    download_neuromaps_file(
      url = "https://example.com/fake",
      destfile = destfile,
      checksum = NULL,
      overwrite = TRUE,
      verbose = FALSE
    )
    expect_true(download_called)
  })

  it("errors on checksum mismatch after download", {
    tmp_dir <- withr::local_tempdir()
    destfile <- file.path(tmp_dir, "test.gii")

    local_mocked_bindings(
      perform_download = function(url, destfile) {
        writeLines("bad content", destfile)
      }
    )

    expect_error(
      download_neuromaps_file(
        url = "https://example.com/fake",
        destfile = destfile,
        checksum = "0000000000000000",
        overwrite = FALSE,
        verbose = FALSE
      ),
      "Checksum validation failed"
    )
    expect_false(file.exists(destfile))
  })

  it("re-downloads when checksum mismatches cached file", {
    tmp_dir <- withr::local_tempdir()
    destfile <- file.path(tmp_dir, "test.gii")
    writeLines("stale content", destfile)
    fresh_file <- withr::local_tempfile()
    writeLines("fresh content", fresh_file)
    fresh_md5 <- unname(tools::md5sum(fresh_file))

    local_mocked_bindings(
      perform_download = function(url, destfile) {
        writeLines("fresh content", destfile)
      }
    )

    expect_message(
      expect_message(
        expect_message(
          download_neuromaps_file(
            url = "https://example.com/fake",
            destfile = destfile,
            checksum = fresh_md5,
            overwrite = FALSE,
            verbose = TRUE
          ),
          "mismatch"
        ),
        "Downloading"
      ),
      "Saved"
    )
  })

  it("prints verbose messages on cache hit", {
    tmp <- withr::local_tempfile()
    writeLines("cached", tmp)
    md5 <- unname(tools::md5sum(tmp))

    expect_message(
      download_neuromaps_file(
        url = "https://example.com/fake",
        destfile = tmp,
        checksum = md5,
        verbose = TRUE
      ),
      "cached"
    )
  })

  it("prints verbose messages during download", {
    tmp_dir <- withr::local_tempdir()
    destfile <- file.path(tmp_dir, "test.gii")

    local_mocked_bindings(
      perform_download = function(url, destfile) {
        writeLines("content", destfile)
      }
    )

    expect_message(
      expect_message(
        download_neuromaps_file(
          url = "https://example.com/fake",
          destfile = destfile,
          checksum = NULL,
          verbose = TRUE
        ),
        "Downloading"
      ),
      "Saved"
    )
  })

  it("creates parent directories", {
    tmp_dir <- withr::local_tempdir()
    destfile <- file.path(tmp_dir, "a", "b", "test.gii")

    local_mocked_bindings(
      perform_download = function(url, destfile) {
        writeLines("content", destfile)
      }
    )

    download_neuromaps_file(
      url = "https://example.com/fake",
      destfile = destfile,
      checksum = NULL,
      overwrite = FALSE,
      verbose = FALSE
    )
    expect_true(file.exists(destfile))
  })
})

#' Variogram-matching null model
#'
#' Generates spatially-constrained surrogate brain maps by matching the
#' empirical variogram of the original data through smoothed random
#' permutations.
#'
#' @template null-params
#' @param distmat Distance matrix between parcels/vertices.
#' @param ns Integer, subsample size for variogram computation.
#' @param nh Integer, number of distance bins for variogram.
#' @param pv Numeric, percentile cutoff for maximum distance in variogram.
#' @param knn Integer, number of nearest neighbors for smoothing.
#' @param deltas Numeric vector of smoothing levels (fractions of `knn`).
#' @param kernel Smoothing kernel function.
#' @param resample Logical. If `FALSE` (default, matches brainsmash Python),
#'   variogram subsample indices are computed once and reused across all
#'   permutations. If `TRUE`, a fresh subsample is drawn each time.
#'
#' @return A [null_distribution] object.
#'
#' @references
#' Burt JB et al. (2020) NeuroImage 220:117038.
#' doi:10.1016/j.neuroimage.2020.117038
#'
#' @examples
#' data <- rnorm(50)
#' distmat <- as.matrix(dist(matrix(rnorm(100), 50, 2)))
#' nd <- null_burt2020(data, distmat, n_perm = 10L, seed = 1L)
#' @export
null_burt2020 <- function(data,
                          distmat,
                          n_perm = 1000L,
                          seed = NULL,
                          ns = 500L,
                          nh = 25L,
                          pv = 25,
                          knn = 1000L,
                          deltas = seq(0.1, 0.9, by = 0.1),
                          kernel = c("exponential", "gaussian", "uniform"),
                          resample = FALSE) {
  validate_data(data)
  n <- length(data)
  validate_distmat(distmat, n)
  kernel <- match.arg(kernel)
  knn <- min(as.integer(knn), n - 1L)

  if (!is.null(seed)) withr::local_seed(seed)

  fixed_idx <- NULL
  if (!resample && ns < n) {
    fixed_idx <- sample.int(n, ns)
  }

  target_variogram <- compute_variogram(
    data, distmat, nh = nh, pv = pv, ns = ns,
    idx = fixed_idx
  )
  nn <- compute_knn(distmat, knn)

  nulls <- matrix(0, nrow = n, ncol = n_perm)

  msg <- "Generating burt2020 nulls"
  for (i in cli::cli_progress_along(seq_len(n_perm), msg)) {
    permuted <- sample(data)
    best_sse <- Inf
    best_surrogate <- permuted

    for (delta in deltas) {
      k <- max(1L, round(delta * knn))
      smoothed <- smooth_surrogate(
        permuted, nn$indices, nn$distances, k, kernel
      )
      smoothed <- rank_match(smoothed, data)
      vg <- compute_variogram(
        smoothed, distmat, nh = nh, pv = pv, ns = ns,
        idx = fixed_idx
      )
      sse <- sum((vg$gamma - target_variogram$gamma)^2)

      if (sse < best_sse) {
        best_sse <- sse
        best_surrogate <- smoothed
      }
    }
    nulls[, i] <- best_surrogate
  }

  new_null_distribution(nulls, "burt2020", data, list(
    ns = ns, nh = nh, pv = pv, knn = knn,
    deltas = deltas, kernel = kernel
  ))
}

#' @noRd
#' @keywords internal
compute_variogram <- function(data, distmat, nh = 25L,
                              pv = 25, ns = 500L,
                              idx = NULL) {
  n <- length(data)
  if (is.null(idx) && ns < n) {
    idx <- sample.int(n, ns)
  }
  if (!is.null(idx)) {
    sub_data <- data[idx]
    sub_dist <- distmat[idx, idx]
  } else {
    sub_data <- data
    sub_dist <- distmat
  }

  upper_idx <- which(upper.tri(sub_dist), arr.ind = TRUE)
  dists <- sub_dist[upper_idx]
  diffs_sq <- 0.5 * (sub_data[upper_idx[, 1]] - sub_data[upper_idx[, 2]])^2

  max_dist <- stats::quantile(dists, pv / 100)
  keep <- dists <= max_dist
  dists <- dists[keep]
  diffs_sq <- diffs_sq[keep]

  breaks <- seq(0, max_dist, length.out = nh + 1L)
  bins <- findInterval(dists, breaks, all.inside = TRUE)

  gamma <- tapply(diffs_sq, bins, mean)
  bin_centers <- (breaks[-length(breaks)] + breaks[-1]) / 2
  present <- as.integer(names(gamma))

  data.frame(
    distance = bin_centers[present],
    gamma = as.numeric(gamma)
  )
}

#' @noRd
#' @keywords internal
smooth_surrogate <- function(permuted, nn_indices, nn_distances, k, kernel) {
  n <- length(permuted)
  k <- min(k, ncol(nn_indices))
  smoothed <- numeric(n)

  for (i in seq_len(n)) {
    idx <- nn_indices[i, seq_len(k)]
    d <- nn_distances[i, seq_len(k)]
    w <- switch(kernel,
      exponential = exp(-d / (max(d) + .Machine$double.eps)),
      gaussian = exp(-0.5 * (d / (max(d) + .Machine$double.eps))^2),
      uniform = rep(1, k)
    )
    smoothed[i] <- stats::weighted.mean(permuted[idx], w)
  }
  smoothed
}

#' @noRd
#' @keywords internal
rank_match <- function(surrogate, target) {
  target_sorted <- sort(target)
  target_sorted[rank(surrogate, ties.method = "first")]
}

describe("compute_variogram", {
  it("returns data frame with distance and gamma columns", {
    set.seed(42)
    n <- 30
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    vg <- compute_variogram(data, distmat, nh = 10, pv = 50)

    expect_s3_class(vg, "data.frame")
    expect_named(vg, c("distance", "gamma"))
    expect_true(all(vg$gamma >= 0))
    expect_true(all(vg$distance > 0))
  })

  it("subsamples when ns < n", {
    set.seed(42)
    n <- 50
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    vg <- compute_variogram(data, distmat, nh = 10, pv = 50, ns = 20)
    expect_s3_class(vg, "data.frame")
  })
})

describe("smooth_surrogate", {
  it("smooths permuted values toward neighbors", {
    set.seed(42)
    n <- 20
    coords <- matrix(rnorm(n * 3), ncol = 3)
    distmat <- as.matrix(dist(coords))
    nn <- compute_knn(distmat, 5)
    permuted <- rnorm(n)
    smoothed <- smooth_surrogate(
      permuted, nn$indices, nn$distances,
      k = 5, kernel = "uniform"
    )

    expect_length(smoothed, n)
    expect_true(stats::sd(smoothed) < stats::sd(permuted))
  })

  it("supports exponential kernel", {
    set.seed(42)
    n <- 20
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    nn <- compute_knn(distmat, 5)
    smoothed <- smooth_surrogate(
      rnorm(n), nn$indices, nn$distances,
      k = 3, kernel = "exponential"
    )
    expect_length(smoothed, n)
  })

  it("supports gaussian kernel", {
    set.seed(42)
    n <- 20
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    nn <- compute_knn(distmat, 5)
    smoothed <- smooth_surrogate(
      rnorm(n), nn$indices, nn$distances,
      k = 3, kernel = "gaussian"
    )
    expect_length(smoothed, n)
  })
})

describe("rank_match", {
  it("preserves the rank distribution of target", {
    target <- c(10, 20, 30, 40, 50)
    surrogate <- c(5, 3, 1, 2, 4)
    matched <- rank_match(surrogate, target)
    expect_equal(sort(matched), sort(target))
  })
})

describe("null_burt2020", {
  it("generates null distribution of correct dimensions", {
    set.seed(42)
    n <- 25
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- null_burt2020(
      data, distmat,
      n_perm = 3L, seed = 1,
      ns = 25, nh = 5, knn = 10,
      deltas = c(0.3, 0.6)
    )

    expect_s3_class(result, "null_distribution")
    expect_equal(result$method, "burt2020")
    expect_equal(result$n_perm, 3)
    expect_equal(result$n, 25)
  })

  it("surrogates have same values as original (rank-matched)", {
    set.seed(42)
    n <- 20
    data <- sort(rnorm(n))
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    result <- null_burt2020(
      data, distmat,
      n_perm = 2L, seed = 1,
      ns = 20, nh = 5, knn = 10,
      deltas = c(0.5)
    )

    for (i in seq_len(2)) {
      expect_equal(sort(result$nulls[, i]), sort(data))
    }
  })

  it("is reproducible with seed", {
    set.seed(42)
    n <- 20
    data <- rnorm(n)
    distmat <- as.matrix(dist(matrix(rnorm(n * 3), ncol = 3)))
    r1 <- null_burt2020(
      data, distmat,
      n_perm = 2L, seed = 99,
      ns = 20, nh = 5, knn = 10, deltas = c(0.5)
    )
    r2 <- null_burt2020(
      data, distmat,
      n_perm = 2L, seed = 99,
      ns = 20, nh = 5, knn = 10, deltas = c(0.5)
    )
    expect_equal(r1$nulls, r2$nulls, tolerance = 1e-10)
  })

  it("errors for dimension mismatch", {
    expect_error(
      null_burt2020(1:5, diag(3)),
      "5 x 5"
    )
  })
})

#' Resample brain maps for comparison
#'
#' Aligns two brain maps into the same coordinate space and density before
#' comparison. Supports multiple strategies for choosing the target space.
#'
#' @param src Character, file path to the source GIFTI.
#' @param trg Character, file path to the target GIFTI.
#' @param src_space Source coordinate space (`"fsaverage"` or `"fsLR"`).
#' @param trg_space Target coordinate space (`"fsaverage"` or `"fsLR"`).
#' @param strategy Resampling strategy. One of `"downsample_only"`,
#'   `"transform_to_src"`, `"transform_to_trg"`, or `"transform_to_alt"`.
#' @param alt_space Alternative space for `"transform_to_alt"` strategy.
#' @param alt_density Alternative density for `"transform_to_alt"` strategy.
#' @param hemisphere Which hemispheres: `"left"`, `"right"`, or both.
#' @param area_surf_current Path to a current-resolution area-correction
#'   surface (e.g. midthickness). Passed to [transform_to_space()].
#' @param area_surf_new Path to a target-resolution area-correction surface.
#'   Passed to [transform_to_space()].
#' @param wb_path Path to `wb_command` executable.
#' @param verbose Logical, print progress messages.
#'
#' @return List with `$src` and `$trg` file paths in the aligned space.
#'
#' @references
#' Markello RD et al. (2022) Nature Methods 19:1472-1480.
#' doi:10.1038/s41592-022-01625-w
#'
#' @examples
#' \dontrun{
#' resample_images("src.func.gii", "trg.func.gii",
#'   src_space = "fsaverage", trg_space = "fsaverage"
#' )
#' }
#' @export
resample_images <- function(src,
                            trg,
                            src_space = c("fsaverage", "fsLR"),
                            trg_space = c("fsaverage", "fsLR"),
                            strategy = c(
                              "downsample_only",
                              "transform_to_src",
                              "transform_to_trg",
                              "transform_to_alt"
                            ),
                            alt_space = NULL,
                            alt_density = NULL,
                            hemisphere = c("left", "right"),
                            area_surf_current = NULL,
                            area_surf_new = NULL,
                            wb_path = NULL,
                            verbose = TRUE) {
  src_space <- match.arg(src_space)
  trg_space <- match.arg(trg_space)
  strategy <- match.arg(strategy)
  hemisphere <- match.arg(hemisphere, several.ok = TRUE)

  if (!file.exists(src)) cli::cli_abort("Source file not found: {.file {src}}")
  if (!file.exists(trg)) cli::cli_abort("Target file not found: {.file {trg}}")

  alt_missing <- is.null(alt_space) || is.null(alt_density)
  if (strategy == "transform_to_alt" && alt_missing) {
    cli::cli_abort(paste(
      "{.arg alt_space} and {.arg alt_density} are required",
      "for {.val transform_to_alt}."
    ))
  }
  if (strategy == "downsample_only" && src_space != trg_space) {
    cli::cli_abort(paste(
      "{.val downsample_only} requires maps in the same space",
      "({src_space} vs {trg_space})."
    ))
  }

  src_density <- get_gifti_density(src)
  trg_density <- get_gifti_density(trg)

  switch(strategy,
    downsample_only = {
      src_n <- density_to_n(src_density)
      trg_n <- density_to_n(trg_density)
      if (src_n > trg_n) {
        src_out <- transform_to_space(
          src, target_space = src_space, target_density = trg_density,
          hemisphere = hemisphere, area_surf_current = area_surf_current,
          area_surf_new = area_surf_new, wb_path = wb_path, verbose = verbose
        )
        list(src = src_out, trg = trg)
      } else if (trg_n > src_n) {
        trg_out <- transform_to_space(
          trg, target_space = trg_space, target_density = src_density,
          hemisphere = hemisphere, area_surf_current = area_surf_current,
          area_surf_new = area_surf_new, wb_path = wb_path, verbose = verbose
        )
        list(src = src, trg = trg_out)
      } else {
        list(src = src, trg = trg)
      }
    },
    transform_to_src = {
      trg_out <- transform_to_space(
        trg, target_space = src_space, target_density = src_density,
        hemisphere = hemisphere, area_surf_current = area_surf_current,
        area_surf_new = area_surf_new, wb_path = wb_path, verbose = verbose
      )
      list(src = src, trg = trg_out)
    },
    transform_to_trg = {
      src_out <- transform_to_space(
        src, target_space = trg_space, target_density = trg_density,
        hemisphere = hemisphere, area_surf_current = area_surf_current,
        area_surf_new = area_surf_new, wb_path = wb_path, verbose = verbose
      )
      list(src = src_out, trg = trg)
    },
    transform_to_alt = {
      src_out <- transform_to_space(
        src, target_space = alt_space, target_density = alt_density,
        hemisphere = hemisphere, area_surf_current = area_surf_current,
        area_surf_new = area_surf_new, wb_path = wb_path, verbose = verbose
      )
      trg_out <- transform_to_space(
        trg, target_space = alt_space, target_density = alt_density,
        hemisphere = hemisphere, area_surf_current = area_surf_current,
        area_surf_new = area_surf_new, wb_path = wb_path, verbose = verbose
      )
      list(src = src_out, trg = trg_out)
    }
  )
}

#' @noRd
#' @keywords internal
get_gifti_density <- function(path) {
  gii <- gifti::read_gifti(path)
  n_vert <- length(gii$data[[1]])
  density_map <- c(
    "163842" = "164k",
    "40962" = "41k",
    "32492" = "32k",
    "10242" = "10k"
  )
  key <- as.character(n_vert)
  if (key %in% names(density_map)) {
    density_map[[key]]
  } else {
    cli::cli_abort("Unknown vertex count {n_vert}, cannot determine density.")
  }
}

#' @noRd
#' @keywords internal
density_to_n <- function(density) {
  n_map <- c("164k" = 163842L, "41k" = 40962L, "32k" = 32492L, "10k" = 10242L)
  result <- unname(n_map[density])
  if (is.na(result)) {
    cli::cli_abort(
      "Unknown density {.val {density}}. Expected one of {.or {names(n_map)}}."
    )
  }
  result
}

describe("get_gifti_density", {
  it("detects density from a real GIFTI", {
    skip_on_cran()
    skip_if_offline()

    cache <- withr::local_tempdir()
    paths <- fetch_neuromaps_annotation(
      "abagen", "genepc1", "fsaverage",
      density = "10k",
      data_dir = cache,
      verbose = FALSE
    )

    expect_equal(get_gifti_density(paths[1]), "10k")
  })
})

describe("density_to_n", {
  it("maps known densities", {
    expect_equal(density_to_n("32k"), 32492L)
    expect_equal(density_to_n("164k"), 163842L)
    expect_equal(density_to_n("10k"), 10242L)
    expect_equal(density_to_n("41k"), 40962L)
  })
})

describe("resample_images", {
  it("errors for missing source file", {
    expect_error(
      resample_images(
        "nonexistent.gii", "also_missing.gii",
        src_space = "fsaverage",
        trg_space = "fsaverage"
      ),
      "not found"
    )
  })

  it("requires alt_space and alt_density for transform_to_alt", {
    src <- withr::local_tempfile(fileext = ".gii")
    trg <- withr::local_tempfile(fileext = ".gii")
    file.create(src)
    file.create(trg)
    expect_error(
      resample_images(
        src, trg,
        src_space = "fsaverage",
        trg_space = "fsLR",
        strategy = "transform_to_alt"
      ),
      "alt_space"
    )
  })
})

---
title: "Working with Parcellated Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Working with Parcellated Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

Most brain imaging analyses eventually need to move between two
resolutions: vertex-level (one value per surface point, typically
32k--164k per hemisphere) and parcel-level (one value per brain region,
typically 50--400). A cortical thickness map is vertex-level. A table of
regional volumes from the Desikan-Killiany atlas is parcel-level.
Comparing the two requires converting one to match the other.

neuromapr provides tools for this conversion---aggregating vertices
into parcels, mapping parcel values back to vertices, and computing
parcel centroids on the cortical surface.

```{r setup}
library(neuromapr)
```

## Vertices to parcels

Suppose you have a vertex-level brain map and a parcellation that
assigns each vertex to a region. The parcellation is an integer vector
where each vertex gets a label, with `0` or `NA` marking the medial
wall (vertices that belong to no region).

```{r vertices-to-parcels}
set.seed(42)
n_vertices <- 1000
vertex_data <- rnorm(n_vertices)

labels <- sample(
  c(0, 1:10), n_vertices,
  replace = TRUE, prob = c(0.1, rep(0.09, 10))
)

parcel_values <- vertices_to_parcels(vertex_data, labels)
parcel_values
```

Each parcel gets the mean of its constituent vertices. Medial wall
vertices (label 0) are excluded automatically.

The summary function is configurable:

```{r custom-summary}
parcel_sd <- vertices_to_parcels(
  vertex_data, labels, summary_func = sd
)
parcel_sd
```

## Parcels back to vertices

The reverse operation paints parcel-level values back onto the vertex
mesh. Every vertex gets its parcel's value; medial wall vertices get a
fill value.

```{r parcels-to-vertices}
vertex_filled <- parcels_to_vertices(parcel_values, labels)
head(vertex_filled, 20)
```

Vertices with label 0 receive `NA` by default. You can change the fill:

```{r custom-fill}
vertex_filled_zero <- parcels_to_vertices(
  parcel_values, labels, fill = 0
)
sum(vertex_filled_zero == 0)
```

## The high-level wrappers

`parcellate()` and `unparcellate()` add file-reading convenience on top
of the low-level functions. If your data lives on disk as GIFTI files,
pass paths directly:

```r
parcel_vals <- parcellate(
  "cortical_thickness.func.gii",
  "aparc.label.gii"
)
```

The function reads the data file, reads the parcellation labels,
validates that they match in length, and calls `vertices_to_parcels()`
underneath.

When working with vectors already in memory, the wrappers still
work---they skip the file-reading step:

```{r parcellate-vectors}
parcellated <- parcellate(vertex_data, labels)
unparcellated <- unparcellate(parcellated, labels)

all.equal(
  parcels_to_vertices(parcel_values, labels),
  unparcellated
)
```

The roundtrip is not lossless---parcellation is lossy aggregation. But
`unparcellate(parcellate(x))` gives you parcel means broadcast back to
vertex positions, which is the expected behaviour.

## Parcel centroids

Some analyses need to know where each parcel lives in 3D space---for
example, to compute a parcel-level distance matrix.
`get_parcel_centroids()` offers three methods for finding the
representative point of each parcel.

```{r centroids-setup}
vertices <- matrix(rnorm(n_vertices * 3), ncol = 3)
```

### Average centroid

Take the mean x, y, z coordinates of all vertices in the parcel. Fast,
but the resulting point may not lie on the actual cortical surface.

```{r centroid-average}
centroids_avg <- get_parcel_centroids(
  vertices, labels, method = "average"
)
head(centroids_avg)
```

### Surface centroid

Find the vertex closest to the average centroid. The result is always
an actual vertex on the surface, which matters when centroids must
respect the cortical geometry.

```{r centroid-surface}
centroids_surf <- get_parcel_centroids(
  vertices, labels, method = "surface"
)
head(centroids_surf)
```

### Geodesic centroid

The most principled approach: find the vertex that minimizes the sum
of geodesic (shortest-path-on-surface) distances to all other vertices
in the parcel. This requires the face matrix of the surface mesh and
the **igraph** package.

```r
centroids_geo <- get_parcel_centroids(
  vertices, labels,
  method = "geodesic",
  faces = faces
)
```

Geodesic centroids are computationally expensive---they compute
pairwise shortest paths within each parcel. For large parcellations on
high-resolution meshes, the surface centroid is a practical
approximation.

## Building a parcel-level distance matrix

A common use case: you have parcellated data and need a distance matrix
for null model generation. Parcel centroids get you there.

```{r parcel-distmat}
parcel_distmat <- as.matrix(dist(centroids_avg))
dim(parcel_distmat)
```

This distance matrix feeds directly into `generate_nulls()`:

```{r parcel-nulls}
parcel_map <- rnorm(nrow(centroids_avg))
nulls <- generate_nulls(
  parcel_map,
  method = "moran",
  distmat = parcel_distmat,
  n_perm = 100L,
  seed = 1
)
nulls
```

## Parcellation and null models together

The parcel spin methods (`baum` and `cornblath`) combine parcellation
with spin-based null generation. They need both spherical coordinates
and the parcellation vector:

```{r parcel-spin}
n_lh <- 680
n_rh <- 680
sphere_coords <- list(
  lh = matrix(rnorm(n_lh * 3), ncol = 3),
  rh = matrix(rnorm(n_rh * 3), ncol = 3)
)
parcellation <- rep(1:68, each = 20)
parcel_data <- rnorm(68)

nulls_baum <- null_baum(
  parcel_data, sphere_coords, parcellation,
  n_perm = 50L, seed = 1
)
nulls_baum
```

The rotation happens at vertex resolution, but the output is
parcel-level: one surrogate value per parcel per permutation. This is
the right approach when your analysis variable is measured at the parcel
level but you want the null model to respect vertex-level spatial
structure.

## Practical considerations

**Parcellation size matters for null models.** With 30--50 parcels,
distance-based methods like `burt2020` and `moran` work well. With
hundreds of parcels (e.g., Schaefer 400), spin-based methods become
more attractive because they leverage the full vertex-level spatial
structure.

**Medial wall handling.** All parcellation functions treat label 0 and
`NA` as medial wall. The `cornblath` null model handles medial wall
vertices gracefully during spin-based reassignment. If your
parcellation has substantial medial wall coverage, prefer `cornblath`
over `baum`.

**The summary function determines what you are testing.** Using `mean`
(the default) tests whether the average value in each region relates to
another variable. Using `sd` would test whether within-region
variability relates to it. Choose deliberately.

## References

Markello RD, Hansen JY, Liu Z-Q, et al. (2022). neuromaps: structural
and functional interpretation of brain maps. *Nature Methods*, 19,
1472--1480. doi:10.1038/s41592-022-01625-w

---
title: "Rotation Methods for Spin Tests"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Rotation Methods for Spin Tests}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| label: setup
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

Every spin test starts the same way: pick a random rotation, apply it to the
cortical sphere, reassign values based on what ended up where.
The rotation is the engine of the whole procedure.
How you generate that rotation turns out to be a choice worth understanding.

neuromapr offers two methods for generating random 3D rotations.
The default (`"euler"`) matches the Python neuromaps implementation.
The alternative (`"rodrigues"`) takes a different mathematical path to the same
goal.
Both work.
They differ in geometry, numerical behaviour, and how they sample the space of all possible rotations.

```{r}
#| label: load
library(neuromapr)
```

## What a random rotation needs to do

A spin test rotates vertex coordinates on the cortical sphere, then checks
whether the spatial relationship between two brain maps survives the scramble.
Do this a thousand times and you have a null distribution.

For the null distribution to be valid, the rotations need two properties:

- _Coverage_: they should explore the full space of 3D rotations, not cluster
  around certain orientations.
- _Axis uniformity_: the rotation axis should be equally likely to point in any
  direction.

The space of all 3D rotations is called SO(3).
It is a three-dimensional manifold, and sampling from it uniformly is not as
simple as drawing three random numbers.

## Euler ZYZ rotations

The default method decomposes a rotation into three sequential turns around
fixed axes: first Z, then Y, then Z again.

The recipe:

- $\alpha$ drawn uniformly from $[0, 2\pi)$
- $\cos(\beta)$ drawn uniformly from $[-1, 1]$
- $\gamma$ drawn uniformly from $[0, 2\pi)$

The trick is in the second step.
Drawing $\cos(\beta)$ uniformly rather than $\beta$ itself corrects for the
non-uniform volume element of the Euler angle parameterisation.
Without this correction, rotations near the poles of the sphere would be
oversampled.

This produces rotations that are _Haar-uniform_ on SO(3) --- the mathematical
gold standard for "every orientation is equally likely."
It is the same method used in Python's neuromaps, so results are directly
comparable when using the same random seed.

```{r}
#| label: euler-demo
set.seed(1)
neuromapr:::random_rotation_euler()
```

The resulting 3x3 matrix is orthogonal with determinant 1 --- a proper rotation.

## Rodrigues axis-angle rotations

The alternative method parameterises a rotation by what it physically does:
spin by some angle around some axis.

The recipe:

- Draw a random axis $\mathbf{u}$ uniformly on the unit sphere (by normalising
  a 3D standard normal vector).
- Draw angle $\theta$ uniformly from $[0, 2\pi)$.
- Build the rotation matrix using the Rodrigues formula.

```{r}
#| label: rodrigues-demo
set.seed(1)
neuromapr:::random_rotation_rodrigues()
```

A different matrix from the same seed, because the two methods consume random
numbers differently.

## Where Rodrigues has the edge

### Geometric transparency

The Euler method describes a rotation as a composition of three turns around
coordinate axes.
The result is correct but opaque --- given the triple $(\alpha, \beta, \gamma)$,
it is not obvious what the rotation _does_ to the sphere.

The Rodrigues method describes a rotation as "turn by $\theta$ around axis
$\mathbf{u}$."
That maps directly to the physical intuition of taking a brain hemisphere and
spinning it.
Each null permutation has a single, interpretable axis and angle.

### Axis uniformity is explicit

For spin tests, the rotation axis is arguably the most important degree of
freedom.
It determines which vertex ends up near which other vertex after rotation ---
the spatial scrambling that the entire null hypothesis rests on.

The Rodrigues method achieves uniform axis sampling directly.
Drawing from $\mathcal{N}(0, I_3)$ and normalising is one of the most
well-studied methods for uniform points on $S^2$, and R's `rnorm()` is heavily
tested.

The Euler method achieves axis uniformity indirectly, as a mathematical
consequence of the ZYZ decomposition.
Both are correct, but the Rodrigues approach makes the property visible in the
code.

### Numerical stability

The Euler method computes $\beta = \arccos(u)$ where $u$ is uniform on
$[-1, 1]$.
When $u$ lands near $\pm 1$ --- which happens regularly --- `acos()` loses
precision because its derivative diverges at the boundaries.
The resulting $\beta$ values near $0$ and $\pi$ carry more floating-point noise
than values in the middle of the range.

The Rodrigues method avoids inverse trigonometric functions entirely in the
sampling step.
The only trig calls are `cos(theta)` and `sin(theta)`, which are
well-conditioned for all $\theta$.

### Idiomatic R

The normalised-Gaussian-on-sphere technique is the standard R idiom for uniform
spherical sampling.
You will find it across _pracma_, _movMF_, and other packages that need random
directions.
The Rodrigues method builds on this established pattern rather than introducing
the less common ZYZ convention.

## Where Euler has the edge

The Euler method produces rotations that are _exactly_ Haar-uniform on SO(3).
The Rodrigues method, as implemented, does not.

The Haar measure in axis-angle coordinates has density proportional to
$(1 - \cos\theta)$ --- it weights angles near $\pi$ more heavily than angles
near $0$.
Sampling $\theta$ uniformly slightly overweights near-identity rotations
(small $\theta$) and underweights half-turn rotations (near $\pi$).

In concrete terms: if you generated a million Rodrigues rotations and looked at
the distribution of rotation angles, you would see a uniform distribution on
$[0, \pi]$.
A million Euler rotations would show the correct Haar density
$(1 - \cos\theta)/\pi$, which rises from zero at $\theta = 0$ to a peak at
$\theta = \pi$.

## How much does the difference matter?

For spin tests on brain maps, not much.

```{r}
#| label: sim-setup
set.seed(42)
n_lh <- 50
n_rh <- 50
coords <- list(
  lh = matrix(rnorm(n_lh * 3), ncol = 3),
  rh = matrix(rnorm(n_rh * 3), ncol = 3)
)
coords$lh <- t(apply(coords$lh, 1, function(x) x / sqrt(sum(x^2))))
coords$rh <- t(apply(coords$rh, 1, function(x) x / sqrt(sum(x^2))))

data_x <- rnorm(n_lh + n_rh)
data_y <- 0.3 * data_x + rnorm(n_lh + n_rh, sd = 0.9)
```

Run the same spin test with both rotation methods:

```{r}
#| label: compare-methods
nulls_euler <- null_alexander_bloch(
  data_x,
  coords,
  n_perm = 500L,
  seed = 1,
  rotation = "euler"
)

nulls_rodrigues <- null_alexander_bloch(
  data_x,
  coords,
  n_perm = 500L,
  seed = 1,
  rotation = "rodrigues"
)
```

Compare the null distributions:

```{r}
#| label: compare-plot
#| fig-alt: "Two overlapping density curves showing the null correlation
#|   distributions from Euler and Rodrigues rotation methods. The
#|   distributions are nearly identical in shape and location."
#| fig-cap: "Null correlation distributions from 500 spin permutations using
#|   Euler (blue) and Rodrigues (orange) rotations. The two methods produce
#|   effectively indistinguishable null distributions."
null_cors_euler <- apply(nulls_euler$nulls, 2, cor, data_y)
null_cors_rodrigues <- apply(nulls_rodrigues$nulls, 2, cor, data_y)

df <- data.frame(
  r = c(null_cors_euler, null_cors_rodrigues),
  method = rep(c("Euler (ZYZ)", "Rodrigues (axis-angle)"), each = 500)
)

ggplot2::ggplot(df, ggplot2::aes(x = r, fill = method)) +
  ggplot2::geom_density(alpha = 0.5) +
  ggplot2::scale_fill_manual(values = c("steelblue", "darkorange")) +
  ggplot2::labs(x = "Null correlation (r)", y = "Density", fill = "Method") +
  ggplot2::theme_minimal()
```

The two distributions overlap almost entirely.
Any difference in the resulting p-values is well within the Monte Carlo noise
of the permutation test itself.

```{r}
#| label: compare-pvalues
obs_r <- cor(data_x, data_y)
p_euler <- mean(abs(null_cors_euler) >= abs(obs_r))
p_rodrigues <- mean(abs(null_cors_rodrigues) >= abs(obs_r))

data.frame(
  method = c("euler", "rodrigues"),
  p_value = c(p_euler, p_rodrigues)
)
```

Three things explain why the theoretical non-uniformity is harmless in practice:

- _Axis uniformity dominates._
  The axis determines which vertex maps near which other vertex.
  Both methods produce perfectly uniform axes.
  The spatial scrambling --- the part that matters for the null hypothesis ---
  is equivalent.

- _Non-trivial rotations are all you need._
  The slight oversampling of small-angle rotations means a few more
  near-identity permutations in the null distribution.
  These are a tiny fraction of 500 or 1000 permutations and do not meaningfully
  shift the distribution.

- _Monte Carlo noise swamps the difference._
  With a finite number of permutations, the sampling variability in the
  estimated p-value is far larger than the bias introduced by non-Haar angle
  sampling.

## Choosing a method

The `rotation` argument is available in all five spin-based null model
functions: `null_alexander_bloch()`, `null_spin_vasa()`,
`null_spin_hungarian()`, `null_baum()`, and `null_cornblath()`.

```{r}
#| label: usage-example
#| eval: false
null_alexander_bloch(
  data,
  coords,
  n_perm = 1000L,
  rotation = "rodrigues"
)
```

When to use which:

- _Reproducing Python neuromaps results_ --- stick with `"euler"` (the default).
  With the same seed, you will get the same rotation matrices as the Python
  implementation.

- _Working purely in R and preferring cleaner numerics_ --- `"rodrigues"` is the
  more natural choice.
  The axis sampling is more transparent, the code avoids `acos()` boundary
  issues, and the practical impact on spin test p-values is negligible.

- _Publishing results_ --- either is defensible.
  If a reviewer asks, Euler is Haar-uniform and Rodrigues is not, but for spin
  test inference the distinction is academic.
  State which you used and move on.

## Summary

|                          | Euler (ZYZ)            | Rodrigues (axis-angle) |
|--------------------------|------------------------|------------------------|
| Haar-uniform on SO(3)    | Yes                    | No (mild angle bias)   |
| Axis uniform on $S^2$    | Yes                    | Yes                    |
| Numerical stability      | `acos()` boundaries    | Stable everywhere      |
| Geometric interpretation | Indirect (3 rotations) | Direct (axis + angle)  |
| Matches Python neuromaps | Yes                    | No                     |

The default is there for cross-language reproducibility.
The alternative is there because sometimes the R-native path is the cleaner
one, and the statistical cost of taking it is effectively zero.

## References

Alexander-Bloch AF, Shou H, Liu S, et al. (2018). On testing for spatial
correspondence between maps of human brain structure and function.
_NeuroImage_, 175, 111--120. doi:10.1016/j.neuroimage.2018.04.023

Markello RD, Hansen JY, Liu Z-Q, et al. (2022). neuromaps: structural and
functional interpretation of brain maps. _Nature Methods_, 19, 1472--1480.
doi:10.1038/s41592-022-01625-w

Markello RD, Misic B (2021). Comparing spatial null models for brain maps.
_NeuroImage_, 236, 118052. doi:10.1016/j.neuroimage.2021.118052

---
title: "Surface Geometry"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Surface Geometry}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 5
)
```

The cortical surface is not flat. Two regions can be close in Euclidean
space---a straight line through the brain---yet far apart along the
cortical sheet. This distinction matters whenever distance drives an
analysis: computing spatial autocorrelation, selecting neighbours, or
estimating how much cortex a region actually covers.

neuromapr provides tools for working with triangular surface meshes:
building graphs, computing geodesic distances, calculating per-vertex
surface areas, and converting between neuroimaging file formats.

```{r setup}
library(neuromapr)
```

## Surface meshes

A cortical surface mesh consists of two pieces:

- **Vertices**: an n x 3 matrix of 3D coordinates, one row per point
  on the surface.
- **Faces**: an m x 3 matrix of integer indices, each row defining a
  triangle by referencing three vertices.

Together they describe the geometry. We build a small mesh by hand to
make the examples concrete:

```{r mesh}
vertices <- matrix(c(
  0, 0, 0,
  1, 0, 0,
  0.5, 1, 0,
  1.5, 1, 0,
  0, 2, 0,
  1, 2, 0
), ncol = 3, byrow = TRUE)

faces <- matrix(c(
  1, 2, 3,
  2, 3, 4,
  3, 4, 6,
  3, 5, 6
), ncol = 3, byrow = TRUE)
```

Six vertices and four triangles forming a small irregular strip. Real
cortical surfaces have tens of thousands of vertices, but the
operations are the same.

## Building a surface graph

`make_surf_graph()` converts a mesh into a weighted igraph object. Each
edge connects two vertices that share a triangle, and the weight is the
Euclidean distance between them.

```{r graph}
g <- make_surf_graph(vertices, faces)
g
```

The graph captures the connectivity structure of the mesh---which
vertices are adjacent and how far apart they are along the surface. It
is the foundation for geodesic distance computation.

Here is what the graph looks like, using the vertex positions as layout
coordinates. Edge labels show the Euclidean distance between connected
vertices:

```{r igraph-plot, fig.cap = "Surface graph of the toy mesh. Vertices are positioned at their x-y coordinates, edges are weighted by Euclidean distance."}
layout_xy <- vertices[, 1:2]
edge_weights <- round(
  igraph::E(g)$weight, 2
)

plot(
  g,
  layout = layout_xy,
  vertex.size = 20,
  vertex.color = "steelblue",
  vertex.label.color = "white",
  vertex.label.cex = 0.9,
  edge.label = edge_weights,
  edge.label.cex = 0.8,
  edge.color = "grey40",
  main = "Surface mesh graph"
)
```

On a real cortical surface, this graph has thousands of nodes and edges.
The structure is the same: vertices connected by triangle edges, weights
encoding surface distances.

## Geodesic distances

Euclidean distance is a straight line through 3D space. Geodesic
distance follows the surface, taking the shortest path along mesh edges.
For a convoluted surface like the cortex, the two can differ
substantially.

`get_surface_distance()` computes geodesic distances via Dijkstra's
algorithm on the surface graph:

```{r geodesic-full}
dmat <- get_surface_distance(vertices, faces)
dmat
```

The result is a symmetric n x n matrix of shortest-path distances.

### Partial distance matrices

Computing the full n x n matrix gets expensive for large meshes. If you
only need distances from a subset of source vertices, pass them
explicitly:

```{r geodesic-partial}
dmat_partial <- get_surface_distance(
  vertices, faces, source_vertices = c(1, 6)
)
dmat_partial
```

A 2 x 6 matrix: distances from vertices 1 and 6 to all six vertices.
For parcellation centroid computations and neighbourhood queries, this
selective approach saves considerable time.

### Euclidean vs geodesic

On a flat mesh the two distance measures coincide. On a folded cortex,
geodesic distances are always greater than or equal to Euclidean, often
substantially so for regions separated by a sulcus.

```{r compare-distances}
euclid <- as.matrix(dist(vertices))
geodesic <- get_surface_distance(vertices, faces)

max(abs(geodesic - euclid))
```

On our flat toy mesh the difference is zero. On real cortical surfaces,
it is not.

## Vertex areas

Each vertex on a surface mesh represents a patch of cortex. Knowing its
area matters for area-weighted averaging and for understanding how much
cortical surface a parcel covers.

`vertex_areas()` distributes each triangle's area equally among its
three vertices:

```{r vertex-areas}
areas <- vertex_areas(vertices, faces)
areas
```

```{r total-area}
sum(areas)
```

The total vertex area equals the total surface area of all triangles.
Corner vertices (part of fewer triangles) get less area than interior
vertices (shared by more triangles).

### Checking the geometry

A sanity check: compute triangle areas directly and verify they sum to
the same total.

```{r triangle-check}
tri_area <- function(v, f) {
  a <- v[f[1], ]
  b <- v[f[2], ]
  cc <- v[f[3], ]
  ab <- b - a
  ac <- cc - a
  cross <- c(
    ab[2] * ac[3] - ab[3] * ac[2],
    ab[3] * ac[1] - ab[1] * ac[3],
    ab[1] * ac[2] - ab[2] * ac[1]
  )
  0.5 * sqrt(sum(cross^2))
}

total_tri <- sum(vapply(
  seq_len(nrow(faces)),
  function(i) tri_area(vertices, faces[i, ]),
  numeric(1)
))

all.equal(sum(areas), total_tri)
```

## Format conversions

Cortical surfaces come in many file formats. FreeSurfer uses its own
binary formats (`.annot` for parcellations, `.curv`/`.thickness`/`.sulc`
for morphometry), while modern tools increasingly expect GIFTI.
neuromapr provides two conversion functions.

### Annotation to GIFTI

`annot_to_gifti()` converts a FreeSurfer `.annot` file (parcellation
labels) to a GIFTI label file:

```r
annot_to_gifti("lh.aparc.annot", "lh.aparc.label.gii")
```

### Morphometry to GIFTI

`fsmorph_to_gifti()` converts FreeSurfer morphometry files to GIFTI
func files:

```r
fsmorph_to_gifti("lh.thickness", "lh.thickness.func.gii")
```

Both functions require the **freesurferformats** and **gifti** packages.
If no output path is given, they replace the file extension
automatically (`.annot` becomes `.label.gii`, `.curv` becomes
`.func.gii`).

## Resampling between coordinate spaces

Brain maps from different studies often live in different coordinate
spaces (fsaverage vs fsLR) and at different mesh densities (32k vs
164k). Before you can compare them, they need to be in the same space
at the same resolution.

`transform_to_space()` handles this resampling using Connectome
Workbench via the **ciftiTools** package:

```r
transformed <- transform_to_space(
  "map.func.gii",
  target_space = "fsLR",
  target_density = "32k"
)
```

For pairwise comparisons, `resample_images()` aligns two maps into a
common space with a single call:

```r
aligned <- resample_images(
  src = "map_a.func.gii",
  trg = "map_b.func.gii",
  src_space = "fsaverage",
  trg_space = "fsLR",
  strategy = "transform_to_trg"
)
```

Four strategies are available:

- `"downsample_only"` --- both maps already share a coordinate space;
  downsample the higher-density one.
- `"transform_to_src"` --- bring the target into the source's space.
- `"transform_to_trg"` --- bring the source into the target's space.
- `"transform_to_alt"` --- bring both into a third space and density.

All resampling functions require Connectome Workbench (`wb_command`).
Use `check_wb_command()` to verify it is installed and locatable:

```r
check_wb_command()
```

## Putting it together

A typical workflow that combines surface geometry with the rest of
neuromapr:

1. Load a surface mesh from a GIFTI `.surf.gii` file.
2. Compute geodesic distances between parcel centroids.
3. Feed that distance matrix into spatially-constrained null model
   generation.

```r
centroids <- get_parcel_centroids(
  vertices, labels, method = "surface"
)
distmat <- as.matrix(dist(centroids))
nulls <- generate_nulls(
  parcel_data,
  method = "burt2020",
  distmat = distmat
)
```

For analyses where the distinction between Euclidean and geodesic
distance matters---highly folded cortex, parcels separated by deep
sulci---computing geodesic distances between centroids can produce
more accurate null models.

## References

Markello RD, Hansen JY, Liu Z-Q, et al. (2022). neuromaps: structural
and functional interpretation of brain maps. *Nature Methods*, 19,
1472--1480. doi:10.1038/s41592-022-01625-w

Robinson EC, Jbabdi S, Glasser MF, et al. (2014). MSM: a new flexible
framework for Multimodal Surface Matching. *NeuroImage*, 100,
414--426. doi:10.1016/j.neuroimage.2014.07.033

---
title: "Getting Started with neuromapr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with neuromapr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 4
)
```

Two brain maps look correlated. But the brain is spatially
autocorrelated---nearby regions tend to have similar values simply
because they are neighbours. A standard correlation test ignores this
structure, which inflates p-values and produces false positives.

neuromapr exists to answer one question honestly: **is the similarity
between two brain maps more than spatial structure alone would predict?**

It generates surrogate maps that preserve the spatial autocorrelation of
the original data, then measures where the real correlation falls in that
null distribution. The framework follows Markello et al. (2022).

```{r setup}
library(neuromapr)
```

## Comparing two brain maps

The core workflow: load data, compare, interpret. We simulate two
brain maps here to keep things self-contained, but in practice you would
read real data from GIFTI or NIfTI files.

```{r simulate-data}
set.seed(42)
n <- 100
coords <- matrix(rnorm(n * 3), ncol = 3)
distmat <- as.matrix(dist(coords))

map_a <- rnorm(n)
map_b <- 0.4 * map_a + rnorm(n, sd = 0.8)
```

Two maps with a planted correlation around 0.4, plus a distance matrix
describing the spatial layout. The simplest comparison:

```{r simple-compare}
result <- compare_maps(map_a, map_b, verbose = FALSE)
result
```

The `r` and parametric `p` tell the familiar story. But that `p`
assumes independence between observations---an assumption the brain
violates everywhere.

## Adding a spatial null model

To get an honest p-value, we generate surrogate versions of `map_a`
that preserve its spatial autocorrelation, then check whether the
observed correlation is unusual against correlations with those
surrogates.

neuromapr provides eight null model methods. For parcellated data with
a distance matrix, variogram matching (`burt2020`) is a strong default:

```{r null-comparison}
result_null <- compare_maps(
  map_a, map_b,
  null_method = "burt2020",
  distmat = distmat,
  n_perm = 500L,
  seed = 1,
  verbose = FALSE
)
result_null
```

The `p_null` value now accounts for spatial autocorrelation. If it
remains small, the relationship between the two maps is unlikely to be
an artifact of shared spatial structure.

## Visualising the null distribution

The comparison object stores all null correlations, making it
straightforward to see where the observed value falls:

```{r null-distribution-plot, fig.cap = "Null correlations from 500 variogram-matching surrogates. The dashed red line marks the observed correlation."}
null_df <- data.frame(r = result_null$null_r)
obs_r <- result_null$r

ggplot2::ggplot(null_df, ggplot2::aes(x = r)) +
  ggplot2::geom_histogram(
    bins = 30, fill = "steelblue", alpha = 0.7
  ) +
  ggplot2::geom_vline(
    xintercept = obs_r,
    linetype = "dashed", color = "red", linewidth = 1
  ) +
  ggplot2::labs(
    x = "Null correlation (r)",
    y = "Count"
  ) +
  ggplot2::theme_minimal()
```

When the observed correlation sits well outside the null distribution,
the relationship is more likely genuine. When it sits inside, spatial
autocorrelation alone can explain the similarity.

## Pre-computing null distributions

Generating nulls is the expensive part. If you want to compare the same
map against several targets, generate once and reuse:

```{r precompute}
nulls <- generate_nulls(
  map_a,
  method = "moran",
  distmat = distmat,
  n_perm = 200L,
  seed = 42
)
nulls
```

The `null_distribution` object has `print`, `summary`, `plot`, and
`as.matrix` methods. The `summary` gives per-element null means and
standard deviations. The `plot` method shows the null distribution for
a chosen element:

```{r plot-nulls, fig.cap = "Built-in plot method for a null_distribution object."}
plot(nulls, parcel = 1L)
```

Pass the pre-computed object to `compare_maps()`:

```{r reuse-nulls}
compare_maps(map_a, map_b, nulls = nulls, verbose = FALSE)
```

No redundant computation when the same source map appears in multiple
comparisons.

## Working with real brain map files

In practice, brain maps live on disk. `compare_maps()` and
`read_brain_map_values()` accept file paths directly:

```r
values <- read_brain_map_values("cortical_thickness.func.gii")

result <- compare_maps(
  "path/to/map_a.func.gii",
  "path/to/map_b.func.gii",
  null_method = "moran",
  distmat = distmat
)
```

Both GIFTI (`.func.gii`) and NIfTI (`.nii.gz`) formats are supported.
No separate reading step needed.

## Accessing the neuromaps atlas collection

The neuromaps project curates a registry of published brain map
annotations---PET receptor maps, gene expression, cortical thickness,
and more. neuromapr provides direct access to this collection:

```r
neuromaps_available()

neuromaps_available(source = "beliveau", tags = "pet")
```

Once you find the annotation you want, download it:

```r
paths <- fetch_neuromaps_annotation(
  "abagen", "genepc1", "fsaverage",
  density = "10k"
)
```

Files are cached locally, so subsequent calls skip the download. The
registry is fetched from the neuromaps GitHub repository and cached for
the session.

## Custom metrics with `permtest_metric()`

`compare_maps()` is built around correlation. For other metrics---mean
absolute error, cosine similarity, anything that takes two vectors and
returns a scalar---`permtest_metric()` provides the same null model
machinery:

```{r permtest}
mae <- function(a, b) mean(abs(a - b))

result_mae <- permtest_metric(
  map_a, rnorm(n),
  metric_func = mae,
  n_perm = 200L,
  seed = 1
)
result_mae$observed
result_mae$p_value
```

Add `null_method` for spatially-constrained surrogates instead of
random permutation:

```{r permtest-spatial}
result_spatial <- permtest_metric(
  map_a, rnorm(n),
  metric_func = mae,
  n_perm = 200L,
  seed = 1,
  null_method = "burt2020",
  distmat = distmat
)
result_spatial$p_value
```

## Choosing a null model

The right model depends on your data:

| Situation | Recommended method |
|---|---|
| Parcellated data with distance matrix | `"burt2020"` or `"moran"` |
| Vertex-level with sphere coordinates | `"spin_hungarian"` or `"alexander_bloch"` |
| Parcellated with sphere coordinates | `"baum"` or `"cornblath"` |
| Spatial autoregressive structure | `"burt2018"` |

`vignette("null-models")` walks through each method in detail.

## What comes next

This vignette covered the essential workflow: compare maps, generate
spatially-constrained null distributions, and interpret results.
From here:

- `vignette("null-models")` dives into all eight null model methods
  and their assumptions.
- `vignette("parcellation")` covers aggregating vertex data into
  parcels and parcel-level null models.
- `vignette("surface-geometry")` explains geodesic distances, surface
  graphs, and vertex area calculations.

## References

Markello RD, Hansen JY, Liu Z-Q, et al. (2022). neuromaps: structural
and functional interpretation of brain maps. *Nature Methods*, 19,
1472--1480. doi:10.1038/s41592-022-01625-w

---
title: "Choosing and Using Null Models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Choosing and Using Null Models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.5
)
```

Correlating two brain maps is easy. Knowing whether that correlation
means anything is the hard part.

The brain's spatial autocorrelation means that even unrelated maps can
appear correlated simply because nearby regions share similar values.
Spatial null models address this by generating surrogate maps that
preserve the spatial structure of the original while destroying the
specific pattern of interest. If the observed correlation still stands
out against these surrogates, you have evidence for a genuine
relationship.

neuromapr implements eight null model methods. They differ in their
assumptions, computational cost, and the kind of data they handle. This
vignette walks through each one, explains when to reach for it, and
names the tradeoffs plainly.

```{r setup}
library(neuromapr)
```

## The landscape

The eight methods split into three families.

**Distance-based methods** work with any distance matrix---Euclidean,
geodesic, or custom. No coordinate system required, and they apply
equally to parcellated and vertex-level data.

- `burt2020` --- variogram matching
- `burt2018` --- spatial autoregressive model
- `moran` --- Moran spectral randomization

**Spin-based methods** require spherical coordinates (one set per
hemisphere) and operate on vertex-level data. They rotate the
coordinates on the sphere, then reassign values based on proximity.

- `alexander_bloch` --- nearest-neighbour reassignment
- `spin_vasa` --- greedy sequential assignment
- `spin_hungarian` --- optimal (Hungarian) assignment

**Parcel spin methods** combine spin rotations with parcellation-level
reassignment. They need both spherical coordinates and a parcellation
scheme.

- `baum` --- maximum-overlap parcel assignment
- `cornblath` --- majority-vote parcel assignment

```{r shared-data}
set.seed(42)
n <- 80
coords_3d <- matrix(rnorm(n * 3), ncol = 3)
distmat <- as.matrix(dist(coords_3d))

map_x <- rnorm(n)
```

## Distance-based methods

### Variogram matching (`burt2020`)

Burt et al. (2020) generate surrogates by randomly permuting the
original values, then smoothing them at different spatial scales and
selecting the surrogate whose variogram best matches the original. The
result is a map with the same rank-order distribution and approximately
the same distance-dependent spatial structure.

```{r burt2020}
nulls_vario <- generate_nulls(
  map_x,
  method = "burt2020",
  distmat = distmat,
  n_perm = 100L,
  seed = 1
)
nulls_vario
```

Variogram matching is the most general-purpose method. It makes no
assumptions about the coordinate system, works on any distance matrix,
and handles both parcellated and vertex-level data. The cost scales
with the number of smoothing parameters tried per permutation, but for
typical parcellation sizes (50--400 regions) it runs quickly.

**When to use it:** Default recommendation when you have a distance
matrix and want minimal assumptions about the spatial generating
process.

### Spatial autoregressive model (`burt2018`)

Burt et al. (2018) take a generative approach: fit a spatial
autoregressive (SAR) model to the data, then draw new samples from
that model. The SAR model captures spatial autocorrelation through a
weight matrix derived from distances and an estimated autocorrelation
parameter.

```{r burt2018}
nulls_sar <- generate_nulls(
  map_x,
  method = "burt2018",
  distmat = distmat,
  n_perm = 100L,
  seed = 1
)
nulls_sar
```

The SAR model estimates two parameters: `rho` (spatial autocorrelation
strength) and `d0` (distance decay scale). These are stored in the
result:

```{r sar-params}
nulls_sar$params$rho
nulls_sar$params$d0
```

**When to use it:** When you believe the data follows an autoregressive
spatial process and want surrogates drawn from that parametric model.
More assumption-heavy than variogram matching, but computationally
cheaper per surrogate once the parameters are estimated.

### Moran spectral randomization (`moran`)

Wagner and Dray (2015) decompose the data onto Moran's eigenvector
maps---spatial eigenvectors of a weighted connectivity matrix---and
then randomly perturb the coefficients. The `"singleton"` procedure
flips the sign of each coefficient at random; the `"pair"` procedure
applies random 2D rotations to pairs of coefficients with similar
eigenvalues.

```{r moran}
nulls_moran <- generate_nulls(
  map_x,
  method = "moran",
  distmat = distmat,
  n_perm = 100L,
  seed = 1,
  procedure = "singleton"
)
nulls_moran
```

**When to use it:** When you want to preserve the spectral properties
of the spatial autocorrelation rather than the variogram. The `"pair"`
procedure is more conservative, preserving the eigenvalue spectrum more
faithfully.

## Spin-based methods

Spin tests work on a fundamentally different principle. Instead of
generating synthetic data from a statistical model, they take the real
data and rotate it on the cortical sphere. Values get reassigned to the
original positions based on proximity to the rotated positions.

All spin methods require coordinates organized as a list with `$lh` and
`$rh` components---one matrix per hemisphere. Each hemisphere rotates
independently.

```{r spin-coords}
n_lh <- 40
n_rh <- 40
sphere_coords <- list(
  lh = matrix(rnorm(n_lh * 3), ncol = 3),
  rh = matrix(rnorm(n_rh * 3), ncol = 3)
)

vertex_data <- rnorm(n_lh + n_rh)
```

### Original spin test (`alexander_bloch`)

The method that started it all (Alexander-Bloch et al., 2018). After
rotating coordinates, each rotated vertex maps to the nearest original
vertex. Multiple rotated vertices can map to the same original, so some
values get duplicated while others are dropped.

```{r alexander-bloch}
nulls_ab <- null_alexander_bloch(
  vertex_data, sphere_coords,
  n_perm = 100L, seed = 1
)
nulls_ab
```

The simplest spin test and the fastest. The tradeoff: the many-to-one
mapping distorts the value distribution. If preserving the distribution
matters, the Hungarian variant is more appropriate.

### Greedy assignment (`spin_vasa`)

Vasa et al. (2018) proposed one-to-one greedy assignment: iterate
through vertices in order, assigning each to the nearest available
(unassigned) rotated vertex. Every value appears exactly once, but
assignment quality depends on vertex ordering.

```{r spin-vasa}
nulls_vasa <- null_spin_vasa(
  vertex_data, sphere_coords,
  n_perm = 100L, seed = 1
)
nulls_vasa
```

### Optimal assignment (`spin_hungarian`)

Markello and Misic (2021) showed that optimal (Hungarian) assignment
removes the ordering bias of the greedy approach. It minimizes total
reassignment distance globally, producing the best possible one-to-one
mapping between original and rotated positions.

```{r spin-hungarian, eval = rlang::is_installed("clue")}
nulls_hung <- null_spin_hungarian(
  vertex_data, sphere_coords,
  n_perm = 100L, seed = 1
)
nulls_hung
```

The Hungarian algorithm requires the **clue** package and runs in cubic
time in the number of vertices. Slower than the greedy approach, but
strictly better assignments.

**When to use it:** The recommended spin method for vertex-level data
when you want one-to-one assignment and can afford the extra
computation.

## Parcel spin methods

When your data is parcellated---one value per brain region rather than
per vertex---the spin rotation still happens at vertex level, but the
reassignment happens at the parcel level. This requires both sphere
coordinates (for the rotation) and a parcellation vector (to know which
vertices belong to which parcel).

```{r parcel-setup}
n_parcel_lh <- 340
n_parcel_rh <- 340
parcel_coords <- list(
  lh = matrix(rnorm(n_parcel_lh * 3), ncol = 3),
  rh = matrix(rnorm(n_parcel_rh * 3), ncol = 3)
)
parcellation <- rep(1:68, each = 10)
parcel_data <- rnorm(68)
```

### Maximum overlap (`baum`)

Baum et al. (2020) proposed that after rotating and reassigning vertices
to their nearest neighbours, each original parcel takes the value of
whichever rotated parcel has the most vertex overlap with it. If most
vertices in parcel 3 now carry the label of parcel 7 after rotation,
parcel 3 gets parcel 7's value.

```{r baum}
nulls_baum <- null_baum(
  parcel_data, parcel_coords, parcellation,
  n_perm = 50L, seed = 1
)
nulls_baum
```

### Majority vote (`cornblath`)

Cornblath et al. (2020) introduced a refinement: each rotated vertex
first receives the label of its nearest *non-medial-wall* original
vertex. Parcels are then reassigned by majority vote among the
resulting vertex labels. This handles the medial wall more gracefully
than Baum's method, since rotated vertices that land on the medial wall
get interpolated to the nearest valid region instead of being lost.

```{r cornblath}
nulls_corn <- null_cornblath(
  parcel_data, parcel_coords, parcellation,
  n_perm = 50L, seed = 1
)
nulls_corn
```

**When to use it:** Prefer Cornblath over Baum when your parcellation
includes medial wall vertices (labeled 0 or `NA`), which is the common
case for standard atlases.

## Comparing null distributions visually

A practical way to understand what each method does: look at the
surrogate values it produces. Here we compare the first 50 surrogates
from two methods on the same data:

```{r comparison-plot, fig.cap = "Surrogate distributions from variogram-matching (left) and Moran spectral randomization (right). The dashed red line is the original data distribution.", fig.width = 7, fig.height = 3.5}
df <- data.frame(
  value = c(
    as.vector(nulls_vario$nulls[, 1:50]),
    as.vector(nulls_moran$nulls[, 1:50])
  ),
  method = rep(
    c("burt2020 (variogram)", "moran (spectral)"),
    each = n * 50
  )
)

ggplot2::ggplot(df, ggplot2::aes(x = value)) +
  ggplot2::geom_density(fill = "steelblue", alpha = 0.5) +
  ggplot2::facet_wrap(~method) +
  ggplot2::geom_density(
    data = data.frame(value = map_x),
    color = "red", linetype = "dashed", linewidth = 0.8
  ) +
  ggplot2::labs(x = "Value", y = "Density") +
  ggplot2::theme_minimal()
```

The red dashed line shows the original data distribution. Good
surrogates should have a similar overall distribution (same value pool)
but different spatial arrangements.

## Using `permtest_metric()` for custom metrics

`compare_maps()` is built around correlation. If you need a different
metric---mean absolute error, cosine similarity, whatever---use
`permtest_metric()` with any function that takes two vectors and returns
a scalar:

```{r permtest}
mae <- function(a, b) mean(abs(a - b))

result <- permtest_metric(
  map_x, rnorm(n),
  metric_func = mae,
  n_perm = 200L,
  seed = 1
)
result$observed
result$p_value
```

Pass `null_method` for spatially-constrained surrogates instead of
random permutation:

```{r permtest-spatial}
result_spatial <- permtest_metric(
  map_x, rnorm(n),
  metric_func = mae,
  n_perm = 200L,
  seed = 1,
  null_method = "burt2020",
  distmat = distmat
)
result_spatial$p_value
```

## Building custom null distributions

If you generate surrogates through some other means---your own
algorithm, or an external tool---you can wrap them in a
`null_distribution` object to use with the rest of neuromapr:

```{r custom-nulls}
custom_nulls <- matrix(rnorm(n * 50), nrow = n, ncol = 50)
nd <- new_null_distribution(
  custom_nulls,
  method = "custom",
  observed = map_x
)
nd
summary(nd)
```

The `new_null_distribution()` constructor produces an object that works
with `compare_maps(nulls = ...)`, `plot()`, `summary()`, and
`as.matrix()`.

## Decision guide

When faced with a choice, these questions narrow it down:

1. **Do you have spherical coordinates?** If not, use a distance-based
   method (`burt2020`, `burt2018`, or `moran`).

2. **Is your data parcellated or vertex-level?** Parcellated + sphere
   coords points to `baum` or `cornblath`. Vertex-level + sphere
   coords points to `spin_hungarian` or `alexander_bloch`.

3. **Does the medial wall matter?** If your parcellation has medial
   wall vertices, `cornblath` handles them more carefully than `baum`.

4. **Do you want a parametric model?** `burt2018` fits an explicit SAR
   model. `moran` preserves spectral properties. `burt2020` is
   distribution-free.

5. **Computational budget?** `alexander_bloch` and `burt2018` are the
   fastest. `spin_hungarian` and `burt2020` are the slowest (but often
   the most principled for their respective data types).

## References

Alexander-Bloch AF, Shou H, Liu S, et al. (2018). On testing for
spatial correspondence between maps of human brain structure and
function. *NeuroImage*, 175, 111--120.

Baum GL, Cui Z, Roalf DR, et al. (2020). Development of
structure--function coupling in human brain networks during youth.
*PNAS*, 117, 21854--21861.

Burt JB, Demirtas M, Eckner WJ, et al. (2018). Hierarchy of
transcriptomic specialization across human cortex captured by
structural connectivity. *Nature Neuroscience*, 21, 1251--1259.

Burt JB, Helmer M, Shinn M, et al. (2020). Generative modeling of
brain maps with spatial autocorrelation. *NeuroImage*, 220, 117038.

Cornblath EJ, Ashourvan A, Kim JZ, et al. (2020). Temporal sequences
of brain activity at rest are constrained by white matter structure and
modulated by cognitive demands. *Communications Biology*, 3, 590.

Markello RD, Misic B (2021). Comparing spatial null models for brain
maps. *NeuroImage*, 236, 118052.

Vasa F, Seidlitz J, Romero-Garcia R, et al. (2018). Adolescent tuning
of association cortex in human structural brain networks. *Cerebral
Cortex*, 28, 3293--3303.

Wagner HH, Dray S (2015). Generating spatially constrained null models
for irregularly spaced data using Moran spectral randomization methods.
*Methods in Ecology and Evolution*, 6, 1169--1178.

# print.null_distribution / prints method, permutations, and observations

    Code
      res <- print(nd)
    Message
      
      -- Null Distribution 
      * Method: moran
      * Permutations: 10
      * Observations: 5

# print.neuromaps_enhanced_comparison / prints without null model info when no nulls used

    Code
      res <- print(result)
    Message
      
      -- Brain Map Comparison 
      Method: pearson
      r = 0.5, p = 0.001
      n = 100

# print.neuromaps_enhanced_comparison / prints null model info when nulls used

    Code
      print(result)
    Message
      
      -- Brain Map Comparison 
      Method: pearson
      r = 0.5, p = 0.001
      n = 100
      --------------------------------------------------------------------------------
      Null model: burt2020 (100 permutations)
      p_null = 0.05

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nulls-baum.R
\name{null_baum}
\alias{null_baum}
\title{Baum spin test null model}
\usage{
null_baum(
  data,
  coords,
  parcellation,
  n_perm = 1000L,
  seed = NULL,
  rotation = c("euler", "rodrigues")
)
}
\arguments{
\item{data}{Numeric vector of brain map values.}

\item{coords}{List with \verb{$lh} and \verb{$rh} matrices of spherical coordinates
(n x 3 each).}

\item{parcellation}{Integer vector of parcel labels for all vertices.
\code{0} and \code{NA} indicate medial wall.}

\item{n_perm}{Integer number of null permutations to generate.}

\item{seed}{Optional integer seed for reproducibility.}

\item{rotation}{Rotation generation method: \code{"euler"} (ZYZ Euler angles,
default, matches neuromaps Python) or \code{"rodrigues"} (Rodrigues axis-angle
formula).}
}
\value{
A \link{null_distribution} object.
}
\description{
Spin-based null model with maximum-overlap parcel reassignment. After
rotating vertex coordinates, each original parcel is assigned the value
of the rotated parcel with the most vertex overlap.
}
\examples{
coords <- list(lh = matrix(rnorm(30), 10, 3), rh = matrix(rnorm(30), 10, 3))
parcellation <- c(rep(1L, 5), rep(2L, 5), rep(3L, 5), rep(4L, 5))
data <- c(1.0, 2.0, 3.0, 4.0)
nd <- null_baum(data, coords, parcellation, n_perm = 10L, seed = 1L)
}
\references{
Baum GL et al. (2020) PNAS 117:21854-21861.
doi:10.1073/pnas.2005518117
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nulls-alexander-bloch.R
\name{null_alexander_bloch}
\alias{null_alexander_bloch}
\title{Alexander-Bloch spin test null model}
\usage{
null_alexander_bloch(
  data,
  coords,
  n_perm = 1000L,
  seed = NULL,
  rotation = c("euler", "rodrigues")
)
}
\arguments{
\item{data}{Numeric vector of brain map values.}

\item{coords}{List with \verb{$lh} and \verb{$rh} matrices of spherical coordinates
(n x 3 each).}

\item{n_perm}{Integer number of null permutations to generate.}

\item{seed}{Optional integer seed for reproducibility.}

\item{rotation}{Rotation generation method: \code{"euler"} (ZYZ Euler angles,
default, matches neuromaps Python) or \code{"rodrigues"} (Rodrigues axis-angle
formula).}
}
\value{
A \link{null_distribution} object.
}
\description{
Original vertex-level spin test: rotates coordinates and assigns each rotated
vertex the value of its nearest original vertex (no optimal matching).
}
\examples{
coords <- list(lh = matrix(rnorm(30), 10, 3), rh = matrix(rnorm(30), 10, 3))
data <- rnorm(20)
nd <- null_alexander_bloch(data, coords, n_perm = 10L, seed = 1L)
}
\references{
Alexander-Bloch AF et al. (2018) NeuroImage 175:111-120.
doi:10.1016/j.neuroimage.2018.04.023
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/surface-graph.R
\name{get_surface_distance}
\alias{get_surface_distance}
\title{Compute geodesic distances on a surface mesh}
\usage{
get_surface_distance(vertices, faces, source_vertices = NULL)
}
\arguments{
\item{vertices}{Numeric matrix (n x 3) of vertex coordinates.}

\item{faces}{Integer matrix (m x 3) of face indices (1-indexed).}

\item{source_vertices}{Optional integer vector of source vertex indices.
If \code{NULL}, computes the full n x n distance matrix.}
}
\value{
Numeric distance matrix. If \code{source_vertices} is provided, returns
a \verb{length(source_vertices) x n} matrix; otherwise an \verb{n x n} matrix.
}
\description{
Builds a graph from a triangular mesh and computes shortest-path (Dijkstra)
distances between vertices.
}
\examples{
vertices <- matrix(
  c(0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1),
  nrow = 4, byrow = TRUE
)
faces <- matrix(c(1L, 2L, 3L, 2L, 3L, 4L), nrow = 2, byrow = TRUE)
get_surface_distance(vertices, faces)
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parcellate.R
\name{parcellate}
\alias{parcellate}
\title{Parcellate brain map data}
\usage{
parcellate(data, parcellation, summary_func = mean)
}
\arguments{
\item{data}{Numeric vector or file path to a GIFTI/NIfTI brain map.}

\item{parcellation}{Integer vector of labels or file path to a GIFTI label
file.}

\item{summary_func}{Function to summarise each parcel (default: \link{mean}).}
}
\value{
Named numeric vector of parcel-level values.
}
\description{
High-level wrapper that reads data and parcellation from file paths or
vectors, then aggregates vertices into parcels.
}
\examples{
data <- c(1.0, 2.0, 3.0, 4.0)
labels <- c(1L, 1L, 2L, 2L)
parcellate(data, labels)
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats.R
\name{permtest_metric}
\alias{permtest_metric}
\title{Permutation test for any metric between brain maps}
\usage{
permtest_metric(
  x,
  y,
  metric_func = stats::cor,
  n_perm = 1000L,
  seed = NULL,
  null_method = NULL,
  distmat = NULL,
  coords = NULL,
  parcellation = NULL,
  ...
)
}
\arguments{
\item{x, y}{Numeric vectors.}

\item{metric_func}{Function taking \verb{(x, y)} and returning a scalar.}

\item{n_perm}{Integer number of permutations.}

\item{seed}{Optional integer seed for reproducibility.}

\item{null_method}{Optional null model method passed to \code{\link[=generate_nulls]{generate_nulls()}}.
If \code{NULL}, uses simple random permutation.}

\item{distmat}{Distance matrix (passed to \code{\link[=generate_nulls]{generate_nulls()}} if needed).}

\item{coords}{Coordinate list (passed to \code{\link[=generate_nulls]{generate_nulls()}} if needed).}

\item{parcellation}{Integer vector (passed to \code{\link[=generate_nulls]{generate_nulls()}} if needed).}

\item{...}{Additional arguments passed to \code{\link[=generate_nulls]{generate_nulls()}}.}
}
\value{
List with \verb{$observed}, \verb{$null_values}, \verb{$p_value}, and \verb{$n_perm}.
}
\description{
Computes a user-specified metric between two vectors and tests significance
using either spatially-constrained null surrogates or simple random
permutation.
}
\examples{
x <- rnorm(100)
y <- x + rnorm(100)
result <- permtest_metric(x, y, n_perm = 99L, seed = 1L)
result$observed
result$p_value
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parcellate.R
\name{unparcellate}
\alias{unparcellate}
\title{Unparcellate brain map data}
\usage{
unparcellate(parcel_data, parcellation, fill = NA_real_)
}
\arguments{
\item{parcel_data}{Named numeric vector of parcel values.}

\item{parcellation}{Integer vector of labels or file path to a GIFTI label
file.}

\item{fill}{Value for medial wall vertices (default: \code{NA_real_}).}
}
\value{
Numeric vector of vertex-level values.
}
\description{
Inverse of \code{\link[=parcellate]{parcellate()}}: maps parcel-level values back to vertices.
}
\examples{
parcel_data <- c("1" = 10, "2" = 20)
labels <- c(1L, 1L, 2L, 2L, 0L)
unparcellate(parcel_data, labels)
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nulls-spin.R
\name{null_spin_vasa}
\alias{null_spin_vasa}
\alias{null_spin_hungarian}
\title{Spin-test null models for brain maps}
\usage{
null_spin_vasa(
  data,
  coords,
  n_perm = 1000L,
  seed = NULL,
  rotation = c("euler", "rodrigues")
)

null_spin_hungarian(
  data,
  coords,
  n_perm = 1000L,
  seed = NULL,
  rotation = c("euler", "rodrigues")
)
}
\arguments{
\item{data}{Numeric vector of brain map values.}

\item{coords}{List with \verb{$lh} and \verb{$rh} matrices of spherical coordinates
(n x 3 each).}

\item{n_perm}{Integer number of null permutations to generate.}

\item{seed}{Optional integer seed for reproducibility.}

\item{rotation}{Rotation generation method: \code{"euler"} (ZYZ Euler angles,
default, matches neuromaps Python) or \code{"rodrigues"} (Rodrigues axis-angle
formula).}
}
\value{
A \link{null_distribution} object.
}
\description{
Generate spatially-constrained null distributions using spin-based
permutation of spherical coordinates.
}
\examples{
coords <- list(lh = matrix(rnorm(30), 10, 3), rh = matrix(rnorm(30), 10, 3))
data <- rnorm(20)
nd <- null_spin_vasa(data, coords, n_perm = 10L, seed = 1L)
}
\references{
Alexander-Bloch AF et al. (2018) NeuroImage 175:111-120.
doi:10.1016/j.neuroimage.2018.04.023

Vasa F et al. (2018) Cerebral Cortex 28:3293-3303.
doi:10.1093/cercor/bhx195

Markello RD, Misic B (2021) NeuroImage 236:118052.
doi:10.1016/j.neuroimage.2021.118052
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nulls-classes.R
\name{null_distribution}
\alias{null_distribution}
\alias{new_null_distribution}
\alias{print.null_distribution}
\alias{summary.null_distribution}
\alias{as.matrix.null_distribution}
\alias{plot.null_distribution}
\title{Create a null distribution object}
\usage{
new_null_distribution(nulls, method, observed, params = list())

\method{print}{null_distribution}(x, ...)

\method{summary}{null_distribution}(object, ...)

\method{as.matrix}{null_distribution}(x, ...)

\method{plot}{null_distribution}(x, parcel = 1L, ...)
}
\arguments{
\item{nulls}{Numeric matrix (n x n_perm) of surrogate values.}

\item{method}{Character string identifying the null model method.}

\item{observed}{Numeric vector of original data values.}

\item{params}{Named list of algorithm parameters.}

\item{x}{A \code{null_distribution} object.}

\item{...}{Ignored.}

\item{object}{A \code{null_distribution} object.}

\item{parcel}{Integer index of the parcel to plot.}
}
\value{
A \code{null_distribution} object.
}
\description{
Create a null distribution object
}
\examples{
nulls <- matrix(rnorm(30), nrow = 3, ncol = 10)
nd <- new_null_distribution(nulls, "test", observed = c(1, 2, 3))
print(nd)
summary(nd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms.R
\name{transform_to_space}
\alias{transform_to_space}
\title{Transform brain maps between coordinate spaces}
\usage{
transform_to_space(
  paths,
  target_space = c("fsLR", "fsaverage"),
  target_density = "32k",
  hemisphere = c("left", "right"),
  method = c("adaptive", "barycentric"),
  area_surf_current = NULL,
  area_surf_new = NULL,
  wb_path = NULL,
  verbose = TRUE
)
}
\arguments{
\item{paths}{Character vector of GIFTI file paths to transform.}

\item{target_space}{Target coordinate space: \code{"fsLR"} or \code{"fsaverage"}.}

\item{target_density}{Target mesh density (e.g., \code{"32k"}, \code{"164k"}).}

\item{hemisphere}{Which hemispheres to transform: \code{"left"}, \code{"right"},
or both.}

\item{method}{Resampling method: \code{"adaptive"} (default) or
\code{"barycentric"}.}

\item{area_surf_current}{Path to the current-resolution area-correction
surface (e.g. midthickness). Recommended when \code{method = "adaptive"} for
proper vertex-area correction (matches neuromaps Python behaviour).}

\item{area_surf_new}{Path to the target-resolution area-correction surface.
If \code{NULL} and \code{area_surf_current} is provided, ciftiTools resamples it
barycentrically.}

\item{wb_path}{Path to \code{wb_command} executable. If \code{NULL}, auto-detected
via \code{ciftiTools}.}

\item{verbose}{Logical, print progress messages.}
}
\value{
Character vector of transformed file paths.
}
\description{
Resamples GIFTI surface files between \code{fsaverage} and \code{fsLR} coordinate
spaces using Connectome Workbench via \code{ciftiTools}.
}
\examples{
\dontrun{
transform_to_space("map.func.gii", target_space = "fsLR")
}
}
\references{
Robinson EC et al. (2014) NeuroImage 100:414-426.
doi:10.1016/j.neuroimage.2014.07.033

Robinson EC et al. (2018) NeuroImage 167:150-165.
doi:10.1016/j.neuroimage.2017.10.037
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/neuromaps-registry.R
\name{clear_neuromaps_cache}
\alias{clear_neuromaps_cache}
\title{Clear cached neuromaps registry data}
\usage{
clear_neuromaps_cache()
}
\value{
\code{NULL}, invisibly.
}
\description{
Removes the session-level cache of the neuromaps annotation registry,
forcing a fresh download on the next call to \code{\link[=neuromaps_available]{neuromaps_available()}} or
\code{\link[=fetch_neuromaps_annotation]{fetch_neuromaps_annotation()}}.
}
\examples{
clear_neuromaps_cache()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parcellate.R
\name{vertices_to_parcels}
\alias{vertices_to_parcels}
\title{Aggregate vertex data into parcels}
\usage{
vertices_to_parcels(data, labels, summary_func = mean)
}
\arguments{
\item{data}{Numeric vector of vertex-level values.}

\item{labels}{Integer vector of parcel labels. \code{0} and \code{NA} are treated as
medial wall and excluded.}

\item{summary_func}{Function to summarise each parcel (default: \link{mean}).}
}
\value{
Named numeric vector of parcel-level values.
}
\description{
Summarises vertex-level data by parcellation labels.
}
\examples{
data <- c(1.0, 2.0, 3.0, 4.0)
labels <- c(1L, 1L, 2L, 2L)
vertices_to_parcels(data, labels)
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nulls-moran.R
\name{null_moran}
\alias{null_moran}
\title{Moran spectral randomization null model}
\usage{
null_moran(
  data,
  distmat,
  n_perm = 1000L,
  seed = NULL,
  procedure = c("pair", "singleton"),
  kernel = c("inverse_distance", "exponential", "gaussian", "bisquare"),
  tol = 1e-06
)
}
\arguments{
\item{data}{Numeric vector of brain map values.}

\item{distmat}{Distance matrix between parcels/vertices.}

\item{n_perm}{Integer number of null permutations to generate.}

\item{seed}{Optional integer seed for reproducibility.}

\item{procedure}{Character, either \code{"pair"} (default, random 2D rotations
of near-degenerate eigenvector pairs, matches neuromaps Python) or
\code{"singleton"} (random sign flips of individual eigenvectors).}

\item{kernel}{Weight matrix kernel: \code{"inverse_distance"} (default, matches
neuromaps Python), \code{"exponential"}, \code{"gaussian"}, or \code{"bisquare"}.}

\item{tol}{Numeric tolerance for eigenvalue comparison.}
}
\value{
A \link{null_distribution} object.
}
\description{
Generates spatially-constrained surrogate brain maps using Moran's
eigenvector maps (MEMs) for spectral randomization.
}
\examples{
data <- rnorm(50)
distmat <- as.matrix(dist(matrix(rnorm(100), 50, 2)))
nd <- null_moran(data, distmat, n_perm = 10L, seed = 1L)
}
\references{
Wagner HH, Dray S (2015) Methods in Ecology and Evolution 6:1169-1178.
doi:10.1111/2041-210X.12407
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image-utils.R
\name{vertex_areas}
\alias{vertex_areas}
\title{Compute per-vertex surface areas}
\usage{
vertex_areas(vertices, faces)
}
\arguments{
\item{vertices}{Numeric matrix (n x 3) of vertex coordinates.}

\item{faces}{Integer matrix (m x 3) of face indices (1-indexed).}
}
\value{
Numeric vector of length \code{nrow(vertices)}.
}
\description{
Each triangle distributes one-third of its area to each of its three
vertices. Triangle area is computed via the cross-product formula.
}
\examples{
vertices <- matrix(c(0, 1, 0, 0, 0, 1, 0, 0, 0), nrow = 3, byrow = TRUE)
faces <- matrix(c(1L, 2L, 3L), nrow = 1)
vertex_areas(vertices, faces)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nulls-variogram.R
\name{null_burt2020}
\alias{null_burt2020}
\title{Variogram-matching null model}
\usage{
null_burt2020(
  data,
  distmat,
  n_perm = 1000L,
  seed = NULL,
  ns = 500L,
  nh = 25L,
  pv = 25,
  knn = 1000L,
  deltas = seq(0.1, 0.9, by = 0.1),
  kernel = c("exponential", "gaussian", "uniform"),
  resample = FALSE
)
}
\arguments{
\item{data}{Numeric vector of brain map values.}

\item{distmat}{Distance matrix between parcels/vertices.}

\item{n_perm}{Integer number of null permutations to generate.}

\item{seed}{Optional integer seed for reproducibility.}

\item{ns}{Integer, subsample size for variogram computation.}

\item{nh}{Integer, number of distance bins for variogram.}

\item{pv}{Numeric, percentile cutoff for maximum distance in variogram.}

\item{knn}{Integer, number of nearest neighbors for smoothing.}

\item{deltas}{Numeric vector of smoothing levels (fractions of \code{knn}).}

\item{kernel}{Smoothing kernel function.}

\item{resample}{Logical. If \code{FALSE} (default, matches brainsmash Python),
variogram subsample indices are computed once and reused across all
permutations. If \code{TRUE}, a fresh subsample is drawn each time.}
}
\value{
A \link{null_distribution} object.
}
\description{
Generates spatially-constrained surrogate brain maps by matching the
empirical variogram of the original data through smoothed random
permutations.
}
\examples{
data <- rnorm(50)
distmat <- as.matrix(dist(matrix(rnorm(100), 50, 2)))
nd <- null_burt2020(data, distmat, n_perm = 10L, seed = 1L)
}
\references{
Burt JB et al. (2020) NeuroImage 220:117038.
doi:10.1016/j.neuroimage.2020.117038
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nulls.R
\name{generate_nulls}
\alias{generate_nulls}
\title{Generate null distributions for brain map data}
\usage{
generate_nulls(
  data,
  method = c("burt2020", "moran", "spin_vasa", "spin_hungarian", "alexander_bloch",
    "baum", "cornblath", "burt2018"),
  n_perm = 1000L,
  distmat = NULL,
  coords = NULL,
  parcellation = NULL,
  seed = NULL,
  ...
)
}
\arguments{
\item{data}{Numeric vector of brain map values.}

\item{method}{Character string specifying the null model method.}

\item{n_perm}{Integer number of null permutations to generate.}

\item{distmat}{Distance matrix (required for \code{"burt2020"}, \code{"burt2018"},
and \code{"moran"}).}

\item{coords}{List with \verb{$lh} and \verb{$rh} coordinate matrices
(required for spin methods).}

\item{parcellation}{Integer vector of parcel labels (required for \code{"baum"}
and \code{"cornblath"}).}

\item{seed}{Optional integer seed for reproducibility.}

\item{...}{Additional arguments passed to the specific null method
(e.g. \code{rotation} for spin methods, \code{kernel} for moran/burt2020).}
}
\value{
A \link{null_distribution} object.
}
\description{
Dispatches to the appropriate null model method for generating
spatially-constrained surrogate brain maps.
}
\examples{
data <- rnorm(100)
distmat <- as.matrix(dist(seq_len(100)))
nd <- generate_nulls(data, method = "moran", distmat = distmat, n_perm = 10L)
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resample.R
\name{resample_images}
\alias{resample_images}
\title{Resample brain maps for comparison}
\usage{
resample_images(
  src,
  trg,
  src_space = c("fsaverage", "fsLR"),
  trg_space = c("fsaverage", "fsLR"),
  strategy = c("downsample_only", "transform_to_src", "transform_to_trg",
    "transform_to_alt"),
  alt_space = NULL,
  alt_density = NULL,
  hemisphere = c("left", "right"),
  area_surf_current = NULL,
  area_surf_new = NULL,
  wb_path = NULL,
  verbose = TRUE
)
}
\arguments{
\item{src}{Character, file path to the source GIFTI.}

\item{trg}{Character, file path to the target GIFTI.}

\item{src_space}{Source coordinate space (\code{"fsaverage"} or \code{"fsLR"}).}

\item{trg_space}{Target coordinate space (\code{"fsaverage"} or \code{"fsLR"}).}

\item{strategy}{Resampling strategy. One of \code{"downsample_only"},
\code{"transform_to_src"}, \code{"transform_to_trg"}, or \code{"transform_to_alt"}.}

\item{alt_space}{Alternative space for \code{"transform_to_alt"} strategy.}

\item{alt_density}{Alternative density for \code{"transform_to_alt"} strategy.}

\item{hemisphere}{Which hemispheres: \code{"left"}, \code{"right"}, or both.}

\item{area_surf_current}{Path to a current-resolution area-correction
surface (e.g. midthickness). Passed to \code{\link[=transform_to_space]{transform_to_space()}}.}

\item{area_surf_new}{Path to a target-resolution area-correction surface.
Passed to \code{\link[=transform_to_space]{transform_to_space()}}.}

\item{wb_path}{Path to \code{wb_command} executable.}

\item{verbose}{Logical, print progress messages.}
}
\value{
List with \verb{$src} and \verb{$trg} file paths in the aligned space.
}
\description{
Aligns two brain maps into the same coordinate space and density before
comparison. Supports multiple strategies for choosing the target space.
}
\examples{
\dontrun{
resample_images("src.func.gii", "trg.func.gii",
  src_space = "fsaverage", trg_space = "fsaverage"
)
}
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/surface-graph.R
\name{make_surf_graph}
\alias{make_surf_graph}
\title{Build an igraph from a triangular surface mesh}
\usage{
make_surf_graph(vertices, faces)
}
\arguments{
\item{vertices}{Numeric matrix (n x 3) of vertex coordinates.}

\item{faces}{Integer matrix (m x 3) of face indices (1-indexed).}
}
\value{
An \code{igraph} graph object with weighted edges.
}
\description{
Extracts unique edges from triangular faces, computes Euclidean edge weights,
and returns an igraph graph object suitable for geodesic distance
computation.
}
\examples{
vertices <- matrix(
  c(0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1),
  nrow = 4, byrow = TRUE
)
faces <- matrix(c(1L, 2L, 3L, 2L, 3L, 4L), nrow = 2, byrow = TRUE)
g <- make_surf_graph(vertices, faces)
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nulls-burt2018.R
\name{null_burt2018}
\alias{null_burt2018}
\title{Burt 2018 spatial autoregressive null model}
\usage{
null_burt2018(data, distmat, n_perm = 1000L, seed = NULL)
}
\arguments{
\item{data}{Numeric vector of brain map values.}

\item{distmat}{Distance matrix between parcels/vertices.}

\item{n_perm}{Integer number of null permutations to generate.}

\item{seed}{Optional integer seed for reproducibility.}
}
\value{
A \link{null_distribution} object.
}
\description{
Generates surrogate brain maps using a spatial autoregressive (SAR) model.
Estimates spatial autocorrelation and distance decay parameters, then
generates surrogates by solving the SAR equation and rank-matching to the
original data.
}
\examples{
data <- rnorm(50)
distmat <- as.matrix(dist(matrix(rnorm(100), 50, 2)))
nd <- null_burt2018(data, distmat, n_perm = 10L, seed = 1L)
}
\references{
Burt JB et al. (2018) Nature Neuroscience 21:1251-1259.
doi:10.1038/s41593-018-0195-0
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image-utils.R
\name{fsmorph_to_gifti}
\alias{fsmorph_to_gifti}
\title{Convert FreeSurfer morphometry to GIFTI}
\usage{
fsmorph_to_gifti(morph_path, output_path = NULL)
}
\arguments{
\item{morph_path}{Path to FreeSurfer morphometry file.}

\item{output_path}{Output GIFTI path. If \code{NULL}, replaces the extension
with \code{.func.gii}.}
}
\value{
The output file path (invisibly).
}
\description{
Reads a FreeSurfer morphometry file (\code{.curv}, \code{.thickness}, \code{.sulc}) and
writes a GIFTI func file.
}
\examples{
\dontrun{
fsmorph_to_gifti("lh.thickness")
fsmorph_to_gifti("lh.thickness", "lh.thickness.func.gii")
}
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{compare_maps}
\alias{compare_maps}
\alias{print.neuromaps_enhanced_comparison}
\title{Compare brain maps with spatial null model significance testing}
\usage{
compare_maps(
  x,
  y,
  method = c("pearson", "spearman"),
  null_method = NULL,
  n_perm = 1000L,
  nulls = NULL,
  distmat = NULL,
  coords = NULL,
  seed = NULL,
  na.rm = TRUE,
  verbose = TRUE,
  ...
)

\method{print}{neuromaps_enhanced_comparison}(x, ...)
}
\arguments{
\item{x, y}{Numeric vectors of brain map values, or file paths to
GIFTI/NIfTI files.}

\item{method}{Correlation method: \code{"pearson"} or \code{"spearman"}.}

\item{null_method}{Optional null model method for empirical p-values.
One of \code{"burt2020"}, \code{"moran"}, \code{"spin_vasa"}, \code{"spin_hungarian"},
\code{"alexander_bloch"}, \code{"baum"}, \code{"cornblath"}, \code{"burt2018"},
or \code{NULL} for parametric only.}

\item{n_perm}{Integer number of null permutations.}

\item{nulls}{Pre-computed \link{null_distribution} object for \code{x}.}

\item{distmat}{Distance matrix (passed to null model if needed).}

\item{coords}{Coordinate list (passed to spin null models if needed).}

\item{seed}{Optional integer seed for reproducibility.}

\item{na.rm}{Logical, remove NA values before computing correlation.}

\item{verbose}{Logical, print progress messages.}

\item{...}{Additional arguments passed to \code{\link[=generate_nulls]{generate_nulls()}}.}
}
\value{
A \code{neuromaps_enhanced_comparison} object (inherits
\code{neuromaps_comparison}) with additional fields \code{p_null}, \code{null_method},
\code{null_r}, and \code{n_perm}.
}
\description{
Computes the correlation between two brain maps and optionally tests
significance using a spatial null model to account for spatial
autocorrelation.
}
\examples{
x <- rnorm(50)
y <- x + rnorm(50)
compare_maps(x, y, verbose = FALSE)
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parcellate.R
\name{get_parcel_centroids}
\alias{get_parcel_centroids}
\title{Compute parcel centroids}
\usage{
get_parcel_centroids(
  vertices,
  labels,
  method = c("average", "surface", "geodesic"),
  faces = NULL
)
}
\arguments{
\item{vertices}{Numeric matrix (n x 3) of vertex coordinates.}

\item{labels}{Integer vector of parcel labels. \code{0} and \code{NA} are medial wall.}

\item{method}{Centroid method: \code{"average"} (coordinate means), \code{"surface"}
(vertex closest to the average centroid), or \code{"geodesic"} (vertex
minimizing sum of geodesic distances within parcel).}

\item{faces}{Integer matrix (m x 3) of face indices. Required for
\code{"geodesic"} method.}
}
\value{
Numeric matrix (n_parcels x 3) with rownames set to parcel labels.
}
\description{
Finds the centroid of each parcel using one of three methods.
}
\examples{
vertices <- matrix(rnorm(30), ncol = 3)
labels <- c(1L, 1L, 1L, 2L, 2L, 2L, 3L, 3L, 3L, 3L)
get_parcel_centroids(vertices, labels, method = "average")
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parcellate.R
\name{parcels_to_vertices}
\alias{parcels_to_vertices}
\title{Map parcel data back to vertices}
\usage{
parcels_to_vertices(parcel_data, labels, fill = NA_real_)
}
\arguments{
\item{parcel_data}{Named numeric vector of parcel values
(names match labels).}

\item{labels}{Integer vector of parcel labels. \code{0} and \code{NA} are medial wall.}

\item{fill}{Value for medial wall vertices (default: \code{NA_real_}).}
}
\value{
Numeric vector of \code{length(labels)}.
}
\description{
Expands parcel-level values to a vertex-level vector using parcellation
labels.
}
\examples{
parcel_data <- c("1" = 10, "2" = 20)
labels <- c(1L, 1L, 2L, 2L, 0L)
parcels_to_vertices(parcel_data, labels)
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/neuromaps.R
\name{fetch_neuromaps_annotation}
\alias{fetch_neuromaps_annotation}
\title{Download a neuromaps annotation}
\usage{
fetch_neuromaps_annotation(
  source,
  desc,
  space,
  density = NULL,
  resolution = NULL,
  hemisphere = NULL,
  data_dir = neuromaps_cache_dir(),
  overwrite = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{source}{Data source identifier (e.g. \code{"abagen"}, \code{"beliveau"}).}

\item{desc}{Map descriptor key (e.g. \code{"genepc1"}, \code{"feobv"}).}

\item{space}{Coordinate space (e.g. \code{"fsaverage"}, \code{"MNI152"}, \code{"fsLR"}).}

\item{density}{Surface vertex density (e.g. \code{"10k"}, \code{"164k"}). Mutually
exclusive with \code{resolution}.}

\item{resolution}{Volume voxel resolution (e.g. \code{"1mm"}, \code{"2mm"}). Mutually
exclusive with \code{density}.}

\item{hemisphere}{Hemisphere (\code{"L"} or \code{"R"}).}

\item{data_dir}{Directory for cached downloads. Defaults to
\code{neuromaps_cache_dir()}.}

\item{overwrite}{Re-download even if cached file exists.}

\item{verbose}{Print progress messages.}
}
\value{
Character vector of downloaded file path(s).
}
\description{
Download brain map annotation files from the neuromaps OSF repository.
Surface annotations return two files (left and right hemispheres),
volume annotations return one.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
fetch_neuromaps_annotation("abagen", "genepc1", "fsaverage", density = "10k")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{read_brain_map_values}
\alias{read_brain_map_values}
\title{Read vertex/voxel values from a brain map file}
\usage{
read_brain_map_values(path)
}
\arguments{
\item{path}{Path to a \code{.func.gii} (GIFTI) or \code{.nii.gz} (NIfTI) file.}
}
\value{
A numeric vector of map values.
}
\description{
Reads GIFTI (\code{.func.gii}) or NIfTI (\code{.nii.gz}) files and returns the
values as a numeric vector. Used internally by \code{\link[=compare_maps]{compare_maps()}} when
file paths are passed instead of numeric vectors.
}
\examples{
\dontrun{
read_brain_map_values("cortical_thickness.func.gii")
read_brain_map_values("brain_volume.nii.gz")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transforms.R
\name{check_wb_command}
\alias{check_wb_command}
\title{Check for Connectome Workbench}
\usage{
check_wb_command(wb_path = NULL)
}
\arguments{
\item{wb_path}{Optional explicit path to \code{wb_command}.}
}
\value{
Path to \code{wb_command} executable.
}
\description{
Verifies that \code{wb_command} is available. If \code{wb_path} is \code{NULL},
checks \code{ciftiTools} default and system PATH.
}
\examples{
\dontrun{
check_wb_command()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/neuromaps.R
\name{neuromaps_available}
\alias{neuromaps_available}
\title{List available neuromaps annotations}
\usage{
neuromaps_available(
  source = NULL,
  desc = NULL,
  space = NULL,
  density = NULL,
  resolution = NULL,
  hemisphere = NULL,
  tags = NULL,
  format = NULL,
  refresh = FALSE,
  fixed = FALSE
)
}
\arguments{
\item{source}{Data source identifier (e.g. \code{"abagen"}, \code{"beliveau"}).}

\item{desc}{Map descriptor key (e.g. \code{"genepc1"}, \code{"feobv"}).}

\item{space}{Coordinate space (e.g. \code{"fsaverage"}, \code{"MNI152"}, \code{"fsLR"}).}

\item{density}{Surface vertex density (e.g. \code{"10k"}, \code{"164k"}). Mutually
exclusive with \code{resolution}.}

\item{resolution}{Volume voxel resolution (e.g. \code{"1mm"}, \code{"2mm"}). Mutually
exclusive with \code{density}.}

\item{hemisphere}{Hemisphere (\code{"L"} or \code{"R"}).}

\item{tags}{Character vector of tags. All must match (AND logic).}

\item{format}{Filter by format (\code{"surface"} or \code{"volume"}).}

\item{refresh}{Logical. If \code{TRUE}, forces a fresh download of the
registry data, ignoring any session cache.}

\item{fixed}{Logical. If \code{TRUE}, filter strings are matched literally
rather than as regular expressions.}
}
\value{
A tibble of available annotations with columns: source, desc,
space, den, res, hemi, format, fname, full_desc, tags, N, age.
}
\description{
Query the neuromaps registry to see which brain map annotations are
available for download. Data is fetched from the
\href{https://github.com/netneurolab/neuromaps}{neuromaps} project's GitHub
repository and cached for the session.
}
\details{
All string filter parameters (\code{source}, \code{desc}, \code{space}, \code{density},
\code{resolution}, \code{hemisphere}, \code{format}) are treated as \strong{R regular
expressions} and matched with \code{\link[=grepl]{grepl()}}. For example,
\code{source = "^beliveau$"} matches exactly, while \code{source = "bel"} matches
any source containing \code{"bel"}. Set \code{fixed = TRUE} for literal string
matching. The \code{tags} parameter always uses exact matching (AND logic).

When used in \code{\link[=fetch_neuromaps_annotation]{fetch_neuromaps_annotation()}}, \code{source}, \code{desc}, and \code{space}
are exact matches.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
neuromaps_available()
neuromaps_available(source = "beliveau")
neuromaps_available(tags = "pet")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image-utils.R
\name{annot_to_gifti}
\alias{annot_to_gifti}
\title{Convert FreeSurfer annotation to GIFTI}
\usage{
annot_to_gifti(annot_path, output_path = NULL)
}
\arguments{
\item{annot_path}{Path to FreeSurfer \code{.annot} file.}

\item{output_path}{Output GIFTI path. If \code{NULL}, replaces the extension
with \code{.label.gii}.}
}
\value{
The output file path (invisibly).
}
\description{
Reads a FreeSurfer \code{.annot} file and writes a GIFTI label file.
}
\examples{
\dontrun{
annot_to_gifti("lh.aparc.annot")
annot_to_gifti("lh.aparc.annot", "lh.aparc.label.gii")
}
}
\references{
Markello RD et al. (2022) Nature Methods 19:1472-1480.
doi:10.1038/s41592-022-01625-w
}
